// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: rpc_messages.proto

#include "rpc_messages.pb.h"

#include <algorithm>

#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/extension_set.h>
#include <google/protobuf/wire_format_lite.h>
#include <google/protobuf/descriptor.h>
#include <google/protobuf/generated_message_reflection.h>
#include <google/protobuf/reflection_ops.h>
#include <google/protobuf/wire_format.h>
// @@protoc_insertion_point(includes)
#include <google/protobuf/port_def.inc>

PROTOBUF_PRAGMA_INIT_SEG

namespace _pb = ::PROTOBUF_NAMESPACE_ID;
namespace _pbi = _pb::internal;

PROTOBUF_CONSTEXPR call_packet_info::call_packet_info(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.xcl_api_)*/0u} {}
struct call_packet_infoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR call_packet_infoDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~call_packet_infoDefaultTypeInternal() {}
  union {
    call_packet_info _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 call_packet_infoDefaultTypeInternal _call_packet_info_default_instance_;
PROTOBUF_CONSTEXPR response_packet_info::response_packet_info(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.xcl_api_)*/0u} {}
struct response_packet_infoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR response_packet_infoDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~response_packet_infoDefaultTypeInternal() {}
  union {
    response_packet_info _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 response_packet_infoDefaultTypeInternal _response_packet_info_default_instance_;
PROTOBUF_CONSTEXPR xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.name_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.value_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}} {}
struct xclSetEnvironment_call_namevaluepairDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_call_namevaluepairDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_call_namevaluepairDefaultTypeInternal() {}
  union {
    xclSetEnvironment_call_namevaluepair _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_call_namevaluepairDefaultTypeInternal _xclSetEnvironment_call_namevaluepair_default_instance_;
PROTOBUF_CONSTEXPR xclSetEnvironment_call::xclSetEnvironment_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_.environment_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}} {}
struct xclSetEnvironment_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_callDefaultTypeInternal() {}
  union {
    xclSetEnvironment_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_callDefaultTypeInternal _xclSetEnvironment_call_default_instance_;
PROTOBUF_CONSTEXPR xclSetEnvironment_response::xclSetEnvironment_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclSetEnvironment_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_responseDefaultTypeInternal() {}
  union {
    xclSetEnvironment_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_responseDefaultTypeInternal _xclSetEnvironment_response_default_instance_;
PROTOBUF_CONSTEXPR xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclLoadBitstream_call_ddrbankDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_call_ddrbankDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_call_ddrbankDefaultTypeInternal() {}
  union {
    xclLoadBitstream_call_ddrbank _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_call_ddrbankDefaultTypeInternal _xclLoadBitstream_call_ddrbank_default_instance_;
PROTOBUF_CONSTEXPR xclLoadBitstream_call::xclLoadBitstream_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ddrbanks_)*/{}
  , /*decltype(_impl_.xmlfile_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.dlopenfilename_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.devicename_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.devicedirectory_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.binarydirectory_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.verbose_)*/false} {}
struct xclLoadBitstream_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_callDefaultTypeInternal() {}
  union {
    xclLoadBitstream_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_callDefaultTypeInternal _xclLoadBitstream_call_default_instance_;
PROTOBUF_CONSTEXPR xclLoadBitstream_response::xclLoadBitstream_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclLoadBitstream_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_responseDefaultTypeInternal() {}
  union {
    xclLoadBitstream_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_responseDefaultTypeInternal _xclLoadBitstream_response_default_instance_;
PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ddraddress_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.peertopeer_)*/false} {}
struct xclAllocDeviceBuffer_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclAllocDeviceBuffer_callDefaultTypeInternal() {}
  union {
    xclAllocDeviceBuffer_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclAllocDeviceBuffer_callDefaultTypeInternal _xclAllocDeviceBuffer_call_default_instance_;
PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.filename_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclAllocDeviceBuffer_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclAllocDeviceBuffer_responseDefaultTypeInternal() {}
  union {
    xclAllocDeviceBuffer_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclAllocDeviceBuffer_responseDefaultTypeInternal _xclAllocDeviceBuffer_response_default_instance_;
PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ddraddress_)*/uint64_t{0u}} {}
struct xclFreeDeviceBuffer_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclFreeDeviceBuffer_callDefaultTypeInternal() {}
  union {
    xclFreeDeviceBuffer_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclFreeDeviceBuffer_callDefaultTypeInternal _xclFreeDeviceBuffer_call_default_instance_;
PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclFreeDeviceBuffer_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclFreeDeviceBuffer_responseDefaultTypeInternal() {}
  union {
    xclFreeDeviceBuffer_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclFreeDeviceBuffer_responseDefaultTypeInternal _xclFreeDeviceBuffer_response_default_instance_;
PROTOBUF_CONSTEXPR xclClose_call::xclClose_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.xcldevicehandle_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.closeall_)*/false} {}
struct xclClose_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclClose_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclClose_callDefaultTypeInternal() {}
  union {
    xclClose_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclClose_callDefaultTypeInternal _xclClose_call_default_instance_;
PROTOBUF_CONSTEXPR xclClose_response::xclClose_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclClose_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclClose_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclClose_responseDefaultTypeInternal() {}
  union {
    xclClose_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclClose_responseDefaultTypeInternal _xclClose_response_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.xcldevicehandle_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.src_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.dest_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.seek_)*/uint64_t{0u}
  , /*decltype(_impl_.space_)*/0u} {}
struct xclCopyBufferHost2Device_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferHost2Device_callDefaultTypeInternal() {}
  union {
    xclCopyBufferHost2Device_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferHost2Device_callDefaultTypeInternal _xclCopyBufferHost2Device_call_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclCopyBufferHost2Device_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferHost2Device_responseDefaultTypeInternal() {}
  union {
    xclCopyBufferHost2Device_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferHost2Device_responseDefaultTypeInternal _xclCopyBufferHost2Device_response_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.xcldevicehandle_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.dest_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.src_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.skip_)*/uint64_t{0u}
  , /*decltype(_impl_.space_)*/0u} {}
struct xclCopyBufferDevice2Host_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferDevice2Host_callDefaultTypeInternal() {}
  union {
    xclCopyBufferDevice2Host_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferDevice2Host_callDefaultTypeInternal _xclCopyBufferDevice2Host_call_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dest_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclCopyBufferDevice2Host_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferDevice2Host_responseDefaultTypeInternal() {}
  union {
    xclCopyBufferDevice2Host_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferDevice2Host_responseDefaultTypeInternal _xclCopyBufferDevice2Host_response_default_instance_;
PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal() {}
  union {
    xclWriteAddrSpaceDeviceRam_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal _xclWriteAddrSpaceDeviceRam_call_default_instance_;
PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal() {}
  union {
    xclWriteAddrSpaceDeviceRam_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal _xclWriteAddrSpaceDeviceRam_response_default_instance_;
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.name_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u} {}
struct xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_call_kernelInfo _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal _xclWriteAddrKernelCtrl_call_kernelInfo_default_instance_;
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.kernel_info_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclWriteAddrKernelCtrl_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_callDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_callDefaultTypeInternal _xclWriteAddrKernelCtrl_call_default_instance_;
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclWriteAddrKernelCtrl_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_responseDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_responseDefaultTypeInternal _xclWriteAddrKernelCtrl_response_default_instance_;
PROTOBUF_CONSTEXPR xclRegWrite_call::xclRegWrite_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.baseaddress_)*/0u
  , /*decltype(_impl_.offset_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclRegWrite_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegWrite_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegWrite_callDefaultTypeInternal() {}
  union {
    xclRegWrite_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegWrite_callDefaultTypeInternal _xclRegWrite_call_default_instance_;
PROTOBUF_CONSTEXPR xclRegWrite_response::xclRegWrite_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclRegWrite_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegWrite_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegWrite_responseDefaultTypeInternal() {}
  union {
    xclRegWrite_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegWrite_responseDefaultTypeInternal _xclRegWrite_response_default_instance_;
PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclReadAddrSpaceDeviceRam_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrSpaceDeviceRam_callDefaultTypeInternal() {}
  union {
    xclReadAddrSpaceDeviceRam_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrSpaceDeviceRam_callDefaultTypeInternal _xclReadAddrSpaceDeviceRam_call_default_instance_;
PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal() {}
  union {
    xclReadAddrSpaceDeviceRam_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal _xclReadAddrSpaceDeviceRam_response_default_instance_;
PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclReadAddrKernelCtrl_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrKernelCtrl_callDefaultTypeInternal() {}
  union {
    xclReadAddrKernelCtrl_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrKernelCtrl_callDefaultTypeInternal _xclReadAddrKernelCtrl_call_default_instance_;
PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclReadAddrKernelCtrl_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrKernelCtrl_responseDefaultTypeInternal() {}
  union {
    xclReadAddrKernelCtrl_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrKernelCtrl_responseDefaultTypeInternal _xclReadAddrKernelCtrl_response_default_instance_;
PROTOBUF_CONSTEXPR xclRegRead_call::xclRegRead_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.baseaddress_)*/0u
  , /*decltype(_impl_.offset_)*/0u
  , /*decltype(_impl_.size_)*/0u
  , /*decltype(_impl_.pf_id_)*/0u
  , /*decltype(_impl_.bar_id_)*/0u} {}
struct xclRegRead_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegRead_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegRead_callDefaultTypeInternal() {}
  union {
    xclRegRead_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegRead_callDefaultTypeInternal _xclRegRead_call_default_instance_;
PROTOBUF_CONSTEXPR xclRegRead_response::xclRegRead_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclRegRead_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegRead_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegRead_responseDefaultTypeInternal() {}
  union {
    xclRegRead_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegRead_responseDefaultTypeInternal _xclRegRead_response_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slotname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.accel_)*/false} {}
struct xclPerfMonReadCounters_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_callDefaultTypeInternal _xclPerfMonReadCounters_call_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.wr_byte_count_)*/0u
  , /*decltype(_impl_.wr_trans_count_)*/0u
  , /*decltype(_impl_.total_wr_latency_)*/0u
  , /*decltype(_impl_.rd_byte_count_)*/0u
  , /*decltype(_impl_.rd_trans_count_)*/0u
  , /*decltype(_impl_.total_rd_latency_)*/0u} {}
struct xclPerfMonReadCounters_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_responseDefaultTypeInternal _xclPerfMonReadCounters_response_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slotname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.ack_)*/false
  , /*decltype(_impl_.accel_)*/false} {}
struct xclPerfMonGetTraceCount_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonGetTraceCount_callDefaultTypeInternal() {}
  union {
    xclPerfMonGetTraceCount_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonGetTraceCount_callDefaultTypeInternal _xclPerfMonGetTraceCount_call_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.no_of_samples_)*/0u} {}
struct xclPerfMonGetTraceCount_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonGetTraceCount_responseDefaultTypeInternal() {}
  union {
    xclPerfMonGetTraceCount_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonGetTraceCount_responseDefaultTypeInternal _xclPerfMonGetTraceCount_response_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slotname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.ack_)*/false
  , /*decltype(_impl_.accel_)*/false} {}
struct xclPerfMonReadTrace_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_callDefaultTypeInternal _xclPerfMonReadTrace_call_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.awlen_)*/0u
  , /*decltype(_impl_.arlen_)*/0u
  , /*decltype(_impl_.eventflags_)*/0u
  , /*decltype(_impl_.timestamp_)*/0u
  , /*decltype(_impl_.host_timestamp_)*/uint64_t{0u}
  , /*decltype(_impl_.rd_bytes_)*/0u
  , /*decltype(_impl_.wr_bytes_)*/0u} {}
struct xclPerfMonReadTrace_response_eventsDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response_eventsDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_response_eventsDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_response_events _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_response_eventsDefaultTypeInternal _xclPerfMonReadTrace_response_events_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_.output_data_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}} {}
struct xclPerfMonReadTrace_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_responseDefaultTypeInternal _xclPerfMonReadTrace_response_default_instance_;
PROTOBUF_CONSTEXPR ddr_mem_msg::ddr_mem_msg(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.strb_data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.fullsize_)*/uint64_t{0u}} {}
struct ddr_mem_msgDefaultTypeInternal {
  PROTOBUF_CONSTEXPR ddr_mem_msgDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~ddr_mem_msgDefaultTypeInternal() {}
  union {
    ddr_mem_msg _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 ddr_mem_msgDefaultTypeInternal _ddr_mem_msg_default_instance_;
PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGetDeviceTimestamp_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDeviceTimestamp_callDefaultTypeInternal() {}
  union {
    xclGetDeviceTimestamp_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDeviceTimestamp_callDefaultTypeInternal _xclGetDeviceTimestamp_call_default_instance_;
PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.device_timestamp_)*/uint64_t{0u}} {}
struct xclGetDeviceTimestamp_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDeviceTimestamp_responseDefaultTypeInternal() {}
  union {
    xclGetDeviceTimestamp_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDeviceTimestamp_responseDefaultTypeInternal _xclGetDeviceTimestamp_response_default_instance_;
PROTOBUF_CONSTEXPR xclReadBusStatus_call::xclReadBusStatus_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slot_n_)*/0} {}
struct xclReadBusStatus_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadBusStatus_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadBusStatus_callDefaultTypeInternal() {}
  union {
    xclReadBusStatus_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadBusStatus_callDefaultTypeInternal _xclReadBusStatus_call_default_instance_;
PROTOBUF_CONSTEXPR xclReadBusStatus_response::xclReadBusStatus_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.uncomplete_rd_cycles_)*/uint64_t{0u}
  , /*decltype(_impl_.uncomplete_wr_cycles_)*/uint64_t{0u}
  , /*decltype(_impl_.idle_bus_cycles_)*/uint64_t{0u}
  , /*decltype(_impl_.uncomplete_rd_id_)*/uint64_t{0u}
  , /*decltype(_impl_.uncomplete_wr_id_)*/uint64_t{0u}} {}
struct xclReadBusStatus_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadBusStatus_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadBusStatus_responseDefaultTypeInternal() {}
  union {
    xclReadBusStatus_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadBusStatus_responseDefaultTypeInternal _xclReadBusStatus_response_default_instance_;
PROTOBUF_CONSTEXPR xclGetDebugMessages_call::xclGetDebugMessages_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false
  , /*decltype(_impl_.force_)*/false} {}
struct xclGetDebugMessages_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDebugMessages_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDebugMessages_callDefaultTypeInternal() {}
  union {
    xclGetDebugMessages_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDebugMessages_callDefaultTypeInternal _xclGetDebugMessages_call_default_instance_;
PROTOBUF_CONSTEXPR xclGetDebugMessages_response::xclGetDebugMessages_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.display_msgs_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.log_msgs_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.stop_msgs_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}} {}
struct xclGetDebugMessages_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDebugMessages_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDebugMessages_responseDefaultTypeInternal() {}
  union {
    xclGetDebugMessages_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDebugMessages_responseDefaultTypeInternal _xclGetDebugMessages_response_default_instance_;
PROTOBUF_CONSTEXPR xclImportBO_call::xclImportBO_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dst_filename_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.offset_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclImportBO_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclImportBO_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclImportBO_callDefaultTypeInternal() {}
  union {
    xclImportBO_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclImportBO_callDefaultTypeInternal _xclImportBO_call_default_instance_;
PROTOBUF_CONSTEXPR xclImportBO_response::xclImportBO_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclImportBO_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclImportBO_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclImportBO_responseDefaultTypeInternal() {}
  union {
    xclImportBO_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclImportBO_responseDefaultTypeInternal _xclImportBO_response_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBO_call::xclCopyBO_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dst_filename_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.src_handle_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.src_offset_)*/uint64_t{0u}
  , /*decltype(_impl_.dst_offset_)*/uint64_t{0u}} {}
struct xclCopyBO_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBO_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBO_callDefaultTypeInternal() {}
  union {
    xclCopyBO_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBO_callDefaultTypeInternal _xclCopyBO_call_default_instance_;
PROTOBUF_CONSTEXPR xclCopyBO_response::xclCopyBO_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclCopyBO_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBO_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBO_responseDefaultTypeInternal() {}
  union {
    xclCopyBO_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBO_responseDefaultTypeInternal _xclCopyBO_response_default_instance_;
PROTOBUF_CONSTEXPR xclCreateQueue_call::xclCreateQueue_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.write_)*/false
  , /*decltype(_impl_.type_)*/0u
  , /*decltype(_impl_.state_)*/0u
  , /*decltype(_impl_.route_)*/0u
  , /*decltype(_impl_.flow_)*/uint64_t{0u}
  , /*decltype(_impl_.qsize_)*/0u
  , /*decltype(_impl_.desc_size_)*/0u
  , /*decltype(_impl_.flags_)*/uint64_t{0u}} {}
struct xclCreateQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCreateQueue_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCreateQueue_callDefaultTypeInternal() {}
  union {
    xclCreateQueue_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCreateQueue_callDefaultTypeInternal _xclCreateQueue_call_default_instance_;
PROTOBUF_CONSTEXPR xclCreateQueue_response::xclCreateQueue_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.q_handle_)*/uint64_t{0u}} {}
struct xclCreateQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCreateQueue_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCreateQueue_responseDefaultTypeInternal() {}
  union {
    xclCreateQueue_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCreateQueue_responseDefaultTypeInternal _xclCreateQueue_response_default_instance_;
PROTOBUF_CONSTEXPR xclWriteQueue_call::xclWriteQueue_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.src_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.q_handle_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.req_)*/uint64_t{0u}
  , /*decltype(_impl_.nonblocking_)*/false
  , /*decltype(_impl_.eot_)*/false} {}
struct xclWriteQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteQueue_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteQueue_callDefaultTypeInternal() {}
  union {
    xclWriteQueue_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteQueue_callDefaultTypeInternal _xclWriteQueue_call_default_instance_;
PROTOBUF_CONSTEXPR xclWriteQueue_response::xclWriteQueue_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.written_size_)*/uint64_t{0u}} {}
struct xclWriteQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteQueue_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteQueue_responseDefaultTypeInternal() {}
  union {
    xclWriteQueue_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteQueue_responseDefaultTypeInternal _xclWriteQueue_response_default_instance_;
PROTOBUF_CONSTEXPR xclReadQueue_call::xclReadQueue_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dest_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.q_handle_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.req_)*/uint64_t{0u}
  , /*decltype(_impl_.nonblocking_)*/false
  , /*decltype(_impl_.eot_)*/false} {}
struct xclReadQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadQueue_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadQueue_callDefaultTypeInternal() {}
  union {
    xclReadQueue_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadQueue_callDefaultTypeInternal _xclReadQueue_call_default_instance_;
PROTOBUF_CONSTEXPR xclReadQueue_response::xclReadQueue_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dest_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclReadQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadQueue_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadQueue_responseDefaultTypeInternal() {}
  union {
    xclReadQueue_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadQueue_responseDefaultTypeInternal _xclReadQueue_response_default_instance_;
PROTOBUF_CONSTEXPR xclPollCompletion_call::xclPollCompletion_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.req_)*/uint64_t{0u}} {}
struct xclPollCompletion_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_callDefaultTypeInternal() {}
  union {
    xclPollCompletion_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_callDefaultTypeInternal _xclPollCompletion_call_default_instance_;
PROTOBUF_CONSTEXPR xclPollCompletion_response_request::xclPollCompletion_response_request(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.dest_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclPollCompletion_response_requestDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_response_requestDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_response_requestDefaultTypeInternal() {}
  union {
    xclPollCompletion_response_request _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_response_requestDefaultTypeInternal _xclPollCompletion_response_request_default_instance_;
PROTOBUF_CONSTEXPR xclPollCompletion_response::xclPollCompletion_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_.fullrequest_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}} {}
struct xclPollCompletion_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_responseDefaultTypeInternal() {}
  union {
    xclPollCompletion_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_responseDefaultTypeInternal _xclPollCompletion_response_default_instance_;
PROTOBUF_CONSTEXPR xclDestroyQueue_call::xclDestroyQueue_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.q_handle_)*/uint64_t{0u}} {}
struct xclDestroyQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclDestroyQueue_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclDestroyQueue_callDefaultTypeInternal() {}
  union {
    xclDestroyQueue_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclDestroyQueue_callDefaultTypeInternal _xclDestroyQueue_call_default_instance_;
PROTOBUF_CONSTEXPR xclDestroyQueue_response::xclDestroyQueue_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.success_)*/false} {}
struct xclDestroyQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclDestroyQueue_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclDestroyQueue_responseDefaultTypeInternal() {}
  union {
    xclDestroyQueue_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclDestroyQueue_responseDefaultTypeInternal _xclDestroyQueue_response_default_instance_;
PROTOBUF_CONSTEXPR xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.tag_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.arg_)*/uint64_t{0u}
  , /*decltype(_impl_.flow_)*/uint64_t{0u}} {}
struct xclSetupInstance_call_argflowpairDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_call_argflowpairDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_call_argflowpairDefaultTypeInternal() {}
  union {
    xclSetupInstance_call_argflowpair _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_call_argflowpairDefaultTypeInternal _xclSetupInstance_call_argflowpair_default_instance_;
PROTOBUF_CONSTEXPR xclSetupInstance_call::xclSetupInstance_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.setup_)*/{}
  , /*decltype(_impl_.route_)*/uint64_t{0u}} {}
struct xclSetupInstance_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_callDefaultTypeInternal() {}
  union {
    xclSetupInstance_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_callDefaultTypeInternal _xclSetupInstance_call_default_instance_;
PROTOBUF_CONSTEXPR xclSetupInstance_response::xclSetupInstance_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.success_)*/false} {}
struct xclSetupInstance_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_responseDefaultTypeInternal() {}
  union {
    xclSetupInstance_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_responseDefaultTypeInternal _xclSetupInstance_response_default_instance_;
PROTOBUF_CONSTEXPR swemuDriverVersion_call::swemuDriverVersion_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.version_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}} {}
struct swemuDriverVersion_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR swemuDriverVersion_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~swemuDriverVersion_callDefaultTypeInternal() {}
  union {
    swemuDriverVersion_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 swemuDriverVersion_callDefaultTypeInternal _swemuDriverVersion_call_default_instance_;
PROTOBUF_CONSTEXPR swemuDriverVersion_response::swemuDriverVersion_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.success_)*/false} {}
struct swemuDriverVersion_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR swemuDriverVersion_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~swemuDriverVersion_responseDefaultTypeInternal() {}
  union {
    swemuDriverVersion_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 swemuDriverVersion_responseDefaultTypeInternal _swemuDriverVersion_response_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slotname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}} {}
struct xclPerfMonReadCounters_Streaming_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_Streaming_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_Streaming_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_Streaming_callDefaultTypeInternal _xclPerfMonReadCounters_Streaming_call_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.str_num_tranx_)*/uint64_t{0u}
  , /*decltype(_impl_.str_data_bytes_)*/uint64_t{0u}
  , /*decltype(_impl_.str_busy_cycles_)*/uint64_t{0u}
  , /*decltype(_impl_.str_stall_cycles_)*/uint64_t{0u}
  , /*decltype(_impl_.str_starve_cycles_)*/uint64_t{0u}} {}
struct xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_Streaming_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal _xclPerfMonReadCounters_Streaming_response_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.slotname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclPerfMonReadTrace_Streaming_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_callDefaultTypeInternal _xclPerfMonReadTrace_Streaming_call_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.eventflags_)*/0u
  , /*decltype(_impl_.timestamp_)*/0u
  , /*decltype(_impl_.host_timestamp_)*/uint64_t{0u}} {}
struct xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_response_events _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal _xclPerfMonReadTrace_Streaming_response_events_default_instance_;
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_.output_data_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}} {}
struct xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal _xclPerfMonReadTrace_Streaming_response_default_instance_;
PROTOBUF_CONSTEXPR xclSlaveReadReq_call::xclSlaveReadReq_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u} {}
struct xclSlaveReadReq_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveReadReq_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveReadReq_callDefaultTypeInternal() {}
  union {
    xclSlaveReadReq_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveReadReq_callDefaultTypeInternal _xclSlaveReadReq_call_default_instance_;
PROTOBUF_CONSTEXPR xclSlaveReadReq_response::xclSlaveReadReq_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclSlaveReadReq_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveReadReq_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveReadReq_responseDefaultTypeInternal() {}
  union {
    xclSlaveReadReq_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveReadReq_responseDefaultTypeInternal _xclSlaveReadReq_response_default_instance_;
PROTOBUF_CONSTEXPR xclSlaveWriteReq_call::xclSlaveWriteReq_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.data_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.addr_)*/uint64_t{0u}
  , /*decltype(_impl_.size_)*/0u} {}
struct xclSlaveWriteReq_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveWriteReq_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveWriteReq_callDefaultTypeInternal() {}
  union {
    xclSlaveWriteReq_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveWriteReq_callDefaultTypeInternal _xclSlaveWriteReq_call_default_instance_;
PROTOBUF_CONSTEXPR xclSlaveWriteReq_response::xclSlaveWriteReq_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclSlaveWriteReq_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveWriteReq_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveWriteReq_responseDefaultTypeInternal() {}
  union {
    xclSlaveWriteReq_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveWriteReq_responseDefaultTypeInternal _xclSlaveWriteReq_response_default_instance_;
PROTOBUF_CONSTEXPR xclInterruptOccured_call::xclInterruptOccured_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.interrupt_line_)*/0u} {}
struct xclInterruptOccured_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclInterruptOccured_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclInterruptOccured_callDefaultTypeInternal() {}
  union {
    xclInterruptOccured_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclInterruptOccured_callDefaultTypeInternal _xclInterruptOccured_call_default_instance_;
PROTOBUF_CONSTEXPR xclInterruptOccured_response::xclInterruptOccured_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.valid_)*/false} {}
struct xclInterruptOccured_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclInterruptOccured_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclInterruptOccured_responseDefaultTypeInternal() {}
  union {
    xclInterruptOccured_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclInterruptOccured_responseDefaultTypeInternal _xclInterruptOccured_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphInit_call::xclGraphInit_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.graphname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphInit_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphInit_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphInit_callDefaultTypeInternal() {}
  union {
    xclGraphInit_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphInit_callDefaultTypeInternal _xclGraphInit_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphInit_response::xclGraphInit_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphInit_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphInit_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphInit_responseDefaultTypeInternal() {}
  union {
    xclGraphInit_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphInit_responseDefaultTypeInternal _xclGraphInit_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphRun_call::xclGraphRun_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gh_)*/0u
  , /*decltype(_impl_.iterations_)*/0u} {}
struct xclGraphRun_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphRun_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphRun_callDefaultTypeInternal() {}
  union {
    xclGraphRun_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphRun_callDefaultTypeInternal _xclGraphRun_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphRun_response::xclGraphRun_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphRun_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphRun_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphRun_responseDefaultTypeInternal() {}
  union {
    xclGraphRun_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphRun_responseDefaultTypeInternal _xclGraphRun_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphWait_call::xclGraphWait_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphWait_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphWait_callDefaultTypeInternal() {}
  union {
    xclGraphWait_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphWait_callDefaultTypeInternal _xclGraphWait_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphWait_response::xclGraphWait_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphWait_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphWait_responseDefaultTypeInternal() {}
  union {
    xclGraphWait_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphWait_responseDefaultTypeInternal _xclGraphWait_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphEnd_call::xclGraphEnd_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphEnd_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphEnd_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphEnd_callDefaultTypeInternal() {}
  union {
    xclGraphEnd_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphEnd_callDefaultTypeInternal _xclGraphEnd_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphEnd_response::xclGraphEnd_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/0u} {}
struct xclGraphEnd_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphEnd_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphEnd_responseDefaultTypeInternal() {}
  union {
    xclGraphEnd_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphEnd_responseDefaultTypeInternal _xclGraphEnd_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.portname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.buffer_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphUpdateRTP_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphUpdateRTP_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphUpdateRTP_callDefaultTypeInternal() {}
  union {
    xclGraphUpdateRTP_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphUpdateRTP_callDefaultTypeInternal _xclGraphUpdateRTP_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclGraphUpdateRTP_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphUpdateRTP_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphUpdateRTP_responseDefaultTypeInternal() {}
  union {
    xclGraphUpdateRTP_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphUpdateRTP_responseDefaultTypeInternal _xclGraphUpdateRTP_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphReadRTP_call::xclGraphReadRTP_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.portname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.buffer_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphReadRTP_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphReadRTP_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphReadRTP_callDefaultTypeInternal() {}
  union {
    xclGraphReadRTP_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphReadRTP_callDefaultTypeInternal _xclGraphReadRTP_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphReadRTP_response::xclGraphReadRTP_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.buffer_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}} {}
struct xclGraphReadRTP_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphReadRTP_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphReadRTP_responseDefaultTypeInternal() {}
  union {
    xclGraphReadRTP_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphReadRTP_responseDefaultTypeInternal _xclGraphReadRTP_response_default_instance_;
PROTOBUF_CONSTEXPR xclSyncBOAIENB_call::xclSyncBOAIENB_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gmioname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.size_)*/uint64_t{0u}
  , /*decltype(_impl_.offset_)*/uint64_t{0u}
  , /*decltype(_impl_.boh_)*/uint64_t{0u}
  , /*decltype(_impl_.dir_)*/0u} {}
struct xclSyncBOAIENB_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSyncBOAIENB_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSyncBOAIENB_callDefaultTypeInternal() {}
  union {
    xclSyncBOAIENB_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSyncBOAIENB_callDefaultTypeInternal _xclSyncBOAIENB_call_default_instance_;
PROTOBUF_CONSTEXPR xclSyncBOAIENB_response::xclSyncBOAIENB_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclSyncBOAIENB_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSyncBOAIENB_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSyncBOAIENB_responseDefaultTypeInternal() {}
  union {
    xclSyncBOAIENB_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSyncBOAIENB_responseDefaultTypeInternal _xclSyncBOAIENB_response_default_instance_;
PROTOBUF_CONSTEXPR xclGMIOWait_call::xclGMIOWait_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gmioname_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}} {}
struct xclGMIOWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGMIOWait_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGMIOWait_callDefaultTypeInternal() {}
  union {
    xclGMIOWait_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGMIOWait_callDefaultTypeInternal _xclGMIOWait_call_default_instance_;
PROTOBUF_CONSTEXPR xclGMIOWait_response::xclGMIOWait_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGMIOWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGMIOWait_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGMIOWait_responseDefaultTypeInternal() {}
  union {
    xclGMIOWait_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGMIOWait_responseDefaultTypeInternal _xclGMIOWait_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphTimedWait_call::xclGraphTimedWait_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.cycletimeout_)*/uint64_t{0u}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphTimedWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedWait_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedWait_callDefaultTypeInternal() {}
  union {
    xclGraphTimedWait_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedWait_callDefaultTypeInternal _xclGraphTimedWait_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphTimedWait_response::xclGraphTimedWait_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphTimedWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedWait_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedWait_responseDefaultTypeInternal() {}
  union {
    xclGraphTimedWait_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedWait_responseDefaultTypeInternal _xclGraphTimedWait_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphTimedEnd_call::xclGraphTimedEnd_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.cycletimeout_)*/uint64_t{0u}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphTimedEnd_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedEnd_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedEnd_callDefaultTypeInternal() {}
  union {
    xclGraphTimedEnd_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedEnd_callDefaultTypeInternal _xclGraphTimedEnd_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphTimedEnd_response::xclGraphTimedEnd_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphTimedEnd_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedEnd_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedEnd_responseDefaultTypeInternal() {}
  union {
    xclGraphTimedEnd_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedEnd_responseDefaultTypeInternal _xclGraphTimedEnd_response_default_instance_;
PROTOBUF_CONSTEXPR xclGraphResume_call::xclGraphResume_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.gh_)*/0u} {}
struct xclGraphResume_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphResume_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphResume_callDefaultTypeInternal() {}
  union {
    xclGraphResume_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphResume_callDefaultTypeInternal _xclGraphResume_call_default_instance_;
PROTOBUF_CONSTEXPR xclGraphResume_response::xclGraphResume_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclGraphResume_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphResume_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphResume_responseDefaultTypeInternal() {}
  union {
    xclGraphResume_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphResume_responseDefaultTypeInternal _xclGraphResume_response_default_instance_;
PROTOBUF_CONSTEXPR xclLoadXclbinContent_call::xclLoadXclbinContent_call(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.xmlbuff_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.sharedbin_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.emuldata_)*/{&::_pbi::fixed_address_empty_string, ::_pbi::ConstantInitialized{}}
  , /*decltype(_impl_.xmlbuffsize_)*/uint64_t{0u}
  , /*decltype(_impl_.sharedbinsize_)*/uint64_t{0u}
  , /*decltype(_impl_.emuldatasize_)*/uint64_t{0u}
  , /*decltype(_impl_.keepdir_)*/false} {}
struct xclLoadXclbinContent_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadXclbinContent_callDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadXclbinContent_callDefaultTypeInternal() {}
  union {
    xclLoadXclbinContent_call _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadXclbinContent_callDefaultTypeInternal _xclLoadXclbinContent_call_default_instance_;
PROTOBUF_CONSTEXPR xclLoadXclbinContent_response::xclLoadXclbinContent_response(
    ::_pbi::ConstantInitialized): _impl_{
    /*decltype(_impl_._has_bits_)*/{}
  , /*decltype(_impl_._cached_size_)*/{}
  , /*decltype(_impl_.ack_)*/false} {}
struct xclLoadXclbinContent_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadXclbinContent_responseDefaultTypeInternal()
      : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadXclbinContent_responseDefaultTypeInternal() {}
  union {
    xclLoadXclbinContent_response _instance;
  };
};
PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadXclbinContent_responseDefaultTypeInternal _xclLoadXclbinContent_response_default_instance_;
static ::_pb::Metadata file_level_metadata_rpc_5fmessages_2eproto[100];
static constexpr ::_pb::EnumDescriptor const** file_level_enum_descriptors_rpc_5fmessages_2eproto = nullptr;
static constexpr ::_pb::ServiceDescriptor const** file_level_service_descriptors_rpc_5fmessages_2eproto = nullptr;

const uint32_t TableStruct_rpc_5fmessages_2eproto::offsets[] PROTOBUF_SECTION_VARIABLE(protodesc_cold) = {
  PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::call_packet_info, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_.xcl_api_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::response_packet_info, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_.xcl_api_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_.name_),
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_.value_),
  0,
  1,
  ~0u,  // no _has_bits_
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call, _impl_.environment_),
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _impl_.size_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.xmlfile_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.dlopenfilename_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.devicename_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.devicedirectory_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.binarydirectory_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.verbose_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.ddrbanks_),
  0,
  1,
  2,
  3,
  4,
  5,
  ~0u,
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.ddraddress_),
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.peertopeer_),
  0,
  1,
  2,
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_.ack_),
  PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_.filename_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _impl_.ddraddress_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclClose_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_.xcldevicehandle_),
  PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_.closeall_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclClose_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclClose_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclClose_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.xcldevicehandle_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.dest_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.src_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.seek_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.space_),
  0,
  2,
  1,
  3,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _impl_.size_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.xcldevicehandle_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.dest_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.src_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.skip_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.space_),
  0,
  1,
  2,
  3,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_.dest_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_),
  1,
  2,
  0,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.name_),
  1,
  2,
  0,
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.kernel_info_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.bar_id_),
  1,
  2,
  0,
  ~0u,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.baseaddress_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.offset_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.bar_id_),
  1,
  2,
  0,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_),
  1,
  2,
  0,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_.valid_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_.data_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.bar_id_),
  1,
  2,
  0,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_.valid_),
  PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_.data_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.baseaddress_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.offset_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.pf_id_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.bar_id_),
  1,
  2,
  3,
  0,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_.valid_),
  PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_.data_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_.slotname_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_.accel_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.wr_byte_count_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.wr_trans_count_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.total_wr_latency_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.rd_byte_count_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.rd_trans_count_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.total_rd_latency_),
  0,
  1,
  2,
  3,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.ack_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.slotname_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.accel_),
  1,
  0,
  2,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _impl_.no_of_samples_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.ack_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.slotname_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.accel_),
  1,
  0,
  2,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.awlen_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.arlen_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.eventflags_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.timestamp_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.host_timestamp_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.rd_bytes_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.wr_bytes_),
  0,
  1,
  2,
  3,
  4,
  5,
  6,
  ~0u,  // no _has_bits_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response, _impl_.output_data_),
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.data_),
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.strb_data_),
  PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.fullsize_),
  0,
  2,
  1,
  3,
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _impl_.device_timestamp_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _impl_.slot_n_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_wr_cycles_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.idle_bus_cycles_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_rd_id_),
  PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_wr_id_),
  0,
  1,
  2,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_.ack_),
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_.force_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.display_msgs_),
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.log_msgs_),
  PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.stop_msgs_),
  0,
  1,
  2,
  PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.dst_filename_),
  PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.offset_),
  PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.size_),
  0,
  1,
  2,
  PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.src_handle_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.dst_filename_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.src_offset_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.dst_offset_),
  1,
  0,
  2,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.write_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.type_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.state_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.route_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.flow_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.qsize_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.desc_size_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.flags_),
  0,
  1,
  2,
  3,
  4,
  5,
  6,
  7,
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _impl_.q_handle_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.q_handle_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.src_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.req_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.nonblocking_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.eot_),
  1,
  0,
  2,
  3,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _impl_.written_size_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.q_handle_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.dest_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.req_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.nonblocking_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.eot_),
  1,
  0,
  2,
  3,
  4,
  5,
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_.dest_),
  PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_.size_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _impl_.req_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_.dest_),
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_.size_),
  0,
  1,
  ~0u,  // no _has_bits_
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response, _impl_.fullrequest_),
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _impl_.q_handle_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _impl_.success_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.arg_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.flow_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.tag_),
  1,
  2,
  0,
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_.route_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_.setup_),
  0,
  ~0u,
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _impl_.success_),
  0,
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _impl_.version_),
  0,
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _impl_.success_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _impl_.slotname_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_data_bytes_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_busy_cycles_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_stall_cycles_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_),
  0,
  1,
  2,
  3,
  4,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_.ack_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_.slotname_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.timestamp_),
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_),
  0,
  1,
  2,
  ~0u,  // no _has_bits_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response, _impl_.output_data_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_.size_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_.valid_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_.data_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.addr_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.data_),
  1,
  2,
  0,
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _impl_.interrupt_line_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _impl_.valid_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_.graphname_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_.iterations_),
  0,
  1,
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _impl_.gh_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _impl_.gh_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.portname_),
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.buffer_),
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.size_),
  3,
  0,
  1,
  2,
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _impl_.size_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.portname_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.buffer_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.size_),
  3,
  0,
  1,
  2,
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_.buffer_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.gmioname_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.dir_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.size_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.offset_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.boh_),
  0,
  4,
  1,
  2,
  3,
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _impl_.gmioname_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_.cycletimeout_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_.gh_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_.cycletimeout_),
  1,
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _impl_.gh_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _impl_.ack_),
  0,
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.xmlbuff_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.xmlbuffsize_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.sharedbin_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.sharedbinsize_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.emuldata_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.emuldatasize_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.keepdir_),
  0,
  3,
  1,
  4,
  2,
  5,
  6,
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _impl_._has_bits_),
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _internal_metadata_),
  ~0u,  // no _extensions_
  ~0u,  // no _oneof_case_
  ~0u,  // no _weak_field_map_
  ~0u,  // no _inlined_string_donated_
  PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _impl_.ack_),
  0,
};
static const ::_pbi::MigrationSchema schemas[] PROTOBUF_SECTION_VARIABLE(protodesc_cold) = {
  { 0, 8, -1, sizeof(::call_packet_info)},
  { 10, 18, -1, sizeof(::response_packet_info)},
  { 20, 28, -1, sizeof(::xclSetEnvironment_call_namevaluepair)},
  { 30, -1, -1, sizeof(::xclSetEnvironment_call)},
  { 37, 44, -1, sizeof(::xclSetEnvironment_response)},
  { 45, 52, -1, sizeof(::xclLoadBitstream_call_ddrbank)},
  { 53, 66, -1, sizeof(::xclLoadBitstream_call)},
  { 73, 80, -1, sizeof(::xclLoadBitstream_response)},
  { 81, 90, -1, sizeof(::xclAllocDeviceBuffer_call)},
  { 93, 101, -1, sizeof(::xclAllocDeviceBuffer_response)},
  { 103, 110, -1, sizeof(::xclFreeDeviceBuffer_call)},
  { 111, 118, -1, sizeof(::xclFreeDeviceBuffer_response)},
  { 119, 127, -1, sizeof(::xclClose_call)},
  { 129, 136, -1, sizeof(::xclClose_response)},
  { 137, 149, -1, sizeof(::xclCopyBufferHost2Device_call)},
  { 155, 162, -1, sizeof(::xclCopyBufferHost2Device_response)},
  { 163, 175, -1, sizeof(::xclCopyBufferDevice2Host_call)},
  { 181, 189, -1, sizeof(::xclCopyBufferDevice2Host_response)},
  { 191, 202, -1, sizeof(::xclWriteAddrSpaceDeviceRam_call)},
  { 207, 214, -1, sizeof(::xclWriteAddrSpaceDeviceRam_response)},
  { 215, 224, -1, sizeof(::xclWriteAddrKernelCtrl_call_kernelInfo)},
  { 227, 239, -1, sizeof(::xclWriteAddrKernelCtrl_call)},
  { 245, 252, -1, sizeof(::xclWriteAddrKernelCtrl_response)},
  { 253, 264, -1, sizeof(::xclRegWrite_call)},
  { 269, 276, -1, sizeof(::xclRegWrite_response)},
  { 277, 288, -1, sizeof(::xclReadAddrSpaceDeviceRam_call)},
  { 293, 301, -1, sizeof(::xclReadAddrSpaceDeviceRam_response)},
  { 303, 314, -1, sizeof(::xclReadAddrKernelCtrl_call)},
  { 319, 327, -1, sizeof(::xclReadAddrKernelCtrl_response)},
  { 329, 341, -1, sizeof(::xclRegRead_call)},
  { 347, 355, -1, sizeof(::xclRegRead_response)},
  { 357, 365, -1, sizeof(::xclPerfMonReadCounters_call)},
  { 367, 379, -1, sizeof(::xclPerfMonReadCounters_response)},
  { 385, 394, -1, sizeof(::xclPerfMonGetTraceCount_call)},
  { 397, 404, -1, sizeof(::xclPerfMonGetTraceCount_response)},
  { 405, 414, -1, sizeof(::xclPerfMonReadTrace_call)},
  { 417, 430, -1, sizeof(::xclPerfMonReadTrace_response_events)},
  { 437, -1, -1, sizeof(::xclPerfMonReadTrace_response)},
  { 444, 454, -1, sizeof(::ddr_mem_msg)},
  { 458, 465, -1, sizeof(::xclGetDeviceTimestamp_call)},
  { 466, 473, -1, sizeof(::xclGetDeviceTimestamp_response)},
  { 474, 481, -1, sizeof(::xclReadBusStatus_call)},
  { 482, 493, -1, sizeof(::xclReadBusStatus_response)},
  { 498, 506, -1, sizeof(::xclGetDebugMessages_call)},
  { 508, 517, -1, sizeof(::xclGetDebugMessages_response)},
  { 520, 529, -1, sizeof(::xclImportBO_call)},
  { 532, 539, -1, sizeof(::xclImportBO_response)},
  { 540, 551, -1, sizeof(::xclCopyBO_call)},
  { 556, 563, -1, sizeof(::xclCopyBO_response)},
  { 564, 578, -1, sizeof(::xclCreateQueue_call)},
  { 586, 593, -1, sizeof(::xclCreateQueue_response)},
  { 594, 606, -1, sizeof(::xclWriteQueue_call)},
  { 612, 619, -1, sizeof(::xclWriteQueue_response)},
  { 620, 632, -1, sizeof(::xclReadQueue_call)},
  { 638, 646, -1, sizeof(::xclReadQueue_response)},
  { 648, 655, -1, sizeof(::xclPollCompletion_call)},
  { 656, 664, -1, sizeof(::xclPollCompletion_response_request)},
  { 666, -1, -1, sizeof(::xclPollCompletion_response)},
  { 673, 680, -1, sizeof(::xclDestroyQueue_call)},
  { 681, 688, -1, sizeof(::xclDestroyQueue_response)},
  { 689, 698, -1, sizeof(::xclSetupInstance_call_argflowpair)},
  { 701, 709, -1, sizeof(::xclSetupInstance_call)},
  { 711, 718, -1, sizeof(::xclSetupInstance_response)},
  { 719, 726, -1, sizeof(::swemuDriverVersion_call)},
  { 727, 734, -1, sizeof(::swemuDriverVersion_response)},
  { 735, 742, -1, sizeof(::xclPerfMonReadCounters_Streaming_call)},
  { 743, 754, -1, sizeof(::xclPerfMonReadCounters_Streaming_response)},
  { 759, 767, -1, sizeof(::xclPerfMonReadTrace_Streaming_call)},
  { 769, 778, -1, sizeof(::xclPerfMonReadTrace_Streaming_response_events)},
  { 781, -1, -1, sizeof(::xclPerfMonReadTrace_Streaming_response)},
  { 788, 796, -1, sizeof(::xclSlaveReadReq_call)},
  { 798, 806, -1, sizeof(::xclSlaveReadReq_response)},
  { 808, 817, -1, sizeof(::xclSlaveWriteReq_call)},
  { 820, 827, -1, sizeof(::xclSlaveWriteReq_response)},
  { 828, 835, -1, sizeof(::xclInterruptOccured_call)},
  { 836, 843, -1, sizeof(::xclInterruptOccured_response)},
  { 844, 852, -1, sizeof(::xclGraphInit_call)},
  { 854, 861, -1, sizeof(::xclGraphInit_response)},
  { 862, 870, -1, sizeof(::xclGraphRun_call)},
  { 872, 879, -1, sizeof(::xclGraphRun_response)},
  { 880, 887, -1, sizeof(::xclGraphWait_call)},
  { 888, 895, -1, sizeof(::xclGraphWait_response)},
  { 896, 903, -1, sizeof(::xclGraphEnd_call)},
  { 904, 911, -1, sizeof(::xclGraphEnd_response)},
  { 912, 922, -1, sizeof(::xclGraphUpdateRTP_call)},
  { 926, 933, -1, sizeof(::xclGraphUpdateRTP_response)},
  { 934, 944, -1, sizeof(::xclGraphReadRTP_call)},
  { 948, 956, -1, sizeof(::xclGraphReadRTP_response)},
  { 958, 969, -1, sizeof(::xclSyncBOAIENB_call)},
  { 974, 981, -1, sizeof(::xclSyncBOAIENB_response)},
  { 982, 989, -1, sizeof(::xclGMIOWait_call)},
  { 990, 997, -1, sizeof(::xclGMIOWait_response)},
  { 998, 1006, -1, sizeof(::xclGraphTimedWait_call)},
  { 1008, 1015, -1, sizeof(::xclGraphTimedWait_response)},
  { 1016, 1024, -1, sizeof(::xclGraphTimedEnd_call)},
  { 1026, 1033, -1, sizeof(::xclGraphTimedEnd_response)},
  { 1034, 1041, -1, sizeof(::xclGraphResume_call)},
  { 1042, 1049, -1, sizeof(::xclGraphResume_response)},
  { 1050, 1063, -1, sizeof(::xclLoadXclbinContent_call)},
  { 1070, 1077, -1, sizeof(::xclLoadXclbinContent_response)},
};

static const ::_pb::Message* const file_default_instances[] = {
  &::_call_packet_info_default_instance_._instance,
  &::_response_packet_info_default_instance_._instance,
  &::_xclSetEnvironment_call_namevaluepair_default_instance_._instance,
  &::_xclSetEnvironment_call_default_instance_._instance,
  &::_xclSetEnvironment_response_default_instance_._instance,
  &::_xclLoadBitstream_call_ddrbank_default_instance_._instance,
  &::_xclLoadBitstream_call_default_instance_._instance,
  &::_xclLoadBitstream_response_default_instance_._instance,
  &::_xclAllocDeviceBuffer_call_default_instance_._instance,
  &::_xclAllocDeviceBuffer_response_default_instance_._instance,
  &::_xclFreeDeviceBuffer_call_default_instance_._instance,
  &::_xclFreeDeviceBuffer_response_default_instance_._instance,
  &::_xclClose_call_default_instance_._instance,
  &::_xclClose_response_default_instance_._instance,
  &::_xclCopyBufferHost2Device_call_default_instance_._instance,
  &::_xclCopyBufferHost2Device_response_default_instance_._instance,
  &::_xclCopyBufferDevice2Host_call_default_instance_._instance,
  &::_xclCopyBufferDevice2Host_response_default_instance_._instance,
  &::_xclWriteAddrSpaceDeviceRam_call_default_instance_._instance,
  &::_xclWriteAddrSpaceDeviceRam_response_default_instance_._instance,
  &::_xclWriteAddrKernelCtrl_call_kernelInfo_default_instance_._instance,
  &::_xclWriteAddrKernelCtrl_call_default_instance_._instance,
  &::_xclWriteAddrKernelCtrl_response_default_instance_._instance,
  &::_xclRegWrite_call_default_instance_._instance,
  &::_xclRegWrite_response_default_instance_._instance,
  &::_xclReadAddrSpaceDeviceRam_call_default_instance_._instance,
  &::_xclReadAddrSpaceDeviceRam_response_default_instance_._instance,
  &::_xclReadAddrKernelCtrl_call_default_instance_._instance,
  &::_xclReadAddrKernelCtrl_response_default_instance_._instance,
  &::_xclRegRead_call_default_instance_._instance,
  &::_xclRegRead_response_default_instance_._instance,
  &::_xclPerfMonReadCounters_call_default_instance_._instance,
  &::_xclPerfMonReadCounters_response_default_instance_._instance,
  &::_xclPerfMonGetTraceCount_call_default_instance_._instance,
  &::_xclPerfMonGetTraceCount_response_default_instance_._instance,
  &::_xclPerfMonReadTrace_call_default_instance_._instance,
  &::_xclPerfMonReadTrace_response_events_default_instance_._instance,
  &::_xclPerfMonReadTrace_response_default_instance_._instance,
  &::_ddr_mem_msg_default_instance_._instance,
  &::_xclGetDeviceTimestamp_call_default_instance_._instance,
  &::_xclGetDeviceTimestamp_response_default_instance_._instance,
  &::_xclReadBusStatus_call_default_instance_._instance,
  &::_xclReadBusStatus_response_default_instance_._instance,
  &::_xclGetDebugMessages_call_default_instance_._instance,
  &::_xclGetDebugMessages_response_default_instance_._instance,
  &::_xclImportBO_call_default_instance_._instance,
  &::_xclImportBO_response_default_instance_._instance,
  &::_xclCopyBO_call_default_instance_._instance,
  &::_xclCopyBO_response_default_instance_._instance,
  &::_xclCreateQueue_call_default_instance_._instance,
  &::_xclCreateQueue_response_default_instance_._instance,
  &::_xclWriteQueue_call_default_instance_._instance,
  &::_xclWriteQueue_response_default_instance_._instance,
  &::_xclReadQueue_call_default_instance_._instance,
  &::_xclReadQueue_response_default_instance_._instance,
  &::_xclPollCompletion_call_default_instance_._instance,
  &::_xclPollCompletion_response_request_default_instance_._instance,
  &::_xclPollCompletion_response_default_instance_._instance,
  &::_xclDestroyQueue_call_default_instance_._instance,
  &::_xclDestroyQueue_response_default_instance_._instance,
  &::_xclSetupInstance_call_argflowpair_default_instance_._instance,
  &::_xclSetupInstance_call_default_instance_._instance,
  &::_xclSetupInstance_response_default_instance_._instance,
  &::_swemuDriverVersion_call_default_instance_._instance,
  &::_swemuDriverVersion_response_default_instance_._instance,
  &::_xclPerfMonReadCounters_Streaming_call_default_instance_._instance,
  &::_xclPerfMonReadCounters_Streaming_response_default_instance_._instance,
  &::_xclPerfMonReadTrace_Streaming_call_default_instance_._instance,
  &::_xclPerfMonReadTrace_Streaming_response_events_default_instance_._instance,
  &::_xclPerfMonReadTrace_Streaming_response_default_instance_._instance,
  &::_xclSlaveReadReq_call_default_instance_._instance,
  &::_xclSlaveReadReq_response_default_instance_._instance,
  &::_xclSlaveWriteReq_call_default_instance_._instance,
  &::_xclSlaveWriteReq_response_default_instance_._instance,
  &::_xclInterruptOccured_call_default_instance_._instance,
  &::_xclInterruptOccured_response_default_instance_._instance,
  &::_xclGraphInit_call_default_instance_._instance,
  &::_xclGraphInit_response_default_instance_._instance,
  &::_xclGraphRun_call_default_instance_._instance,
  &::_xclGraphRun_response_default_instance_._instance,
  &::_xclGraphWait_call_default_instance_._instance,
  &::_xclGraphWait_response_default_instance_._instance,
  &::_xclGraphEnd_call_default_instance_._instance,
  &::_xclGraphEnd_response_default_instance_._instance,
  &::_xclGraphUpdateRTP_call_default_instance_._instance,
  &::_xclGraphUpdateRTP_response_default_instance_._instance,
  &::_xclGraphReadRTP_call_default_instance_._instance,
  &::_xclGraphReadRTP_response_default_instance_._instance,
  &::_xclSyncBOAIENB_call_default_instance_._instance,
  &::_xclSyncBOAIENB_response_default_instance_._instance,
  &::_xclGMIOWait_call_default_instance_._instance,
  &::_xclGMIOWait_response_default_instance_._instance,
  &::_xclGraphTimedWait_call_default_instance_._instance,
  &::_xclGraphTimedWait_response_default_instance_._instance,
  &::_xclGraphTimedEnd_call_default_instance_._instance,
  &::_xclGraphTimedEnd_response_default_instance_._instance,
  &::_xclGraphResume_call_default_instance_._instance,
  &::_xclGraphResume_response_default_instance_._instance,
  &::_xclLoadXclbinContent_call_default_instance_._instance,
  &::_xclLoadXclbinContent_response_default_instance_._instance,
};

const char descriptor_table_protodef_rpc_5fmessages_2eproto[] PROTOBUF_SECTION_VARIABLE(protodesc_cold) =
  "\n\022rpc_messages.proto\"1\n\020call_packet_info"
  "\022\014\n\004size\030\001 \002(\006\022\017\n\007xcl_api\030\002 \001(\007\"5\n\024respo"
  "nse_packet_info\022\014\n\004size\030\001 \002(\006\022\017\n\007xcl_api"
  "\030\002 \001(\007\"\202\001\n\026xclSetEnvironment_call\022:\n\013env"
  "ironment\030\003 \003(\0132%.xclSetEnvironment_call."
  "namevaluepair\032,\n\rnamevaluepair\022\014\n\004name\030\001"
  " \001(\t\022\r\n\005value\030\002 \001(\t\")\n\032xclSetEnvironment"
  "_response\022\013\n\003ack\030\001 \001(\010\"\342\001\n\025xclLoadBitstr"
  "eam_call\022\017\n\007xmlfile\030\001 \002(\014\022\026\n\016dlopenfilen"
  "ame\030\002 \002(\014\022\022\n\ndevicename\030\003 \002(\014\022\027\n\017deviced"
  "irectory\030\004 \002(\014\022\027\n\017binarydirectory\030\005 \002(\014\022"
  "\017\n\007verbose\030\006 \002(\010\0220\n\010ddrbanks\030\010 \003(\0132\036.xcl"
  "LoadBitstream_call.ddrbank\032\027\n\007ddrbank\022\014\n"
  "\004size\030\007 \001(\004\"(\n\031xclLoadBitstream_response"
  "\022\013\n\003ack\030\001 \002(\010\"Q\n\031xclAllocDeviceBuffer_ca"
  "ll\022\022\n\nddraddress\030\001 \002(\004\022\014\n\004size\030\002 \002(\004\022\022\n\n"
  "peertopeer\030\003 \001(\010\">\n\035xclAllocDeviceBuffer"
  "_response\022\013\n\003ack\030\001 \002(\010\022\020\n\010filename\030\002 \001(\t"
  "\".\n\030xclFreeDeviceBuffer_call\022\022\n\nddraddre"
  "ss\030\001 \002(\004\"+\n\034xclFreeDeviceBuffer_response"
  "\022\013\n\003ack\030\001 \002(\010\":\n\rxclClose_call\022\027\n\017xclDev"
  "iceHandle\030\002 \001(\014\022\020\n\010closeall\030\003 \001(\010\"\"\n\021xcl"
  "Close_response\022\r\n\005valid\030\001 \002(\010\"~\n\035xclCopy"
  "BufferHost2Device_call\022\027\n\017xclDeviceHandl"
  "e\030\002 \002(\014\022\014\n\004dest\030\003 \002(\004\022\013\n\003src\030\004 \002(\014\022\014\n\004si"
  "ze\030\005 \002(\004\022\014\n\004seek\030\006 \002(\004\022\r\n\005space\030\007 \001(\r\"1\n"
  "!xclCopyBufferHost2Device_response\022\014\n\004si"
  "ze\030\001 \002(\004\"~\n\035xclCopyBufferDevice2Host_cal"
  "l\022\027\n\017xclDeviceHandle\030\002 \002(\014\022\014\n\004dest\030\003 \002(\014"
  "\022\013\n\003src\030\004 \002(\004\022\014\n\004size\030\005 \002(\004\022\014\n\004skip\030\006 \002("
  "\004\022\r\n\005space\030\007 \001(\r\"\?\n!xclCopyBufferDevice2"
  "Host_response\022\014\n\004size\030\001 \002(\004\022\014\n\004dest\030\002 \002("
  "\014\"j\n\037xclWriteAddrSpaceDeviceRam_call\022\014\n\004"
  "addr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \002(\014\022\r"
  "\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006 \001(\r\"4\n#xclWrit"
  "eAddrSpaceDeviceRam_response\022\r\n\005valid\030\001 "
  "\002(\010\"\334\001\n\033xclWriteAddrKernelCtrl_call\022\014\n\004a"
  "ddr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \002(\014\022<\n"
  "\013kernel_info\030\005 \003(\0132\'.xclWriteAddrKernelC"
  "trl_call.kernelInfo\022\r\n\005PF_ID\030\006 \001(\r\022\016\n\006BA"
  "R_ID\030\007 \001(\r\0326\n\nkernelInfo\022\014\n\004addr\030\001 \001(\004\022\014"
  "\n\004size\030\002 \001(\r\022\014\n\004name\030\003 \001(\t\"0\n\037xclWriteAd"
  "drKernelCtrl_response\022\r\n\005valid\030\001 \002(\010\"d\n\020"
  "xclRegWrite_call\022\023\n\013baseaddress\030\001 \002(\r\022\016\n"
  "\006offset\030\002 \002(\r\022\014\n\004data\030\003 \002(\014\022\r\n\005PF_ID\030\004 \001"
  "(\r\022\016\n\006BAR_ID\030\005 \001(\r\"%\n\024xclRegWrite_respon"
  "se\022\r\n\005valid\030\001 \002(\010\"i\n\036xclReadAddrSpaceDev"
  "iceRam_call\022\014\n\004addr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022"
  "\014\n\004data\030\004 \001(\014\022\r\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006"
  " \001(\r\"A\n\"xclReadAddrSpaceDeviceRam_respon"
  "se\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002 \001(\014\"e\n\032xclRe"
  "adAddrKernelCtrl_call\022\014\n\004addr\030\002 \002(\004\022\014\n\004s"
  "ize\030\003 \002(\r\022\014\n\004data\030\004 \001(\014\022\r\n\005PF_ID\030\005 \001(\r\022\016"
  "\n\006BAR_ID\030\006 \001(\r\"=\n\036xclReadAddrKernelCtrl_"
  "response\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002 \001(\014\"q\n"
  "\017xclRegRead_call\022\023\n\013baseaddress\030\001 \002(\r\022\016\n"
  "\006offset\030\002 \002(\r\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \001("
  "\014\022\r\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006 \001(\r\"2\n\023xclR"
  "egRead_response\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002"
  " \001(\014\">\n\033xclPerfMonReadCounters_call\022\020\n\010s"
  "lotname\030\001 \002(\t\022\r\n\005accel\030\002 \001(\010\"\263\001\n\037xclPerf"
  "MonReadCounters_response\022\025\n\rwr_byte_coun"
  "t\030\001 \001(\r\022\026\n\016wr_trans_count\030\002 \001(\r\022\030\n\020total"
  "_wr_latency\030\003 \001(\r\022\025\n\rrd_byte_count\030\004 \001(\r"
  "\022\026\n\016rd_trans_count\030\005 \001(\r\022\030\n\020total_rd_lat"
  "ency\030\006 \001(\r\"L\n\034xclPerfMonGetTraceCount_ca"
  "ll\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname\030\002 \001(\t\022\r\n\005acc"
  "el\030\003 \001(\010\"9\n xclPerfMonGetTraceCount_resp"
  "onse\022\025\n\rno_of_samples\030\001 \001(\r\"H\n\030xclPerfMo"
  "nReadTrace_call\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname"
  "\030\002 \001(\t\022\r\n\005accel\030\003 \001(\010\"\345\001\n\034xclPerfMonRead"
  "Trace_response\0229\n\013output_data\030\010 \003(\0132$.xc"
  "lPerfMonReadTrace_response.events\032\211\001\n\006ev"
  "ents\022\r\n\005awlen\030\001 \001(\r\022\r\n\005arlen\030\002 \001(\r\022\022\n\nev"
  "entflags\030\003 \001(\r\022\021\n\ttimestamp\030\004 \001(\r\022\026\n\016hos"
  "t_timestamp\030\005 \001(\004\022\020\n\010rd_bytes\030\006 \001(\r\022\020\n\010w"
  "r_bytes\030\007 \001(\r\"N\n\013ddr_mem_msg\022\014\n\004data\030\001 \002"
  "(\014\022\014\n\004size\030\002 \001(\004\022\021\n\tstrb_data\030\003 \001(\014\022\020\n\010f"
  "ullsize\030\004 \001(\004\")\n\032xclGetDeviceTimestamp_c"
  "all\022\013\n\003ack\030\001 \001(\010\":\n\036xclGetDeviceTimestam"
  "p_response\022\030\n\020device_timestamp\030\001 \002(\004\"\'\n\025"
  "xclReadBusStatus_call\022\016\n\006slot_n\030\001 \001(\005\"\244\001"
  "\n\031xclReadBusStatus_response\022\034\n\024uncomplet"
  "e_rd_cycles\030\001 \001(\004\022\034\n\024uncomplete_wr_cycle"
  "s\030\002 \001(\004\022\027\n\017idle_bus_cycles\030\003 \001(\004\022\030\n\020unco"
  "mplete_rd_id\030\004 \001(\004\022\030\n\020uncomplete_wr_id\030\005"
  " \001(\004\"6\n\030xclGetDebugMessages_call\022\013\n\003ack\030"
  "\001 \001(\010\022\r\n\005force\030\002 \001(\010\"Y\n\034xclGetDebugMessa"
  "ges_response\022\024\n\014display_msgs\030\001 \001(\014\022\020\n\010lo"
  "g_msgs\030\002 \001(\014\022\021\n\tstop_msgs\030\003 \001(\014\"F\n\020xclIm"
  "portBO_call\022\024\n\014dst_filename\030\001 \001(\t\022\016\n\006off"
  "set\030\002 \001(\004\022\014\n\004size\030\003 \001(\004\"#\n\024xclImportBO_r"
  "esponse\022\013\n\003ack\030\001 \001(\010\"p\n\016xclCopyBO_call\022\022"
  "\n\nsrc_handle\030\001 \001(\004\022\024\n\014dst_filename\030\002 \001(\t"
  "\022\014\n\004size\030\003 \001(\004\022\022\n\nsrc_offset\030\004 \001(\004\022\022\n\nds"
  "t_offset\030\005 \001(\004\"!\n\022xclCopyBO_response\022\013\n\003"
  "ack\030\001 \001(\010\"\217\001\n\023xclCreateQueue_call\022\r\n\005wri"
  "te\030\001 \001(\010\022\014\n\004type\030\002 \001(\r\022\r\n\005state\030\003 \001(\r\022\r\n"
  "\005route\030\004 \001(\r\022\014\n\004flow\030\005 \001(\004\022\r\n\005qsize\030\006 \001("
  "\r\022\021\n\tdesc_size\030\007 \001(\r\022\r\n\005flags\030\010 \001(\004\"+\n\027x"
  "clCreateQueue_response\022\020\n\010q_handle\030\001 \001(\004"
  "\"p\n\022xclWriteQueue_call\022\020\n\010q_handle\030\001 \001(\004"
  "\022\013\n\003src\030\002 \001(\014\022\014\n\004size\030\003 \001(\004\022\013\n\003req\030\004 \001(\004"
  "\022\023\n\013nonblocking\030\005 \001(\010\022\013\n\003eot\030\006 \001(\010\".\n\026xc"
  "lWriteQueue_response\022\024\n\014written_size\030\001 \001"
  "(\004\"p\n\021xclReadQueue_call\022\020\n\010q_handle\030\001 \001("
  "\004\022\014\n\004dest\030\002 \001(\014\022\014\n\004size\030\003 \001(\004\022\013\n\003req\030\004 \001"
  "(\004\022\023\n\013nonblocking\030\005 \001(\010\022\013\n\003eot\030\006 \001(\010\"3\n\025"
  "xclReadQueue_response\022\014\n\004dest\030\001 \001(\014\022\014\n\004s"
  "ize\030\002 \001(\004\"%\n\026xclPollCompletion_call\022\013\n\003r"
  "eq\030\001 \001(\004\"}\n\032xclPollCompletion_response\0228"
  "\n\013fullrequest\030\003 \003(\0132#.xclPollCompletion_"
  "response.request\032%\n\007request\022\014\n\004dest\030\001 \001("
  "\014\022\014\n\004size\030\002 \001(\004\"(\n\024xclDestroyQueue_call\022"
  "\020\n\010q_handle\030\001 \001(\004\"+\n\030xclDestroyQueue_res"
  "ponse\022\017\n\007success\030\001 \001(\010\"\220\001\n\025xclSetupInsta"
  "nce_call\022\r\n\005route\030\001 \001(\004\0221\n\005setup\030\005 \003(\0132\""
  ".xclSetupInstance_call.argflowpair\0325\n\013ar"
  "gflowpair\022\013\n\003arg\030\002 \001(\004\022\014\n\004flow\030\003 \001(\004\022\013\n\003"
  "tag\030\004 \001(\t\",\n\031xclSetupInstance_response\022\017"
  "\n\007success\030\001 \001(\010\"*\n\027swemuDriverVersion_ca"
  "ll\022\017\n\007version\030\001 \002(\t\".\n\033swemuDriverVersio"
  "n_response\022\017\n\007success\030\001 \001(\010\"9\n%xclPerfMo"
  "nReadCounters_Streaming_call\022\020\n\010slotname"
  "\030\001 \002(\t\"\250\001\n)xclPerfMonReadCounters_Stream"
  "ing_response\022\025\n\rstr_num_tranx\030\007 \001(\004\022\026\n\016s"
  "tr_data_bytes\030\010 \001(\004\022\027\n\017str_busy_cycles\030\t"
  " \001(\004\022\030\n\020str_stall_cycles\030\n \001(\004\022\031\n\021str_st"
  "arve_cycles\030\013 \001(\004\"C\n\"xclPerfMonReadTrace"
  "_Streaming_call\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname"
  "\030\002 \001(\t\"\266\001\n&xclPerfMonReadTrace_Streaming"
  "_response\022C\n\013output_data\030\010 \003(\0132..xclPerf"
  "MonReadTrace_Streaming_response.events\032G"
  "\n\006events\022\022\n\neventflags\030\003 \001(\r\022\021\n\ttimestam"
  "p\030\004 \001(\r\022\026\n\016host_timestamp\030\005 \001(\004\"2\n\024xclSl"
  "aveReadReq_call\022\014\n\004addr\030\001 \002(\004\022\014\n\004size\030\002 "
  "\002(\r\"7\n\030xclSlaveReadReq_response\022\r\n\005valid"
  "\030\001 \002(\010\022\014\n\004data\030\002 \002(\014\"A\n\025xclSlaveWriteReq"
  "_call\022\014\n\004addr\030\001 \002(\004\022\014\n\004size\030\002 \002(\r\022\014\n\004dat"
  "a\030\003 \002(\014\"*\n\031xclSlaveWriteReq_response\022\r\n\005"
  "valid\030\001 \002(\010\"2\n\030xclInterruptOccured_call\022"
  "\026\n\016interrupt_line\030\001 \002(\r\"-\n\034xclInterruptO"
  "ccured_response\022\r\n\005valid\030\001 \002(\010\"2\n\021xclGra"
  "phInit_call\022\n\n\002gh\030\002 \002(\r\022\021\n\tgraphname\030\003 \002"
  "(\014\"$\n\025xclGraphInit_response\022\013\n\003ack\030\001 \002(\010"
  "\"2\n\020xclGraphRun_call\022\n\n\002gh\030\002 \002(\r\022\022\n\niter"
  "ations\030\003 \002(\r\"#\n\024xclGraphRun_response\022\013\n\003"
  "ack\030\001 \002(\010\"\037\n\021xclGraphWait_call\022\n\n\002gh\030\002 \002"
  "(\r\"$\n\025xclGraphWait_response\022\013\n\003ack\030\001 \002(\010"
  "\"\036\n\020xclGraphEnd_call\022\n\n\002gh\030\002 \002(\r\"#\n\024xclG"
  "raphEnd_response\022\013\n\003ack\030\001 \002(\r\"T\n\026xclGrap"
  "hUpdateRTP_call\022\n\n\002gh\030\002 \002(\r\022\020\n\010portname\030"
  "\003 \002(\014\022\016\n\006buffer\030\004 \002(\014\022\014\n\004size\030\005 \002(\004\"*\n\032x"
  "clGraphUpdateRTP_response\022\014\n\004size\030\001 \002(\004\""
  "R\n\024xclGraphReadRTP_call\022\n\n\002gh\030\002 \002(\r\022\020\n\010p"
  "ortname\030\003 \002(\014\022\016\n\006buffer\030\004 \002(\014\022\014\n\004size\030\005 "
  "\002(\004\"8\n\030xclGraphReadRTP_response\022\014\n\004size\030"
  "\001 \002(\004\022\016\n\006buffer\030\002 \002(\014\"_\n\023xclSyncBOAIENB_"
  "call\022\020\n\010gmioname\030\002 \002(\014\022\013\n\003dir\030\003 \002(\r\022\014\n\004s"
  "ize\030\004 \002(\004\022\016\n\006offset\030\005 \002(\004\022\013\n\003boh\030\006 \002(\004\"&"
  "\n\027xclSyncBOAIENB_response\022\013\n\003ack\030\001 \002(\010\"$"
  "\n\020xclGMIOWait_call\022\020\n\010gmioname\030\002 \002(\014\"#\n\024"
  "xclGMIOWait_response\022\013\n\003ack\030\001 \002(\010\":\n\026xcl"
  "GraphTimedWait_call\022\n\n\002gh\030\002 \002(\r\022\024\n\014cycle"
  "timeout\030\003 \002(\004\")\n\032xclGraphTimedWait_respo"
  "nse\022\013\n\003ack\030\001 \002(\010\"9\n\025xclGraphTimedEnd_cal"
  "l\022\n\n\002gh\030\002 \002(\r\022\024\n\014cycletimeout\030\003 \002(\004\"(\n\031x"
  "clGraphTimedEnd_response\022\013\n\003ack\030\001 \002(\010\"!\n"
  "\023xclGraphResume_call\022\n\n\002gh\030\002 \002(\r\"&\n\027xclG"
  "raphResume_response\022\013\n\003ack\030\001 \002(\010\"\244\001\n\031xcl"
  "LoadXclbinContent_call\022\017\n\007xmlbuff\030\002 \002(\014\022"
  "\023\n\013xmlbuffsize\030\003 \002(\004\022\021\n\tsharedbin\030\004 \002(\014\022"
  "\025\n\rsharedbinsize\030\005 \002(\004\022\020\n\010emuldata\030\006 \002(\014"
  "\022\024\n\014emuldatasize\030\007 \002(\004\022\017\n\007keepdir\030\010 \002(\010\""
  ",\n\035xclLoadXclbinContent_response\022\013\n\003ack\030"
  "\001 \002(\010"
  ;
static ::_pbi::once_flag descriptor_table_rpc_5fmessages_2eproto_once;
const ::_pbi::DescriptorTable descriptor_table_rpc_5fmessages_2eproto = {
    false, false, 7005, descriptor_table_protodef_rpc_5fmessages_2eproto,
    "rpc_messages.proto",
    &descriptor_table_rpc_5fmessages_2eproto_once, nullptr, 0, 100,
    schemas, file_default_instances, TableStruct_rpc_5fmessages_2eproto::offsets,
    file_level_metadata_rpc_5fmessages_2eproto, file_level_enum_descriptors_rpc_5fmessages_2eproto,
    file_level_service_descriptors_rpc_5fmessages_2eproto,
};
PROTOBUF_ATTRIBUTE_WEAK const ::_pbi::DescriptorTable* descriptor_table_rpc_5fmessages_2eproto_getter() {
  return &descriptor_table_rpc_5fmessages_2eproto;
}

// Force running AddDescriptors() at dynamic initialization time.
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::_pbi::AddDescriptorsRunner dynamic_init_dummy_rpc_5fmessages_2eproto(&descriptor_table_rpc_5fmessages_2eproto);

// ===================================================================

class call_packet_info::_Internal {
 public:
  using HasBits = decltype(std::declval<call_packet_info>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xcl_api(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

call_packet_info::call_packet_info(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:call_packet_info)
}
call_packet_info::call_packet_info(const call_packet_info& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  call_packet_info* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){}
    , decltype(_impl_.xcl_api_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.xcl_api_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  // @@protoc_insertion_point(copy_constructor:call_packet_info)
}

inline void call_packet_info::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.xcl_api_){0u}
  };
}

call_packet_info::~call_packet_info() {
  // @@protoc_insertion_point(destructor:call_packet_info)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void call_packet_info::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void call_packet_info::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void call_packet_info::Clear() {
// @@protoc_insertion_point(message_clear_start:call_packet_info)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.xcl_api_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* call_packet_info::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required fixed64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 9)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::UnalignedLoad<uint64_t>(ptr);
          ptr += sizeof(uint64_t);
        } else
          goto handle_unusual;
        continue;
      // optional fixed32 xcl_api = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 21)) {
          _Internal::set_has_xcl_api(&has_bits);
          _impl_.xcl_api_ = ::PROTOBUF_NAMESPACE_ID::internal::UnalignedLoad<uint32_t>(ptr);
          ptr += sizeof(uint32_t);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* call_packet_info::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:call_packet_info)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required fixed64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed64ToArray(1, this->_internal_size(), target);
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed32ToArray(2, this->_internal_xcl_api(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:call_packet_info)
  return target;
}

size_t call_packet_info::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:call_packet_info)
  size_t total_size = 0;

  // required fixed64 size = 1;
  if (_internal_has_size()) {
    total_size += 1 + 8;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional fixed32 xcl_api = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000002u) {
    total_size += 1 + 4;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData call_packet_info::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    call_packet_info::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*call_packet_info::GetClassData() const { return &_class_data_; }


void call_packet_info::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<call_packet_info*>(&to_msg);
  auto& from = static_cast<const call_packet_info&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:call_packet_info)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.xcl_api_ = from._impl_.xcl_api_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void call_packet_info::CopyFrom(const call_packet_info& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:call_packet_info)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool call_packet_info::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void call_packet_info::InternalSwap(call_packet_info* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.xcl_api_)
      + sizeof(call_packet_info::_impl_.xcl_api_)
      - PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata call_packet_info::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[0]);
}

// ===================================================================

class response_packet_info::_Internal {
 public:
  using HasBits = decltype(std::declval<response_packet_info>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xcl_api(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

response_packet_info::response_packet_info(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:response_packet_info)
}
response_packet_info::response_packet_info(const response_packet_info& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  response_packet_info* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){}
    , decltype(_impl_.xcl_api_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.xcl_api_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  // @@protoc_insertion_point(copy_constructor:response_packet_info)
}

inline void response_packet_info::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.xcl_api_){0u}
  };
}

response_packet_info::~response_packet_info() {
  // @@protoc_insertion_point(destructor:response_packet_info)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void response_packet_info::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void response_packet_info::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void response_packet_info::Clear() {
// @@protoc_insertion_point(message_clear_start:response_packet_info)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.xcl_api_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* response_packet_info::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required fixed64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 9)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::UnalignedLoad<uint64_t>(ptr);
          ptr += sizeof(uint64_t);
        } else
          goto handle_unusual;
        continue;
      // optional fixed32 xcl_api = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 21)) {
          _Internal::set_has_xcl_api(&has_bits);
          _impl_.xcl_api_ = ::PROTOBUF_NAMESPACE_ID::internal::UnalignedLoad<uint32_t>(ptr);
          ptr += sizeof(uint32_t);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* response_packet_info::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:response_packet_info)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required fixed64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed64ToArray(1, this->_internal_size(), target);
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed32ToArray(2, this->_internal_xcl_api(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:response_packet_info)
  return target;
}

size_t response_packet_info::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:response_packet_info)
  size_t total_size = 0;

  // required fixed64 size = 1;
  if (_internal_has_size()) {
    total_size += 1 + 8;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional fixed32 xcl_api = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000002u) {
    total_size += 1 + 4;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData response_packet_info::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    response_packet_info::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*response_packet_info::GetClassData() const { return &_class_data_; }


void response_packet_info::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<response_packet_info*>(&to_msg);
  auto& from = static_cast<const response_packet_info&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:response_packet_info)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.xcl_api_ = from._impl_.xcl_api_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void response_packet_info::CopyFrom(const response_packet_info& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:response_packet_info)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool response_packet_info::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void response_packet_info::InternalSwap(response_packet_info* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.xcl_api_)
      + sizeof(response_packet_info::_impl_.xcl_api_)
      - PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata response_packet_info::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[1]);
}

// ===================================================================

class xclSetEnvironment_call_namevaluepair::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetEnvironment_call_namevaluepair>()._impl_._has_bits_);
  static void set_has_name(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_value(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_call.namevaluepair)
}
xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(const xclSetEnvironment_call_namevaluepair& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetEnvironment_call_namevaluepair* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.name_){}
    , decltype(_impl_.value_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.name_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.name_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_name()) {
    _this->_impl_.name_.Set(from._internal_name(), 
      _this->GetArenaForAllocation());
  }
  _impl_.value_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.value_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_value()) {
    _this->_impl_.value_.Set(from._internal_value(), 
      _this->GetArenaForAllocation());
  }
  // @@protoc_insertion_point(copy_constructor:xclSetEnvironment_call.namevaluepair)
}

inline void xclSetEnvironment_call_namevaluepair::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.name_){}
    , decltype(_impl_.value_){}
  };
  _impl_.name_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.name_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.value_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.value_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclSetEnvironment_call_namevaluepair::~xclSetEnvironment_call_namevaluepair() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_call.namevaluepair)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetEnvironment_call_namevaluepair::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.name_.Destroy();
  _impl_.value_.Destroy();
}

void xclSetEnvironment_call_namevaluepair::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetEnvironment_call_namevaluepair::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_call.namevaluepair)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.value_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetEnvironment_call_namevaluepair::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional string name = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_name();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclSetEnvironment_call.namevaluepair.name");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional string value = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_value();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclSetEnvironment_call.namevaluepair.value");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetEnvironment_call_namevaluepair::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_call.namevaluepair)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional string name = 1;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_name().data(), static_cast<int>(this->_internal_name().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclSetEnvironment_call.namevaluepair.name");
    target = stream->WriteStringMaybeAliased(
        1, this->_internal_name(), target);
  }

  // optional string value = 2;
  if (cached_has_bits & 0x00000002u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_value().data(), static_cast<int>(this->_internal_value().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclSetEnvironment_call.namevaluepair.value");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_value(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_call.namevaluepair)
  return target;
}

size_t xclSetEnvironment_call_namevaluepair::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_call.namevaluepair)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional string name = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_name());
    }

    // optional string value = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_value());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetEnvironment_call_namevaluepair::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetEnvironment_call_namevaluepair::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetEnvironment_call_namevaluepair::GetClassData() const { return &_class_data_; }


void xclSetEnvironment_call_namevaluepair::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_call_namevaluepair*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_call_namevaluepair&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_call.namevaluepair)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_name(from._internal_name());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_value(from._internal_value());
    }
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_call_namevaluepair::CopyFrom(const xclSetEnvironment_call_namevaluepair& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_call.namevaluepair)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetEnvironment_call_namevaluepair::IsInitialized() const {
  return true;
}

void xclSetEnvironment_call_namevaluepair::InternalSwap(xclSetEnvironment_call_namevaluepair* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.name_, lhs_arena,
      &other->_impl_.name_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.value_, lhs_arena,
      &other->_impl_.value_, rhs_arena
  );
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetEnvironment_call_namevaluepair::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[2]);
}

// ===================================================================

class xclSetEnvironment_call::_Internal {
 public:
};

xclSetEnvironment_call::xclSetEnvironment_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_call)
}
xclSetEnvironment_call::xclSetEnvironment_call(const xclSetEnvironment_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetEnvironment_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_.environment_){from._impl_.environment_}
    , /*decltype(_impl_._cached_size_)*/{}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  // @@protoc_insertion_point(copy_constructor:xclSetEnvironment_call)
}

inline void xclSetEnvironment_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_.environment_){arena}
    , /*decltype(_impl_._cached_size_)*/{}
  };
}

xclSetEnvironment_call::~xclSetEnvironment_call() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetEnvironment_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.environment_.~RepeatedPtrField();
}

void xclSetEnvironment_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetEnvironment_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.environment_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetEnvironment_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_environment(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<26>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetEnvironment_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_environment_size()); i < n; i++) {
    const auto& repfield = this->_internal_environment(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(3, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_call)
  return target;
}

size_t xclSetEnvironment_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
  total_size += 1UL * this->_internal_environment_size();
  for (const auto& msg : this->_impl_.environment_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetEnvironment_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetEnvironment_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetEnvironment_call::GetClassData() const { return &_class_data_; }


void xclSetEnvironment_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_call*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.environment_.MergeFrom(from._impl_.environment_);
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_call::CopyFrom(const xclSetEnvironment_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetEnvironment_call::IsInitialized() const {
  return true;
}

void xclSetEnvironment_call::InternalSwap(xclSetEnvironment_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.environment_.InternalSwap(&other->_impl_.environment_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetEnvironment_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[3]);
}

// ===================================================================

class xclSetEnvironment_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetEnvironment_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetEnvironment_response::xclSetEnvironment_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_response)
}
xclSetEnvironment_response::xclSetEnvironment_response(const xclSetEnvironment_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetEnvironment_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclSetEnvironment_response)
}

inline void xclSetEnvironment_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclSetEnvironment_response::~xclSetEnvironment_response() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetEnvironment_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclSetEnvironment_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetEnvironment_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetEnvironment_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetEnvironment_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_response)
  return target;
}

size_t xclSetEnvironment_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetEnvironment_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetEnvironment_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetEnvironment_response::GetClassData() const { return &_class_data_; }


void xclSetEnvironment_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_response*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_response::CopyFrom(const xclSetEnvironment_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetEnvironment_response::IsInitialized() const {
  return true;
}

void xclSetEnvironment_response::InternalSwap(xclSetEnvironment_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetEnvironment_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[4]);
}

// ===================================================================

class xclLoadBitstream_call_ddrbank::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_call_ddrbank>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_call.ddrbank)
}
xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(const xclLoadBitstream_call_ddrbank& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclLoadBitstream_call_ddrbank* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclLoadBitstream_call.ddrbank)
}

inline void xclLoadBitstream_call_ddrbank::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
}

xclLoadBitstream_call_ddrbank::~xclLoadBitstream_call_ddrbank() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_call.ddrbank)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclLoadBitstream_call_ddrbank::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclLoadBitstream_call_ddrbank::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclLoadBitstream_call_ddrbank::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_call.ddrbank)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclLoadBitstream_call_ddrbank::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 size = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclLoadBitstream_call_ddrbank::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_call.ddrbank)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 size = 7;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(7, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_call.ddrbank)
  return target;
}

size_t xclLoadBitstream_call_ddrbank::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_call.ddrbank)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 size = 7;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclLoadBitstream_call_ddrbank::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclLoadBitstream_call_ddrbank::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclLoadBitstream_call_ddrbank::GetClassData() const { return &_class_data_; }


void xclLoadBitstream_call_ddrbank::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_call_ddrbank*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_call_ddrbank&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_call.ddrbank)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_size()) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_call_ddrbank::CopyFrom(const xclLoadBitstream_call_ddrbank& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_call.ddrbank)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclLoadBitstream_call_ddrbank::IsInitialized() const {
  return true;
}

void xclLoadBitstream_call_ddrbank::InternalSwap(xclLoadBitstream_call_ddrbank* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclLoadBitstream_call_ddrbank::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[5]);
}

// ===================================================================

class xclLoadBitstream_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_call>()._impl_._has_bits_);
  static void set_has_xmlfile(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dlopenfilename(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_devicename(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_devicedirectory(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_binarydirectory(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_verbose(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000003f) ^ 0x0000003f) != 0;
  }
};

xclLoadBitstream_call::xclLoadBitstream_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_call)
}
xclLoadBitstream_call::xclLoadBitstream_call(const xclLoadBitstream_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclLoadBitstream_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddrbanks_){from._impl_.ddrbanks_}
    , decltype(_impl_.xmlfile_){}
    , decltype(_impl_.dlopenfilename_){}
    , decltype(_impl_.devicename_){}
    , decltype(_impl_.devicedirectory_){}
    , decltype(_impl_.binarydirectory_){}
    , decltype(_impl_.verbose_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.xmlfile_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xmlfile_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_xmlfile()) {
    _this->_impl_.xmlfile_.Set(from._internal_xmlfile(), 
      _this->GetArenaForAllocation());
  }
  _impl_.dlopenfilename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dlopenfilename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dlopenfilename()) {
    _this->_impl_.dlopenfilename_.Set(from._internal_dlopenfilename(), 
      _this->GetArenaForAllocation());
  }
  _impl_.devicename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.devicename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_devicename()) {
    _this->_impl_.devicename_.Set(from._internal_devicename(), 
      _this->GetArenaForAllocation());
  }
  _impl_.devicedirectory_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.devicedirectory_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_devicedirectory()) {
    _this->_impl_.devicedirectory_.Set(from._internal_devicedirectory(), 
      _this->GetArenaForAllocation());
  }
  _impl_.binarydirectory_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.binarydirectory_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_binarydirectory()) {
    _this->_impl_.binarydirectory_.Set(from._internal_binarydirectory(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.verbose_ = from._impl_.verbose_;
  // @@protoc_insertion_point(copy_constructor:xclLoadBitstream_call)
}

inline void xclLoadBitstream_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddrbanks_){arena}
    , decltype(_impl_.xmlfile_){}
    , decltype(_impl_.dlopenfilename_){}
    , decltype(_impl_.devicename_){}
    , decltype(_impl_.devicedirectory_){}
    , decltype(_impl_.binarydirectory_){}
    , decltype(_impl_.verbose_){false}
  };
  _impl_.xmlfile_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xmlfile_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.dlopenfilename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dlopenfilename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.devicename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.devicename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.devicedirectory_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.devicedirectory_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.binarydirectory_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.binarydirectory_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclLoadBitstream_call::~xclLoadBitstream_call() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclLoadBitstream_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.ddrbanks_.~RepeatedPtrField();
  _impl_.xmlfile_.Destroy();
  _impl_.dlopenfilename_.Destroy();
  _impl_.devicename_.Destroy();
  _impl_.devicedirectory_.Destroy();
  _impl_.binarydirectory_.Destroy();
}

void xclLoadBitstream_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclLoadBitstream_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ddrbanks_.Clear();
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xmlfile_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.dlopenfilename_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.devicename_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000008u) {
      _impl_.devicedirectory_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000010u) {
      _impl_.binarydirectory_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.verbose_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclLoadBitstream_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes xmlfile = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_xmlfile();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes dlopenfilename = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_dlopenfilename();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes devicename = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_devicename();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes devicedirectory = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_devicedirectory();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes binarydirectory = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 42)) {
          auto str = _internal_mutable_binarydirectory();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bool verbose = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_verbose(&has_bits);
          _impl_.verbose_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 66)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_ddrbanks(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<66>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclLoadBitstream_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xmlfile = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        1, this->_internal_xmlfile(), target);
  }

  // required bytes dlopenfilename = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_dlopenfilename(), target);
  }

  // required bytes devicename = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_devicename(), target);
  }

  // required bytes devicedirectory = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_devicedirectory(), target);
  }

  // required bytes binarydirectory = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->WriteBytesMaybeAliased(
        5, this->_internal_binarydirectory(), target);
  }

  // required bool verbose = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(6, this->_internal_verbose(), target);
  }

  // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_ddrbanks_size()); i < n; i++) {
    const auto& repfield = this->_internal_ddrbanks(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_call)
  return target;
}

size_t xclLoadBitstream_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclLoadBitstream_call)
  size_t total_size = 0;

  if (_internal_has_xmlfile()) {
    // required bytes xmlfile = 1;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xmlfile());
  }

  if (_internal_has_dlopenfilename()) {
    // required bytes dlopenfilename = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dlopenfilename());
  }

  if (_internal_has_devicename()) {
    // required bytes devicename = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_devicename());
  }

  if (_internal_has_devicedirectory()) {
    // required bytes devicedirectory = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_devicedirectory());
  }

  if (_internal_has_binarydirectory()) {
    // required bytes binarydirectory = 5;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_binarydirectory());
  }

  if (_internal_has_verbose()) {
    // required bool verbose = 6;
    total_size += 1 + 1;
  }

  return total_size;
}
size_t xclLoadBitstream_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000003f) ^ 0x0000003f) == 0) {  // All required fields are present.
    // required bytes xmlfile = 1;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xmlfile());

    // required bytes dlopenfilename = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dlopenfilename());

    // required bytes devicename = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_devicename());

    // required bytes devicedirectory = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_devicedirectory());

    // required bytes binarydirectory = 5;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_binarydirectory());

    // required bool verbose = 6;
    total_size += 1 + 1;

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
  total_size += 1UL * this->_internal_ddrbanks_size();
  for (const auto& msg : this->_impl_.ddrbanks_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclLoadBitstream_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclLoadBitstream_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclLoadBitstream_call::GetClassData() const { return &_class_data_; }


void xclLoadBitstream_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_call*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.ddrbanks_.MergeFrom(from._impl_.ddrbanks_);
  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xmlfile(from._internal_xmlfile());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_dlopenfilename(from._internal_dlopenfilename());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_devicename(from._internal_devicename());
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_internal_set_devicedirectory(from._internal_devicedirectory());
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_internal_set_binarydirectory(from._internal_binarydirectory());
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.verbose_ = from._impl_.verbose_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_call::CopyFrom(const xclLoadBitstream_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclLoadBitstream_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclLoadBitstream_call::InternalSwap(xclLoadBitstream_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.ddrbanks_.InternalSwap(&other->_impl_.ddrbanks_);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.xmlfile_, lhs_arena,
      &other->_impl_.xmlfile_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dlopenfilename_, lhs_arena,
      &other->_impl_.dlopenfilename_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.devicename_, lhs_arena,
      &other->_impl_.devicename_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.devicedirectory_, lhs_arena,
      &other->_impl_.devicedirectory_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.binarydirectory_, lhs_arena,
      &other->_impl_.binarydirectory_, rhs_arena
  );
  swap(_impl_.verbose_, other->_impl_.verbose_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclLoadBitstream_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[6]);
}

// ===================================================================

class xclLoadBitstream_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclLoadBitstream_response::xclLoadBitstream_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_response)
}
xclLoadBitstream_response::xclLoadBitstream_response(const xclLoadBitstream_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclLoadBitstream_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclLoadBitstream_response)
}

inline void xclLoadBitstream_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclLoadBitstream_response::~xclLoadBitstream_response() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclLoadBitstream_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclLoadBitstream_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclLoadBitstream_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclLoadBitstream_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclLoadBitstream_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_response)
  return target;
}

size_t xclLoadBitstream_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclLoadBitstream_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclLoadBitstream_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclLoadBitstream_response::GetClassData() const { return &_class_data_; }


void xclLoadBitstream_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_response*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_response::CopyFrom(const xclLoadBitstream_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclLoadBitstream_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclLoadBitstream_response::InternalSwap(xclLoadBitstream_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclLoadBitstream_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[7]);
}

// ===================================================================

class xclAllocDeviceBuffer_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclAllocDeviceBuffer_call>()._impl_._has_bits_);
  static void set_has_ddraddress(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_peertopeer(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclAllocDeviceBuffer_call)
}
xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(const xclAllocDeviceBuffer_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclAllocDeviceBuffer_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddraddress_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.peertopeer_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.ddraddress_, &from._impl_.ddraddress_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.peertopeer_) -
    reinterpret_cast<char*>(&_impl_.ddraddress_)) + sizeof(_impl_.peertopeer_));
  // @@protoc_insertion_point(copy_constructor:xclAllocDeviceBuffer_call)
}

inline void xclAllocDeviceBuffer_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddraddress_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.peertopeer_){false}
  };
}

xclAllocDeviceBuffer_call::~xclAllocDeviceBuffer_call() {
  // @@protoc_insertion_point(destructor:xclAllocDeviceBuffer_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclAllocDeviceBuffer_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclAllocDeviceBuffer_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclAllocDeviceBuffer_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclAllocDeviceBuffer_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    ::memset(&_impl_.ddraddress_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.peertopeer_) -
        reinterpret_cast<char*>(&_impl_.ddraddress_)) + sizeof(_impl_.peertopeer_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclAllocDeviceBuffer_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 ddraddress = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ddraddress(&has_bits);
          _impl_.ddraddress_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool peertopeer = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_peertopeer(&has_bits);
          _impl_.peertopeer_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclAllocDeviceBuffer_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclAllocDeviceBuffer_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 ddraddress = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_ddraddress(), target);
  }

  // required uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_size(), target);
  }

  // optional bool peertopeer = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(3, this->_internal_peertopeer(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclAllocDeviceBuffer_call)
  return target;
}

size_t xclAllocDeviceBuffer_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclAllocDeviceBuffer_call)
  size_t total_size = 0;

  if (_internal_has_ddraddress()) {
    // required uint64 ddraddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_ddraddress());
  }

  if (_internal_has_size()) {
    // required uint64 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclAllocDeviceBuffer_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclAllocDeviceBuffer_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required uint64 ddraddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_ddraddress());

    // required uint64 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool peertopeer = 3;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000004u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclAllocDeviceBuffer_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclAllocDeviceBuffer_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclAllocDeviceBuffer_call::GetClassData() const { return &_class_data_; }


void xclAllocDeviceBuffer_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclAllocDeviceBuffer_call*>(&to_msg);
  auto& from = static_cast<const xclAllocDeviceBuffer_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclAllocDeviceBuffer_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.ddraddress_ = from._impl_.ddraddress_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.peertopeer_ = from._impl_.peertopeer_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclAllocDeviceBuffer_call::CopyFrom(const xclAllocDeviceBuffer_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclAllocDeviceBuffer_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclAllocDeviceBuffer_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclAllocDeviceBuffer_call::InternalSwap(xclAllocDeviceBuffer_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.peertopeer_)
      + sizeof(xclAllocDeviceBuffer_call::_impl_.peertopeer_)
      - PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.ddraddress_)>(
          reinterpret_cast<char*>(&_impl_.ddraddress_),
          reinterpret_cast<char*>(&other->_impl_.ddraddress_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclAllocDeviceBuffer_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[8]);
}

// ===================================================================

class xclAllocDeviceBuffer_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclAllocDeviceBuffer_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclAllocDeviceBuffer_response)
}
xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(const xclAllocDeviceBuffer_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclAllocDeviceBuffer_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.filename_){}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_filename()) {
    _this->_impl_.filename_.Set(from._internal_filename(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclAllocDeviceBuffer_response)
}

inline void xclAllocDeviceBuffer_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.filename_){}
    , decltype(_impl_.ack_){false}
  };
  _impl_.filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclAllocDeviceBuffer_response::~xclAllocDeviceBuffer_response() {
  // @@protoc_insertion_point(destructor:xclAllocDeviceBuffer_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclAllocDeviceBuffer_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.filename_.Destroy();
}

void xclAllocDeviceBuffer_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclAllocDeviceBuffer_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclAllocDeviceBuffer_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.filename_.ClearNonDefaultToEmpty();
  }
  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclAllocDeviceBuffer_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string filename = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_filename();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclAllocDeviceBuffer_response.filename");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclAllocDeviceBuffer_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclAllocDeviceBuffer_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  // optional string filename = 2;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_filename().data(), static_cast<int>(this->_internal_filename().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclAllocDeviceBuffer_response.filename");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_filename(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclAllocDeviceBuffer_response)
  return target;
}

size_t xclAllocDeviceBuffer_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclAllocDeviceBuffer_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional string filename = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
        this->_internal_filename());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclAllocDeviceBuffer_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclAllocDeviceBuffer_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclAllocDeviceBuffer_response::GetClassData() const { return &_class_data_; }


void xclAllocDeviceBuffer_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclAllocDeviceBuffer_response*>(&to_msg);
  auto& from = static_cast<const xclAllocDeviceBuffer_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclAllocDeviceBuffer_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_filename(from._internal_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclAllocDeviceBuffer_response::CopyFrom(const xclAllocDeviceBuffer_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclAllocDeviceBuffer_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclAllocDeviceBuffer_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclAllocDeviceBuffer_response::InternalSwap(xclAllocDeviceBuffer_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.filename_, lhs_arena,
      &other->_impl_.filename_, rhs_arena
  );
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclAllocDeviceBuffer_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[9]);
}

// ===================================================================

class xclFreeDeviceBuffer_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclFreeDeviceBuffer_call>()._impl_._has_bits_);
  static void set_has_ddraddress(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclFreeDeviceBuffer_call)
}
xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(const xclFreeDeviceBuffer_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclFreeDeviceBuffer_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddraddress_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ddraddress_ = from._impl_.ddraddress_;
  // @@protoc_insertion_point(copy_constructor:xclFreeDeviceBuffer_call)
}

inline void xclFreeDeviceBuffer_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ddraddress_){uint64_t{0u}}
  };
}

xclFreeDeviceBuffer_call::~xclFreeDeviceBuffer_call() {
  // @@protoc_insertion_point(destructor:xclFreeDeviceBuffer_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclFreeDeviceBuffer_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclFreeDeviceBuffer_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclFreeDeviceBuffer_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclFreeDeviceBuffer_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ddraddress_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclFreeDeviceBuffer_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 ddraddress = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ddraddress(&has_bits);
          _impl_.ddraddress_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclFreeDeviceBuffer_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclFreeDeviceBuffer_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 ddraddress = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_ddraddress(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclFreeDeviceBuffer_call)
  return target;
}

size_t xclFreeDeviceBuffer_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclFreeDeviceBuffer_call)
  size_t total_size = 0;

  // required uint64 ddraddress = 1;
  if (_internal_has_ddraddress()) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_ddraddress());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclFreeDeviceBuffer_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclFreeDeviceBuffer_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclFreeDeviceBuffer_call::GetClassData() const { return &_class_data_; }


void xclFreeDeviceBuffer_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclFreeDeviceBuffer_call*>(&to_msg);
  auto& from = static_cast<const xclFreeDeviceBuffer_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclFreeDeviceBuffer_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ddraddress()) {
    _this->_internal_set_ddraddress(from._internal_ddraddress());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclFreeDeviceBuffer_call::CopyFrom(const xclFreeDeviceBuffer_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclFreeDeviceBuffer_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclFreeDeviceBuffer_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclFreeDeviceBuffer_call::InternalSwap(xclFreeDeviceBuffer_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ddraddress_, other->_impl_.ddraddress_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclFreeDeviceBuffer_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[10]);
}

// ===================================================================

class xclFreeDeviceBuffer_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclFreeDeviceBuffer_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclFreeDeviceBuffer_response)
}
xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(const xclFreeDeviceBuffer_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclFreeDeviceBuffer_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclFreeDeviceBuffer_response)
}

inline void xclFreeDeviceBuffer_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclFreeDeviceBuffer_response::~xclFreeDeviceBuffer_response() {
  // @@protoc_insertion_point(destructor:xclFreeDeviceBuffer_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclFreeDeviceBuffer_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclFreeDeviceBuffer_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclFreeDeviceBuffer_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclFreeDeviceBuffer_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclFreeDeviceBuffer_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclFreeDeviceBuffer_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclFreeDeviceBuffer_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclFreeDeviceBuffer_response)
  return target;
}

size_t xclFreeDeviceBuffer_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclFreeDeviceBuffer_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclFreeDeviceBuffer_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclFreeDeviceBuffer_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclFreeDeviceBuffer_response::GetClassData() const { return &_class_data_; }


void xclFreeDeviceBuffer_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclFreeDeviceBuffer_response*>(&to_msg);
  auto& from = static_cast<const xclFreeDeviceBuffer_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclFreeDeviceBuffer_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclFreeDeviceBuffer_response::CopyFrom(const xclFreeDeviceBuffer_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclFreeDeviceBuffer_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclFreeDeviceBuffer_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclFreeDeviceBuffer_response::InternalSwap(xclFreeDeviceBuffer_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclFreeDeviceBuffer_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[11]);
}

// ===================================================================

class xclClose_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclClose_call>()._impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_closeall(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclClose_call::xclClose_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclClose_call)
}
xclClose_call::xclClose_call(const xclClose_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclClose_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.closeall_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_xcldevicehandle()) {
    _this->_impl_.xcldevicehandle_.Set(from._internal_xcldevicehandle(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.closeall_ = from._impl_.closeall_;
  // @@protoc_insertion_point(copy_constructor:xclClose_call)
}

inline void xclClose_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.closeall_){false}
  };
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclClose_call::~xclClose_call() {
  // @@protoc_insertion_point(destructor:xclClose_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclClose_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
}

void xclClose_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclClose_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclClose_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
  }
  _impl_.closeall_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclClose_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bytes xclDeviceHandle = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_xcldevicehandle();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool closeall = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_closeall(&has_bits);
          _impl_.closeall_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclClose_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclClose_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_xcldevicehandle(), target);
  }

  // optional bool closeall = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(3, this->_internal_closeall(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclClose_call)
  return target;
}

size_t xclClose_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclClose_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes xclDeviceHandle = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_xcldevicehandle());
    }

    // optional bool closeall = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclClose_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclClose_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclClose_call::GetClassData() const { return &_class_data_; }


void xclClose_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclClose_call*>(&to_msg);
  auto& from = static_cast<const xclClose_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclClose_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.closeall_ = from._impl_.closeall_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclClose_call::CopyFrom(const xclClose_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclClose_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclClose_call::IsInitialized() const {
  return true;
}

void xclClose_call::InternalSwap(xclClose_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.xcldevicehandle_, lhs_arena,
      &other->_impl_.xcldevicehandle_, rhs_arena
  );
  swap(_impl_.closeall_, other->_impl_.closeall_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclClose_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[12]);
}

// ===================================================================

class xclClose_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclClose_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclClose_response::xclClose_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclClose_response)
}
xclClose_response::xclClose_response(const xclClose_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclClose_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclClose_response)
}

inline void xclClose_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclClose_response::~xclClose_response() {
  // @@protoc_insertion_point(destructor:xclClose_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclClose_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclClose_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclClose_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclClose_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclClose_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclClose_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclClose_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclClose_response)
  return target;
}

size_t xclClose_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclClose_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclClose_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclClose_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclClose_response::GetClassData() const { return &_class_data_; }


void xclClose_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclClose_response*>(&to_msg);
  auto& from = static_cast<const xclClose_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclClose_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclClose_response::CopyFrom(const xclClose_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclClose_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclClose_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclClose_response::InternalSwap(xclClose_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclClose_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[13]);
}

// ===================================================================

class xclCopyBufferHost2Device_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferHost2Device_call>()._impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_seek(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_space(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferHost2Device_call)
}
xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(const xclCopyBufferHost2Device_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBufferHost2Device_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.src_){}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.seek_){}
    , decltype(_impl_.space_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_xcldevicehandle()) {
    _this->_impl_.xcldevicehandle_.Set(from._internal_xcldevicehandle(), 
      _this->GetArenaForAllocation());
  }
  _impl_.src_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.src_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_src()) {
    _this->_impl_.src_.Set(from._internal_src(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.dest_, &from._impl_.dest_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.space_) -
    reinterpret_cast<char*>(&_impl_.dest_)) + sizeof(_impl_.space_));
  // @@protoc_insertion_point(copy_constructor:xclCopyBufferHost2Device_call)
}

inline void xclCopyBufferHost2Device_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.src_){}
    , decltype(_impl_.dest_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.seek_){uint64_t{0u}}
    , decltype(_impl_.space_){0u}
  };
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.src_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.src_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclCopyBufferHost2Device_call::~xclCopyBufferHost2Device_call() {
  // @@protoc_insertion_point(destructor:xclCopyBufferHost2Device_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBufferHost2Device_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
  _impl_.src_.Destroy();
}

void xclCopyBufferHost2Device_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBufferHost2Device_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferHost2Device_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.src_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000003cu) {
    ::memset(&_impl_.dest_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.space_) -
        reinterpret_cast<char*>(&_impl_.dest_)) + sizeof(_impl_.space_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBufferHost2Device_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes xclDeviceHandle = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_xcldevicehandle();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 dest = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_dest(&has_bits);
          _impl_.dest_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes src = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_src();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 seek = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_seek(&has_bits);
          _impl_.seek_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 space = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_space(&has_bits);
          _impl_.space_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBufferHost2Device_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferHost2Device_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_xcldevicehandle(), target);
  }

  // required uint64 dest = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_dest(), target);
  }

  // required bytes src = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_src(), target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_size(), target);
  }

  // required uint64 seek = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(6, this->_internal_seek(), target);
  }

  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(7, this->_internal_space(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferHost2Device_call)
  return target;
}

size_t xclCopyBufferHost2Device_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclCopyBufferHost2Device_call)
  size_t total_size = 0;

  if (_internal_has_xcldevicehandle()) {
    // required bytes xclDeviceHandle = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xcldevicehandle());
  }

  if (_internal_has_src()) {
    // required bytes src = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_src());
  }

  if (_internal_has_dest()) {
    // required uint64 dest = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_dest());
  }

  if (_internal_has_size()) {
    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  if (_internal_has_seek()) {
    // required uint64 seek = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_seek());
  }

  return total_size;
}
size_t xclCopyBufferHost2Device_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferHost2Device_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000001f) ^ 0x0000001f) == 0) {  // All required fields are present.
    // required bytes xclDeviceHandle = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xcldevicehandle());

    // required bytes src = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_src());

    // required uint64 dest = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_dest());

    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

    // required uint64 seek = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_seek());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint32 space = 7;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000020u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_space());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBufferHost2Device_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBufferHost2Device_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBufferHost2Device_call::GetClassData() const { return &_class_data_; }


void xclCopyBufferHost2Device_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferHost2Device_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferHost2Device_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferHost2Device_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_src(from._internal_src());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.dest_ = from._impl_.dest_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.seek_ = from._impl_.seek_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.space_ = from._impl_.space_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferHost2Device_call::CopyFrom(const xclCopyBufferHost2Device_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferHost2Device_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBufferHost2Device_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclCopyBufferHost2Device_call::InternalSwap(xclCopyBufferHost2Device_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.xcldevicehandle_, lhs_arena,
      &other->_impl_.xcldevicehandle_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.src_, lhs_arena,
      &other->_impl_.src_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.space_)
      + sizeof(xclCopyBufferHost2Device_call::_impl_.space_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.dest_)>(
          reinterpret_cast<char*>(&_impl_.dest_),
          reinterpret_cast<char*>(&other->_impl_.dest_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBufferHost2Device_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[14]);
}

// ===================================================================

class xclCopyBufferHost2Device_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferHost2Device_response>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferHost2Device_response)
}
xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(const xclCopyBufferHost2Device_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBufferHost2Device_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclCopyBufferHost2Device_response)
}

inline void xclCopyBufferHost2Device_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
}

xclCopyBufferHost2Device_response::~xclCopyBufferHost2Device_response() {
  // @@protoc_insertion_point(destructor:xclCopyBufferHost2Device_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBufferHost2Device_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclCopyBufferHost2Device_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBufferHost2Device_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferHost2Device_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBufferHost2Device_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBufferHost2Device_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferHost2Device_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferHost2Device_response)
  return target;
}

size_t xclCopyBufferHost2Device_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferHost2Device_response)
  size_t total_size = 0;

  // required uint64 size = 1;
  if (_internal_has_size()) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBufferHost2Device_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBufferHost2Device_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBufferHost2Device_response::GetClassData() const { return &_class_data_; }


void xclCopyBufferHost2Device_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferHost2Device_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferHost2Device_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferHost2Device_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_size()) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferHost2Device_response::CopyFrom(const xclCopyBufferHost2Device_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferHost2Device_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBufferHost2Device_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclCopyBufferHost2Device_response::InternalSwap(xclCopyBufferHost2Device_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBufferHost2Device_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[15]);
}

// ===================================================================

class xclCopyBufferDevice2Host_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferDevice2Host_call>()._impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_skip(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_space(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferDevice2Host_call)
}
xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(const xclCopyBufferDevice2Host_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBufferDevice2Host_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.src_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.skip_){}
    , decltype(_impl_.space_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_xcldevicehandle()) {
    _this->_impl_.xcldevicehandle_.Set(from._internal_xcldevicehandle(), 
      _this->GetArenaForAllocation());
  }
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dest()) {
    _this->_impl_.dest_.Set(from._internal_dest(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.src_, &from._impl_.src_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.space_) -
    reinterpret_cast<char*>(&_impl_.src_)) + sizeof(_impl_.space_));
  // @@protoc_insertion_point(copy_constructor:xclCopyBufferDevice2Host_call)
}

inline void xclCopyBufferDevice2Host_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xcldevicehandle_){}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.src_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.skip_){uint64_t{0u}}
    , decltype(_impl_.space_){0u}
  };
  _impl_.xcldevicehandle_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xcldevicehandle_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclCopyBufferDevice2Host_call::~xclCopyBufferDevice2Host_call() {
  // @@protoc_insertion_point(destructor:xclCopyBufferDevice2Host_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBufferDevice2Host_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
  _impl_.dest_.Destroy();
}

void xclCopyBufferDevice2Host_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBufferDevice2Host_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferDevice2Host_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.dest_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000003cu) {
    ::memset(&_impl_.src_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.space_) -
        reinterpret_cast<char*>(&_impl_.src_)) + sizeof(_impl_.space_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBufferDevice2Host_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes xclDeviceHandle = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_xcldevicehandle();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes dest = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_dest();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 src = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_src(&has_bits);
          _impl_.src_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 skip = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_skip(&has_bits);
          _impl_.skip_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 space = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_space(&has_bits);
          _impl_.space_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBufferDevice2Host_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferDevice2Host_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_xcldevicehandle(), target);
  }

  // required bytes dest = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_dest(), target);
  }

  // required uint64 src = 4;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_src(), target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_size(), target);
  }

  // required uint64 skip = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(6, this->_internal_skip(), target);
  }

  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(7, this->_internal_space(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferDevice2Host_call)
  return target;
}

size_t xclCopyBufferDevice2Host_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclCopyBufferDevice2Host_call)
  size_t total_size = 0;

  if (_internal_has_xcldevicehandle()) {
    // required bytes xclDeviceHandle = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xcldevicehandle());
  }

  if (_internal_has_dest()) {
    // required bytes dest = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dest());
  }

  if (_internal_has_src()) {
    // required uint64 src = 4;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_src());
  }

  if (_internal_has_size()) {
    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  if (_internal_has_skip()) {
    // required uint64 skip = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_skip());
  }

  return total_size;
}
size_t xclCopyBufferDevice2Host_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferDevice2Host_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000001f) ^ 0x0000001f) == 0) {  // All required fields are present.
    // required bytes xclDeviceHandle = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xcldevicehandle());

    // required bytes dest = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dest());

    // required uint64 src = 4;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_src());

    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

    // required uint64 skip = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_skip());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint32 space = 7;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000020u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_space());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBufferDevice2Host_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBufferDevice2Host_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBufferDevice2Host_call::GetClassData() const { return &_class_data_; }


void xclCopyBufferDevice2Host_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferDevice2Host_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferDevice2Host_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferDevice2Host_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.src_ = from._impl_.src_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.skip_ = from._impl_.skip_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.space_ = from._impl_.space_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferDevice2Host_call::CopyFrom(const xclCopyBufferDevice2Host_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferDevice2Host_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBufferDevice2Host_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclCopyBufferDevice2Host_call::InternalSwap(xclCopyBufferDevice2Host_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.xcldevicehandle_, lhs_arena,
      &other->_impl_.xcldevicehandle_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dest_, lhs_arena,
      &other->_impl_.dest_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.space_)
      + sizeof(xclCopyBufferDevice2Host_call::_impl_.space_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.src_)>(
          reinterpret_cast<char*>(&_impl_.src_),
          reinterpret_cast<char*>(&other->_impl_.src_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBufferDevice2Host_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[16]);
}

// ===================================================================

class xclCopyBufferDevice2Host_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferDevice2Host_response>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferDevice2Host_response)
}
xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(const xclCopyBufferDevice2Host_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBufferDevice2Host_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dest()) {
    _this->_impl_.dest_.Set(from._internal_dest(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclCopyBufferDevice2Host_response)
}

inline void xclCopyBufferDevice2Host_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclCopyBufferDevice2Host_response::~xclCopyBufferDevice2Host_response() {
  // @@protoc_insertion_point(destructor:xclCopyBufferDevice2Host_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBufferDevice2Host_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dest_.Destroy();
}

void xclCopyBufferDevice2Host_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBufferDevice2Host_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferDevice2Host_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBufferDevice2Host_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes dest = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_dest();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBufferDevice2Host_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferDevice2Host_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_size(), target);
  }

  // required bytes dest = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_dest(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferDevice2Host_response)
  return target;
}

size_t xclCopyBufferDevice2Host_response::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclCopyBufferDevice2Host_response)
  size_t total_size = 0;

  if (_internal_has_dest()) {
    // required bytes dest = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dest());
  }

  if (_internal_has_size()) {
    // required uint64 size = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclCopyBufferDevice2Host_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferDevice2Host_response)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required bytes dest = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_dest());

    // required uint64 size = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBufferDevice2Host_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBufferDevice2Host_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBufferDevice2Host_response::GetClassData() const { return &_class_data_; }


void xclCopyBufferDevice2Host_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferDevice2Host_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferDevice2Host_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferDevice2Host_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferDevice2Host_response::CopyFrom(const xclCopyBufferDevice2Host_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferDevice2Host_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBufferDevice2Host_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclCopyBufferDevice2Host_response::InternalSwap(xclCopyBufferDevice2Host_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dest_, lhs_arena,
      &other->_impl_.dest_, rhs_arena
  );
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBufferDevice2Host_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[17]);
}

// ===================================================================

class xclWriteAddrSpaceDeviceRam_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrSpaceDeviceRam_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrSpaceDeviceRam_call)
}
xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(const xclWriteAddrSpaceDeviceRam_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteAddrSpaceDeviceRam_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclWriteAddrSpaceDeviceRam_call)
}

inline void xclWriteAddrSpaceDeviceRam_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclWriteAddrSpaceDeviceRam_call::~xclWriteAddrSpaceDeviceRam_call() {
  // @@protoc_insertion_point(destructor:xclWriteAddrSpaceDeviceRam_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteAddrSpaceDeviceRam_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclWriteAddrSpaceDeviceRam_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteAddrSpaceDeviceRam_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrSpaceDeviceRam_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteAddrSpaceDeviceRam_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes data = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteAddrSpaceDeviceRam_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrSpaceDeviceRam_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_size(), target);
  }

  // required bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_data(), target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrSpaceDeviceRam_call)
  return target;
}

size_t xclWriteAddrSpaceDeviceRam_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclWriteAddrSpaceDeviceRam_call)
  size_t total_size = 0;

  if (_internal_has_data()) {
    // required bytes data = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (_internal_has_addr()) {
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclWriteAddrSpaceDeviceRam_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrSpaceDeviceRam_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000007) ^ 0x00000007) == 0) {  // All required fields are present.
    // required bytes data = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());

    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteAddrSpaceDeviceRam_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteAddrSpaceDeviceRam_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteAddrSpaceDeviceRam_call::GetClassData() const { return &_class_data_; }


void xclWriteAddrSpaceDeviceRam_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrSpaceDeviceRam_call*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrSpaceDeviceRam_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrSpaceDeviceRam_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrSpaceDeviceRam_call::CopyFrom(const xclWriteAddrSpaceDeviceRam_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrSpaceDeviceRam_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteAddrSpaceDeviceRam_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclWriteAddrSpaceDeviceRam_call::InternalSwap(xclWriteAddrSpaceDeviceRam_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_)
      + sizeof(xclWriteAddrSpaceDeviceRam_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteAddrSpaceDeviceRam_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[18]);
}

// ===================================================================

class xclWriteAddrSpaceDeviceRam_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrSpaceDeviceRam_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrSpaceDeviceRam_response)
}
xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(const xclWriteAddrSpaceDeviceRam_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteAddrSpaceDeviceRam_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclWriteAddrSpaceDeviceRam_response)
}

inline void xclWriteAddrSpaceDeviceRam_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclWriteAddrSpaceDeviceRam_response::~xclWriteAddrSpaceDeviceRam_response() {
  // @@protoc_insertion_point(destructor:xclWriteAddrSpaceDeviceRam_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteAddrSpaceDeviceRam_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclWriteAddrSpaceDeviceRam_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteAddrSpaceDeviceRam_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrSpaceDeviceRam_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteAddrSpaceDeviceRam_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteAddrSpaceDeviceRam_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrSpaceDeviceRam_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrSpaceDeviceRam_response)
  return target;
}

size_t xclWriteAddrSpaceDeviceRam_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrSpaceDeviceRam_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteAddrSpaceDeviceRam_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteAddrSpaceDeviceRam_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteAddrSpaceDeviceRam_response::GetClassData() const { return &_class_data_; }


void xclWriteAddrSpaceDeviceRam_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrSpaceDeviceRam_response*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrSpaceDeviceRam_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrSpaceDeviceRam_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrSpaceDeviceRam_response::CopyFrom(const xclWriteAddrSpaceDeviceRam_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrSpaceDeviceRam_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteAddrSpaceDeviceRam_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclWriteAddrSpaceDeviceRam_response::InternalSwap(xclWriteAddrSpaceDeviceRam_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteAddrSpaceDeviceRam_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[19]);
}

// ===================================================================

class xclWriteAddrKernelCtrl_call_kernelInfo::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_call_kernelInfo>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_name(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_call.kernelInfo)
}
xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(const xclWriteAddrKernelCtrl_call_kernelInfo& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteAddrKernelCtrl_call_kernelInfo* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.name_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.name_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.name_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_name()) {
    _this->_impl_.name_.Set(from._internal_name(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.size_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  // @@protoc_insertion_point(copy_constructor:xclWriteAddrKernelCtrl_call.kernelInfo)
}

inline void xclWriteAddrKernelCtrl_call_kernelInfo::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.name_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
  };
  _impl_.name_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.name_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclWriteAddrKernelCtrl_call_kernelInfo::~xclWriteAddrKernelCtrl_call_kernelInfo() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_call.kernelInfo)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteAddrKernelCtrl_call_kernelInfo::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.name_.Destroy();
}

void xclWriteAddrKernelCtrl_call_kernelInfo::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteAddrKernelCtrl_call_kernelInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.name_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_call_kernelInfo::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 addr = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string name = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_name();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclWriteAddrKernelCtrl_call.kernelInfo.name");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteAddrKernelCtrl_call_kernelInfo::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 addr = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_addr(), target);
  }

  // optional uint32 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_size(), target);
  }

  // optional string name = 3;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_name().data(), static_cast<int>(this->_internal_name().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclWriteAddrKernelCtrl_call.kernelInfo.name");
    target = stream->WriteStringMaybeAliased(
        3, this->_internal_name(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_call.kernelInfo)
  return target;
}

size_t xclWriteAddrKernelCtrl_call_kernelInfo::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string name = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_name());
    }

    // optional uint64 addr = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
    }

    // optional uint32 size = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteAddrKernelCtrl_call_kernelInfo::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteAddrKernelCtrl_call_kernelInfo::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteAddrKernelCtrl_call_kernelInfo::GetClassData() const { return &_class_data_; }


void xclWriteAddrKernelCtrl_call_kernelInfo::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_call_kernelInfo*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_call_kernelInfo&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_name(from._internal_name());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_call_kernelInfo::CopyFrom(const xclWriteAddrKernelCtrl_call_kernelInfo& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteAddrKernelCtrl_call_kernelInfo::IsInitialized() const {
  return true;
}

void xclWriteAddrKernelCtrl_call_kernelInfo::InternalSwap(xclWriteAddrKernelCtrl_call_kernelInfo* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.name_, lhs_arena,
      &other->_impl_.name_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_)
      + sizeof(xclWriteAddrKernelCtrl_call_kernelInfo::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteAddrKernelCtrl_call_kernelInfo::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[20]);
}

// ===================================================================

class xclWriteAddrKernelCtrl_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_call)
}
xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(const xclWriteAddrKernelCtrl_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteAddrKernelCtrl_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.kernel_info_){from._impl_.kernel_info_}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclWriteAddrKernelCtrl_call)
}

inline void xclWriteAddrKernelCtrl_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.kernel_info_){arena}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclWriteAddrKernelCtrl_call::~xclWriteAddrKernelCtrl_call() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteAddrKernelCtrl_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.kernel_info_.~RepeatedPtrField();
  _impl_.data_.Destroy();
}

void xclWriteAddrKernelCtrl_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteAddrKernelCtrl_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.kernel_info_.Clear();
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes data = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 42)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_kernel_info(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<42>(ptr));
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteAddrKernelCtrl_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_size(), target);
  }

  // required bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_data(), target);
  }

  // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_kernel_info_size()); i < n; i++) {
    const auto& repfield = this->_internal_kernel_info(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(5, repfield, repfield.GetCachedSize(), target, stream);
  }

  // optional uint32 PF_ID = 6;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 7;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(7, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_call)
  return target;
}

size_t xclWriteAddrKernelCtrl_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclWriteAddrKernelCtrl_call)
  size_t total_size = 0;

  if (_internal_has_data()) {
    // required bytes data = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (_internal_has_addr()) {
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclWriteAddrKernelCtrl_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000007) ^ 0x00000007) == 0) {  // All required fields are present.
    // required bytes data = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());

    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
  total_size += 1UL * this->_internal_kernel_info_size();
  for (const auto& msg : this->_impl_.kernel_info_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 6;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 7;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteAddrKernelCtrl_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteAddrKernelCtrl_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteAddrKernelCtrl_call::GetClassData() const { return &_class_data_; }


void xclWriteAddrKernelCtrl_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_call*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.kernel_info_.MergeFrom(from._impl_.kernel_info_);
  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_call::CopyFrom(const xclWriteAddrKernelCtrl_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteAddrKernelCtrl_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclWriteAddrKernelCtrl_call::InternalSwap(xclWriteAddrKernelCtrl_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.kernel_info_.InternalSwap(&other->_impl_.kernel_info_);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.bar_id_)
      + sizeof(xclWriteAddrKernelCtrl_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteAddrKernelCtrl_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[21]);
}

// ===================================================================

class xclWriteAddrKernelCtrl_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_response)
}
xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(const xclWriteAddrKernelCtrl_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteAddrKernelCtrl_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclWriteAddrKernelCtrl_response)
}

inline void xclWriteAddrKernelCtrl_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclWriteAddrKernelCtrl_response::~xclWriteAddrKernelCtrl_response() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteAddrKernelCtrl_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclWriteAddrKernelCtrl_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteAddrKernelCtrl_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteAddrKernelCtrl_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_response)
  return target;
}

size_t xclWriteAddrKernelCtrl_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteAddrKernelCtrl_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteAddrKernelCtrl_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteAddrKernelCtrl_response::GetClassData() const { return &_class_data_; }


void xclWriteAddrKernelCtrl_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_response*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_response::CopyFrom(const xclWriteAddrKernelCtrl_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteAddrKernelCtrl_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclWriteAddrKernelCtrl_response::InternalSwap(xclWriteAddrKernelCtrl_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteAddrKernelCtrl_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[22]);
}

// ===================================================================

class xclRegWrite_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegWrite_call>()._impl_._has_bits_);
  static void set_has_baseaddress(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclRegWrite_call::xclRegWrite_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclRegWrite_call)
}
xclRegWrite_call::xclRegWrite_call(const xclRegWrite_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclRegWrite_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.baseaddress_){}
    , decltype(_impl_.offset_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.baseaddress_, &from._impl_.baseaddress_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclRegWrite_call)
}

inline void xclRegWrite_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.baseaddress_){0u}
    , decltype(_impl_.offset_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclRegWrite_call::~xclRegWrite_call() {
  // @@protoc_insertion_point(destructor:xclRegWrite_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclRegWrite_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclRegWrite_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclRegWrite_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegWrite_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.baseaddress_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclRegWrite_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 baseaddress = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_baseaddress(&has_bits);
          _impl_.baseaddress_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 offset = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_offset(&has_bits);
          _impl_.offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes data = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclRegWrite_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegWrite_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 baseaddress = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_baseaddress(), target);
  }

  // required uint32 offset = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_offset(), target);
  }

  // required bytes data = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_data(), target);
  }

  // optional uint32 PF_ID = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(4, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegWrite_call)
  return target;
}

size_t xclRegWrite_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclRegWrite_call)
  size_t total_size = 0;

  if (_internal_has_data()) {
    // required bytes data = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (_internal_has_baseaddress()) {
    // required uint32 baseaddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_baseaddress());
  }

  if (_internal_has_offset()) {
    // required uint32 offset = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_offset());
  }

  return total_size;
}
size_t xclRegWrite_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegWrite_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000007) ^ 0x00000007) == 0) {  // All required fields are present.
    // required bytes data = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());

    // required uint32 baseaddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_baseaddress());

    // required uint32 offset = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_offset());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclRegWrite_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclRegWrite_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclRegWrite_call::GetClassData() const { return &_class_data_; }


void xclRegWrite_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclRegWrite_call*>(&to_msg);
  auto& from = static_cast<const xclRegWrite_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegWrite_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.baseaddress_ = from._impl_.baseaddress_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegWrite_call::CopyFrom(const xclRegWrite_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegWrite_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclRegWrite_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclRegWrite_call::InternalSwap(xclRegWrite_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.bar_id_)
      + sizeof(xclRegWrite_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.baseaddress_)>(
          reinterpret_cast<char*>(&_impl_.baseaddress_),
          reinterpret_cast<char*>(&other->_impl_.baseaddress_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclRegWrite_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[23]);
}

// ===================================================================

class xclRegWrite_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegWrite_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclRegWrite_response::xclRegWrite_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclRegWrite_response)
}
xclRegWrite_response::xclRegWrite_response(const xclRegWrite_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclRegWrite_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclRegWrite_response)
}

inline void xclRegWrite_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclRegWrite_response::~xclRegWrite_response() {
  // @@protoc_insertion_point(destructor:xclRegWrite_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclRegWrite_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclRegWrite_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclRegWrite_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegWrite_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclRegWrite_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclRegWrite_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegWrite_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegWrite_response)
  return target;
}

size_t xclRegWrite_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegWrite_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclRegWrite_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclRegWrite_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclRegWrite_response::GetClassData() const { return &_class_data_; }


void xclRegWrite_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclRegWrite_response*>(&to_msg);
  auto& from = static_cast<const xclRegWrite_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegWrite_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegWrite_response::CopyFrom(const xclRegWrite_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegWrite_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclRegWrite_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclRegWrite_response::InternalSwap(xclRegWrite_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclRegWrite_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[24]);
}

// ===================================================================

class xclReadAddrSpaceDeviceRam_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrSpaceDeviceRam_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000006) ^ 0x00000006) != 0;
  }
};

xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrSpaceDeviceRam_call)
}
xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(const xclReadAddrSpaceDeviceRam_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadAddrSpaceDeviceRam_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclReadAddrSpaceDeviceRam_call)
}

inline void xclReadAddrSpaceDeviceRam_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadAddrSpaceDeviceRam_call::~xclReadAddrSpaceDeviceRam_call() {
  // @@protoc_insertion_point(destructor:xclReadAddrSpaceDeviceRam_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadAddrSpaceDeviceRam_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclReadAddrSpaceDeviceRam_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadAddrSpaceDeviceRam_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrSpaceDeviceRam_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadAddrSpaceDeviceRam_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadAddrSpaceDeviceRam_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrSpaceDeviceRam_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_data(), target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrSpaceDeviceRam_call)
  return target;
}

size_t xclReadAddrSpaceDeviceRam_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclReadAddrSpaceDeviceRam_call)
  size_t total_size = 0;

  if (_internal_has_addr()) {
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclReadAddrSpaceDeviceRam_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrSpaceDeviceRam_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000006) ^ 0x00000006) == 0) {  // All required fields are present.
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadAddrSpaceDeviceRam_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadAddrSpaceDeviceRam_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadAddrSpaceDeviceRam_call::GetClassData() const { return &_class_data_; }


void xclReadAddrSpaceDeviceRam_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrSpaceDeviceRam_call*>(&to_msg);
  auto& from = static_cast<const xclReadAddrSpaceDeviceRam_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrSpaceDeviceRam_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrSpaceDeviceRam_call::CopyFrom(const xclReadAddrSpaceDeviceRam_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrSpaceDeviceRam_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadAddrSpaceDeviceRam_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclReadAddrSpaceDeviceRam_call::InternalSwap(xclReadAddrSpaceDeviceRam_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_)
      + sizeof(xclReadAddrSpaceDeviceRam_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadAddrSpaceDeviceRam_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[25]);
}

// ===================================================================

class xclReadAddrSpaceDeviceRam_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrSpaceDeviceRam_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrSpaceDeviceRam_response)
}
xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(const xclReadAddrSpaceDeviceRam_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadAddrSpaceDeviceRam_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclReadAddrSpaceDeviceRam_response)
}

inline void xclReadAddrSpaceDeviceRam_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){false}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadAddrSpaceDeviceRam_response::~xclReadAddrSpaceDeviceRam_response() {
  // @@protoc_insertion_point(destructor:xclReadAddrSpaceDeviceRam_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadAddrSpaceDeviceRam_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclReadAddrSpaceDeviceRam_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadAddrSpaceDeviceRam_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrSpaceDeviceRam_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadAddrSpaceDeviceRam_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadAddrSpaceDeviceRam_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrSpaceDeviceRam_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_data(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrSpaceDeviceRam_response)
  return target;
}

size_t xclReadAddrSpaceDeviceRam_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrSpaceDeviceRam_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadAddrSpaceDeviceRam_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadAddrSpaceDeviceRam_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadAddrSpaceDeviceRam_response::GetClassData() const { return &_class_data_; }


void xclReadAddrSpaceDeviceRam_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrSpaceDeviceRam_response*>(&to_msg);
  auto& from = static_cast<const xclReadAddrSpaceDeviceRam_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrSpaceDeviceRam_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrSpaceDeviceRam_response::CopyFrom(const xclReadAddrSpaceDeviceRam_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrSpaceDeviceRam_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadAddrSpaceDeviceRam_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclReadAddrSpaceDeviceRam_response::InternalSwap(xclReadAddrSpaceDeviceRam_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadAddrSpaceDeviceRam_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[26]);
}

// ===================================================================

class xclReadAddrKernelCtrl_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrKernelCtrl_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000006) ^ 0x00000006) != 0;
  }
};

xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrKernelCtrl_call)
}
xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(const xclReadAddrKernelCtrl_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadAddrKernelCtrl_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclReadAddrKernelCtrl_call)
}

inline void xclReadAddrKernelCtrl_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadAddrKernelCtrl_call::~xclReadAddrKernelCtrl_call() {
  // @@protoc_insertion_point(destructor:xclReadAddrKernelCtrl_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadAddrKernelCtrl_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclReadAddrKernelCtrl_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadAddrKernelCtrl_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrKernelCtrl_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadAddrKernelCtrl_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadAddrKernelCtrl_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrKernelCtrl_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_data(), target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrKernelCtrl_call)
  return target;
}

size_t xclReadAddrKernelCtrl_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclReadAddrKernelCtrl_call)
  size_t total_size = 0;

  if (_internal_has_addr()) {
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclReadAddrKernelCtrl_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrKernelCtrl_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000006) ^ 0x00000006) == 0) {  // All required fields are present.
    // required uint64 addr = 2;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadAddrKernelCtrl_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadAddrKernelCtrl_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadAddrKernelCtrl_call::GetClassData() const { return &_class_data_; }


void xclReadAddrKernelCtrl_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrKernelCtrl_call*>(&to_msg);
  auto& from = static_cast<const xclReadAddrKernelCtrl_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrKernelCtrl_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrKernelCtrl_call::CopyFrom(const xclReadAddrKernelCtrl_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrKernelCtrl_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadAddrKernelCtrl_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclReadAddrKernelCtrl_call::InternalSwap(xclReadAddrKernelCtrl_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.bar_id_)
      + sizeof(xclReadAddrKernelCtrl_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadAddrKernelCtrl_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[27]);
}

// ===================================================================

class xclReadAddrKernelCtrl_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrKernelCtrl_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrKernelCtrl_response)
}
xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(const xclReadAddrKernelCtrl_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadAddrKernelCtrl_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclReadAddrKernelCtrl_response)
}

inline void xclReadAddrKernelCtrl_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){false}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadAddrKernelCtrl_response::~xclReadAddrKernelCtrl_response() {
  // @@protoc_insertion_point(destructor:xclReadAddrKernelCtrl_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadAddrKernelCtrl_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclReadAddrKernelCtrl_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadAddrKernelCtrl_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrKernelCtrl_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadAddrKernelCtrl_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadAddrKernelCtrl_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrKernelCtrl_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_data(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrKernelCtrl_response)
  return target;
}

size_t xclReadAddrKernelCtrl_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrKernelCtrl_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadAddrKernelCtrl_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadAddrKernelCtrl_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadAddrKernelCtrl_response::GetClassData() const { return &_class_data_; }


void xclReadAddrKernelCtrl_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrKernelCtrl_response*>(&to_msg);
  auto& from = static_cast<const xclReadAddrKernelCtrl_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrKernelCtrl_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrKernelCtrl_response::CopyFrom(const xclReadAddrKernelCtrl_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrKernelCtrl_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadAddrKernelCtrl_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclReadAddrKernelCtrl_response::InternalSwap(xclReadAddrKernelCtrl_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadAddrKernelCtrl_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[28]);
}

// ===================================================================

class xclRegRead_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegRead_call>()._impl_._has_bits_);
  static void set_has_baseaddress(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000e) ^ 0x0000000e) != 0;
  }
};

xclRegRead_call::xclRegRead_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclRegRead_call)
}
xclRegRead_call::xclRegRead_call(const xclRegRead_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclRegRead_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.baseaddress_){}
    , decltype(_impl_.offset_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.pf_id_){}
    , decltype(_impl_.bar_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.baseaddress_, &from._impl_.baseaddress_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.bar_id_) -
    reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  // @@protoc_insertion_point(copy_constructor:xclRegRead_call)
}

inline void xclRegRead_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.baseaddress_){0u}
    , decltype(_impl_.offset_){0u}
    , decltype(_impl_.size_){0u}
    , decltype(_impl_.pf_id_){0u}
    , decltype(_impl_.bar_id_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclRegRead_call::~xclRegRead_call() {
  // @@protoc_insertion_point(destructor:xclRegRead_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclRegRead_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclRegRead_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclRegRead_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegRead_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.baseaddress_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclRegRead_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 baseaddress = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_baseaddress(&has_bits);
          _impl_.baseaddress_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 offset = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_offset(&has_bits);
          _impl_.offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 PF_ID = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_pf_id(&has_bits);
          _impl_.pf_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 BAR_ID = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_bar_id(&has_bits);
          _impl_.bar_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclRegRead_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegRead_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 baseaddress = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_baseaddress(), target);
  }

  // required uint32 offset = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_offset(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_data(), target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegRead_call)
  return target;
}

size_t xclRegRead_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclRegRead_call)
  size_t total_size = 0;

  if (_internal_has_baseaddress()) {
    // required uint32 baseaddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_baseaddress());
  }

  if (_internal_has_offset()) {
    // required uint32 offset = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_offset());
  }

  if (_internal_has_size()) {
    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclRegRead_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegRead_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000000e) ^ 0x0000000e) == 0) {  // All required fields are present.
    // required uint32 baseaddress = 1;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_baseaddress());

    // required uint32 offset = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_offset());

    // required uint32 size = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (cached_has_bits & 0x00000030u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclRegRead_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclRegRead_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclRegRead_call::GetClassData() const { return &_class_data_; }


void xclRegRead_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclRegRead_call*>(&to_msg);
  auto& from = static_cast<const xclRegRead_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegRead_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.baseaddress_ = from._impl_.baseaddress_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegRead_call::CopyFrom(const xclRegRead_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegRead_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclRegRead_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclRegRead_call::InternalSwap(xclRegRead_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.bar_id_)
      + sizeof(xclRegRead_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.baseaddress_)>(
          reinterpret_cast<char*>(&_impl_.baseaddress_),
          reinterpret_cast<char*>(&other->_impl_.baseaddress_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclRegRead_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[29]);
}

// ===================================================================

class xclRegRead_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegRead_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclRegRead_response::xclRegRead_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclRegRead_response)
}
xclRegRead_response::xclRegRead_response(const xclRegRead_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclRegRead_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclRegRead_response)
}

inline void xclRegRead_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){false}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclRegRead_response::~xclRegRead_response() {
  // @@protoc_insertion_point(destructor:xclRegRead_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclRegRead_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclRegRead_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclRegRead_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegRead_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclRegRead_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes data = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclRegRead_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegRead_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_data(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegRead_response)
  return target;
}

size_t xclRegRead_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegRead_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclRegRead_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclRegRead_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclRegRead_response::GetClassData() const { return &_class_data_; }


void xclRegRead_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclRegRead_response*>(&to_msg);
  auto& from = static_cast<const xclRegRead_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegRead_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegRead_response::CopyFrom(const xclRegRead_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegRead_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclRegRead_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclRegRead_response::InternalSwap(xclRegRead_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclRegRead_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[30]);
}

// ===================================================================

class xclPerfMonReadCounters_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_call>()._impl_._has_bits_);
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_call)
}
xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(const xclPerfMonReadCounters_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadCounters_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.accel_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_slotname()) {
    _this->_impl_.slotname_.Set(from._internal_slotname(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.accel_ = from._impl_.accel_;
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_call)
}

inline void xclPerfMonReadCounters_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.accel_){false}
  };
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPerfMonReadCounters_call::~xclPerfMonReadCounters_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadCounters_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.slotname_.Destroy();
}

void xclPerfMonReadCounters_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadCounters_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_.accel_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required string slotname = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_slotname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclPerfMonReadCounters_call.slotname");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional bool accel = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_accel(&has_bits);
          _impl_.accel_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadCounters_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string slotname = 1;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_slotname().data(), static_cast<int>(this->_internal_slotname().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclPerfMonReadCounters_call.slotname");
    target = stream->WriteStringMaybeAliased(
        1, this->_internal_slotname(), target);
  }

  // optional bool accel = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(2, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_call)
  return target;
}

size_t xclPerfMonReadCounters_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_call)
  size_t total_size = 0;

  // required string slotname = 1;
  if (_internal_has_slotname()) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
        this->_internal_slotname());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool accel = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000002u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadCounters_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadCounters_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadCounters_call::GetClassData() const { return &_class_data_; }


void xclPerfMonReadCounters_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_call::CopyFrom(const xclPerfMonReadCounters_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadCounters_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclPerfMonReadCounters_call::InternalSwap(xclPerfMonReadCounters_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.slotname_, lhs_arena,
      &other->_impl_.slotname_, rhs_arena
  );
  swap(_impl_.accel_, other->_impl_.accel_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadCounters_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[31]);
}

// ===================================================================

class xclPerfMonReadCounters_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_response>()._impl_._has_bits_);
  static void set_has_wr_byte_count(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_wr_trans_count(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_total_wr_latency(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_rd_byte_count(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_rd_trans_count(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_total_rd_latency(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_response)
}
xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(const xclPerfMonReadCounters_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadCounters_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.wr_byte_count_){}
    , decltype(_impl_.wr_trans_count_){}
    , decltype(_impl_.total_wr_latency_){}
    , decltype(_impl_.rd_byte_count_){}
    , decltype(_impl_.rd_trans_count_){}
    , decltype(_impl_.total_rd_latency_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.wr_byte_count_, &from._impl_.wr_byte_count_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.total_rd_latency_) -
    reinterpret_cast<char*>(&_impl_.wr_byte_count_)) + sizeof(_impl_.total_rd_latency_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_response)
}

inline void xclPerfMonReadCounters_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.wr_byte_count_){0u}
    , decltype(_impl_.wr_trans_count_){0u}
    , decltype(_impl_.total_wr_latency_){0u}
    , decltype(_impl_.rd_byte_count_){0u}
    , decltype(_impl_.rd_trans_count_){0u}
    , decltype(_impl_.total_rd_latency_){0u}
  };
}

xclPerfMonReadCounters_response::~xclPerfMonReadCounters_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadCounters_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPerfMonReadCounters_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadCounters_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    ::memset(&_impl_.wr_byte_count_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.total_rd_latency_) -
        reinterpret_cast<char*>(&_impl_.wr_byte_count_)) + sizeof(_impl_.total_rd_latency_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint32 wr_byte_count = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_wr_byte_count(&has_bits);
          _impl_.wr_byte_count_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 wr_trans_count = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_wr_trans_count(&has_bits);
          _impl_.wr_trans_count_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 total_wr_latency = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_total_wr_latency(&has_bits);
          _impl_.total_wr_latency_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 rd_byte_count = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_rd_byte_count(&has_bits);
          _impl_.rd_byte_count_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 rd_trans_count = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_rd_trans_count(&has_bits);
          _impl_.rd_trans_count_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 total_rd_latency = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_total_rd_latency(&has_bits);
          _impl_.total_rd_latency_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadCounters_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 wr_byte_count = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_wr_byte_count(), target);
  }

  // optional uint32 wr_trans_count = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_wr_trans_count(), target);
  }

  // optional uint32 total_wr_latency = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_total_wr_latency(), target);
  }

  // optional uint32 rd_byte_count = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(4, this->_internal_rd_byte_count(), target);
  }

  // optional uint32 rd_trans_count = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(5, this->_internal_rd_trans_count(), target);
  }

  // optional uint32 total_rd_latency = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_total_rd_latency(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_response)
  return target;
}

size_t xclPerfMonReadCounters_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional uint32 wr_byte_count = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_wr_byte_count());
    }

    // optional uint32 wr_trans_count = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_wr_trans_count());
    }

    // optional uint32 total_wr_latency = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_total_wr_latency());
    }

    // optional uint32 rd_byte_count = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_rd_byte_count());
    }

    // optional uint32 rd_trans_count = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_rd_trans_count());
    }

    // optional uint32 total_rd_latency = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_total_rd_latency());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadCounters_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadCounters_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadCounters_response::GetClassData() const { return &_class_data_; }


void xclPerfMonReadCounters_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.wr_byte_count_ = from._impl_.wr_byte_count_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.wr_trans_count_ = from._impl_.wr_trans_count_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.total_wr_latency_ = from._impl_.total_wr_latency_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.rd_byte_count_ = from._impl_.rd_byte_count_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.rd_trans_count_ = from._impl_.rd_trans_count_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.total_rd_latency_ = from._impl_.total_rd_latency_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_response::CopyFrom(const xclPerfMonReadCounters_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadCounters_response::IsInitialized() const {
  return true;
}

void xclPerfMonReadCounters_response::InternalSwap(xclPerfMonReadCounters_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_rd_latency_)
      + sizeof(xclPerfMonReadCounters_response::_impl_.total_rd_latency_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_byte_count_)>(
          reinterpret_cast<char*>(&_impl_.wr_byte_count_),
          reinterpret_cast<char*>(&other->_impl_.wr_byte_count_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadCounters_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[32]);
}

// ===================================================================

class xclPerfMonGetTraceCount_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonGetTraceCount_call>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonGetTraceCount_call)
}
xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(const xclPerfMonGetTraceCount_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonGetTraceCount_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){}
    , decltype(_impl_.accel_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_slotname()) {
    _this->_impl_.slotname_.Set(from._internal_slotname(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.ack_, &from._impl_.ack_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.accel_) -
    reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonGetTraceCount_call)
}

inline void xclPerfMonGetTraceCount_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){false}
    , decltype(_impl_.accel_){false}
  };
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPerfMonGetTraceCount_call::~xclPerfMonGetTraceCount_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonGetTraceCount_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonGetTraceCount_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.slotname_.Destroy();
}

void xclPerfMonGetTraceCount_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonGetTraceCount_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonGetTraceCount_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  ::memset(&_impl_.ack_, 0, static_cast<size_t>(
      reinterpret_cast<char*>(&_impl_.accel_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonGetTraceCount_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string slotname = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_slotname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclPerfMonGetTraceCount_call.slotname");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional bool accel = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_accel(&has_bits);
          _impl_.accel_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonGetTraceCount_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonGetTraceCount_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_slotname().data(), static_cast<int>(this->_internal_slotname().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclPerfMonGetTraceCount_call.slotname");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_slotname(), target);
  }

  // optional bool accel = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(3, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonGetTraceCount_call)
  return target;
}

size_t xclPerfMonGetTraceCount_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonGetTraceCount_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + 1;
    }

    // optional bool accel = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonGetTraceCount_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonGetTraceCount_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonGetTraceCount_call::GetClassData() const { return &_class_data_; }


void xclPerfMonGetTraceCount_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonGetTraceCount_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonGetTraceCount_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonGetTraceCount_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonGetTraceCount_call::CopyFrom(const xclPerfMonGetTraceCount_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonGetTraceCount_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonGetTraceCount_call::IsInitialized() const {
  return true;
}

void xclPerfMonGetTraceCount_call::InternalSwap(xclPerfMonGetTraceCount_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.slotname_, lhs_arena,
      &other->_impl_.slotname_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.accel_)
      + sizeof(xclPerfMonGetTraceCount_call::_impl_.accel_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonGetTraceCount_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[33]);
}

// ===================================================================

class xclPerfMonGetTraceCount_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonGetTraceCount_response>()._impl_._has_bits_);
  static void set_has_no_of_samples(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonGetTraceCount_response)
}
xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(const xclPerfMonGetTraceCount_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonGetTraceCount_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.no_of_samples_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.no_of_samples_ = from._impl_.no_of_samples_;
  // @@protoc_insertion_point(copy_constructor:xclPerfMonGetTraceCount_response)
}

inline void xclPerfMonGetTraceCount_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.no_of_samples_){0u}
  };
}

xclPerfMonGetTraceCount_response::~xclPerfMonGetTraceCount_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonGetTraceCount_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonGetTraceCount_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPerfMonGetTraceCount_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonGetTraceCount_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonGetTraceCount_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.no_of_samples_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonGetTraceCount_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint32 no_of_samples = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_no_of_samples(&has_bits);
          _impl_.no_of_samples_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonGetTraceCount_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonGetTraceCount_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 no_of_samples = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_no_of_samples(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonGetTraceCount_response)
  return target;
}

size_t xclPerfMonGetTraceCount_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonGetTraceCount_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint32 no_of_samples = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_no_of_samples());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonGetTraceCount_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonGetTraceCount_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonGetTraceCount_response::GetClassData() const { return &_class_data_; }


void xclPerfMonGetTraceCount_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonGetTraceCount_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonGetTraceCount_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonGetTraceCount_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_no_of_samples()) {
    _this->_internal_set_no_of_samples(from._internal_no_of_samples());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonGetTraceCount_response::CopyFrom(const xclPerfMonGetTraceCount_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonGetTraceCount_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonGetTraceCount_response::IsInitialized() const {
  return true;
}

void xclPerfMonGetTraceCount_response::InternalSwap(xclPerfMonGetTraceCount_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.no_of_samples_, other->_impl_.no_of_samples_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonGetTraceCount_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[34]);
}

// ===================================================================

class xclPerfMonReadTrace_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_call>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_call)
}
xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(const xclPerfMonReadTrace_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){}
    , decltype(_impl_.accel_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_slotname()) {
    _this->_impl_.slotname_.Set(from._internal_slotname(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.ack_, &from._impl_.ack_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.accel_) -
    reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_call)
}

inline void xclPerfMonReadTrace_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){false}
    , decltype(_impl_.accel_){false}
  };
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPerfMonReadTrace_call::~xclPerfMonReadTrace_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.slotname_.Destroy();
}

void xclPerfMonReadTrace_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  ::memset(&_impl_.ack_, 0, static_cast<size_t>(
      reinterpret_cast<char*>(&_impl_.accel_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string slotname = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_slotname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclPerfMonReadTrace_call.slotname");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional bool accel = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_accel(&has_bits);
          _impl_.accel_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_slotname().data(), static_cast<int>(this->_internal_slotname().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclPerfMonReadTrace_call.slotname");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_slotname(), target);
  }

  // optional bool accel = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(3, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_call)
  return target;
}

size_t xclPerfMonReadTrace_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + 1;
    }

    // optional bool accel = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_call::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_call::CopyFrom(const xclPerfMonReadTrace_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_call::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_call::InternalSwap(xclPerfMonReadTrace_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.slotname_, lhs_arena,
      &other->_impl_.slotname_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.accel_)
      + sizeof(xclPerfMonReadTrace_call::_impl_.accel_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[35]);
}

// ===================================================================

class xclPerfMonReadTrace_response_events::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_response_events>()._impl_._has_bits_);
  static void set_has_awlen(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_arlen(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_eventflags(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_host_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_rd_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_wr_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
};

xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_response.events)
}
xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(const xclPerfMonReadTrace_response_events& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_response_events* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.awlen_){}
    , decltype(_impl_.arlen_){}
    , decltype(_impl_.eventflags_){}
    , decltype(_impl_.timestamp_){}
    , decltype(_impl_.host_timestamp_){}
    , decltype(_impl_.rd_bytes_){}
    , decltype(_impl_.wr_bytes_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.awlen_, &from._impl_.awlen_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.wr_bytes_) -
    reinterpret_cast<char*>(&_impl_.awlen_)) + sizeof(_impl_.wr_bytes_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_response.events)
}

inline void xclPerfMonReadTrace_response_events::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.awlen_){0u}
    , decltype(_impl_.arlen_){0u}
    , decltype(_impl_.eventflags_){0u}
    , decltype(_impl_.timestamp_){0u}
    , decltype(_impl_.host_timestamp_){uint64_t{0u}}
    , decltype(_impl_.rd_bytes_){0u}
    , decltype(_impl_.wr_bytes_){0u}
  };
}

xclPerfMonReadTrace_response_events::~xclPerfMonReadTrace_response_events() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_response.events)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_response_events::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPerfMonReadTrace_response_events::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_response_events::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_response.events)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    ::memset(&_impl_.awlen_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.wr_bytes_) -
        reinterpret_cast<char*>(&_impl_.awlen_)) + sizeof(_impl_.wr_bytes_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_response_events::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint32 awlen = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_awlen(&has_bits);
          _impl_.awlen_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 arlen = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_arlen(&has_bits);
          _impl_.arlen_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 eventflags = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_eventflags(&has_bits);
          _impl_.eventflags_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 timestamp = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_timestamp(&has_bits);
          _impl_.timestamp_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 host_timestamp = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_host_timestamp(&has_bits);
          _impl_.host_timestamp_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 rd_bytes = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_rd_bytes(&has_bits);
          _impl_.rd_bytes_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 wr_bytes = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_wr_bytes(&has_bits);
          _impl_.wr_bytes_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_response_events::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_response.events)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 awlen = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_awlen(), target);
  }

  // optional uint32 arlen = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_arlen(), target);
  }

  // optional uint32 eventflags = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_eventflags(), target);
  }

  // optional uint32 timestamp = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(4, this->_internal_timestamp(), target);
  }

  // optional uint64 host_timestamp = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_host_timestamp(), target);
  }

  // optional uint32 rd_bytes = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_rd_bytes(), target);
  }

  // optional uint32 wr_bytes = 7;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(7, this->_internal_wr_bytes(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_response.events)
  return target;
}

size_t xclPerfMonReadTrace_response_events::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_response.events)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    // optional uint32 awlen = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_awlen());
    }

    // optional uint32 arlen = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_arlen());
    }

    // optional uint32 eventflags = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_eventflags());
    }

    // optional uint32 timestamp = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_timestamp());
    }

    // optional uint64 host_timestamp = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_host_timestamp());
    }

    // optional uint32 rd_bytes = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_rd_bytes());
    }

    // optional uint32 wr_bytes = 7;
    if (cached_has_bits & 0x00000040u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_wr_bytes());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_response_events::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_response_events::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_response_events::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_response_events::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_response_events*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_response_events&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_response.events)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.awlen_ = from._impl_.awlen_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.arlen_ = from._impl_.arlen_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.eventflags_ = from._impl_.eventflags_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.timestamp_ = from._impl_.timestamp_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.host_timestamp_ = from._impl_.host_timestamp_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.rd_bytes_ = from._impl_.rd_bytes_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.wr_bytes_ = from._impl_.wr_bytes_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_response_events::CopyFrom(const xclPerfMonReadTrace_response_events& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_response.events)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_response_events::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_response_events::InternalSwap(xclPerfMonReadTrace_response_events* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.wr_bytes_)
      + sizeof(xclPerfMonReadTrace_response_events::_impl_.wr_bytes_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.awlen_)>(
          reinterpret_cast<char*>(&_impl_.awlen_),
          reinterpret_cast<char*>(&other->_impl_.awlen_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_response_events::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[36]);
}

// ===================================================================

class xclPerfMonReadTrace_response::_Internal {
 public:
};

xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_response)
}
xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(const xclPerfMonReadTrace_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_.output_data_){from._impl_.output_data_}
    , /*decltype(_impl_._cached_size_)*/{}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_response)
}

inline void xclPerfMonReadTrace_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_.output_data_){arena}
    , /*decltype(_impl_._cached_size_)*/{}
  };
}

xclPerfMonReadTrace_response::~xclPerfMonReadTrace_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.output_data_.~RepeatedPtrField();
}

void xclPerfMonReadTrace_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.output_data_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // repeated .xclPerfMonReadTrace_response.events output_data = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 66)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_output_data(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<66>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_response.events output_data = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_output_data_size()); i < n; i++) {
    const auto& repfield = this->_internal_output_data(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_response)
  return target;
}

size_t xclPerfMonReadTrace_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_response.events output_data = 8;
  total_size += 1UL * this->_internal_output_data_size();
  for (const auto& msg : this->_impl_.output_data_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_response::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.output_data_.MergeFrom(from._impl_.output_data_);
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_response::CopyFrom(const xclPerfMonReadTrace_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_response::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_response::InternalSwap(xclPerfMonReadTrace_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.output_data_.InternalSwap(&other->_impl_.output_data_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[37]);
}

// ===================================================================

class ddr_mem_msg::_Internal {
 public:
  using HasBits = decltype(std::declval<ddr_mem_msg>()._impl_._has_bits_);
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_strb_data(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_fullsize(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

ddr_mem_msg::ddr_mem_msg(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:ddr_mem_msg)
}
ddr_mem_msg::ddr_mem_msg(const ddr_mem_msg& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  ddr_mem_msg* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.strb_data_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.fullsize_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  _impl_.strb_data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.strb_data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_strb_data()) {
    _this->_impl_.strb_data_.Set(from._internal_strb_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.fullsize_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.fullsize_));
  // @@protoc_insertion_point(copy_constructor:ddr_mem_msg)
}

inline void ddr_mem_msg::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.strb_data_){}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.fullsize_){uint64_t{0u}}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.strb_data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.strb_data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

ddr_mem_msg::~ddr_mem_msg() {
  // @@protoc_insertion_point(destructor:ddr_mem_msg)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void ddr_mem_msg::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
  _impl_.strb_data_.Destroy();
}

void ddr_mem_msg::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void ddr_mem_msg::Clear() {
// @@protoc_insertion_point(message_clear_start:ddr_mem_msg)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.data_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.strb_data_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.fullsize_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.fullsize_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* ddr_mem_msg::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes data = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes strb_data = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_strb_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 fullsize = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_fullsize(&has_bits);
          _impl_.fullsize_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* ddr_mem_msg::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:ddr_mem_msg)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes data = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        1, this->_internal_data(), target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_size(), target);
  }

  // optional bytes strb_data = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_strb_data(), target);
  }

  // optional uint64 fullsize = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_fullsize(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:ddr_mem_msg)
  return target;
}

size_t ddr_mem_msg::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:ddr_mem_msg)
  size_t total_size = 0;

  // required bytes data = 1;
  if (_internal_has_data()) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000eu) {
    // optional bytes strb_data = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_strb_data());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

    // optional uint64 fullsize = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_fullsize());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData ddr_mem_msg::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    ddr_mem_msg::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*ddr_mem_msg::GetClassData() const { return &_class_data_; }


void ddr_mem_msg::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<ddr_mem_msg*>(&to_msg);
  auto& from = static_cast<const ddr_mem_msg&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:ddr_mem_msg)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_strb_data(from._internal_strb_data());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.fullsize_ = from._impl_.fullsize_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void ddr_mem_msg::CopyFrom(const ddr_mem_msg& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:ddr_mem_msg)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool ddr_mem_msg::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void ddr_mem_msg::InternalSwap(ddr_mem_msg* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.strb_data_, lhs_arena,
      &other->_impl_.strb_data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.fullsize_)
      + sizeof(ddr_mem_msg::_impl_.fullsize_)
      - PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata ddr_mem_msg::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[38]);
}

// ===================================================================

class xclGetDeviceTimestamp_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDeviceTimestamp_call>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGetDeviceTimestamp_call)
}
xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(const xclGetDeviceTimestamp_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGetDeviceTimestamp_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGetDeviceTimestamp_call)
}

inline void xclGetDeviceTimestamp_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGetDeviceTimestamp_call::~xclGetDeviceTimestamp_call() {
  // @@protoc_insertion_point(destructor:xclGetDeviceTimestamp_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGetDeviceTimestamp_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGetDeviceTimestamp_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGetDeviceTimestamp_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDeviceTimestamp_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGetDeviceTimestamp_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGetDeviceTimestamp_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDeviceTimestamp_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDeviceTimestamp_call)
  return target;
}

size_t xclGetDeviceTimestamp_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDeviceTimestamp_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGetDeviceTimestamp_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGetDeviceTimestamp_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGetDeviceTimestamp_call::GetClassData() const { return &_class_data_; }


void xclGetDeviceTimestamp_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGetDeviceTimestamp_call*>(&to_msg);
  auto& from = static_cast<const xclGetDeviceTimestamp_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDeviceTimestamp_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDeviceTimestamp_call::CopyFrom(const xclGetDeviceTimestamp_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDeviceTimestamp_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGetDeviceTimestamp_call::IsInitialized() const {
  return true;
}

void xclGetDeviceTimestamp_call::InternalSwap(xclGetDeviceTimestamp_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGetDeviceTimestamp_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[39]);
}

// ===================================================================

class xclGetDeviceTimestamp_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDeviceTimestamp_response>()._impl_._has_bits_);
  static void set_has_device_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGetDeviceTimestamp_response)
}
xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(const xclGetDeviceTimestamp_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGetDeviceTimestamp_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.device_timestamp_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.device_timestamp_ = from._impl_.device_timestamp_;
  // @@protoc_insertion_point(copy_constructor:xclGetDeviceTimestamp_response)
}

inline void xclGetDeviceTimestamp_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.device_timestamp_){uint64_t{0u}}
  };
}

xclGetDeviceTimestamp_response::~xclGetDeviceTimestamp_response() {
  // @@protoc_insertion_point(destructor:xclGetDeviceTimestamp_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGetDeviceTimestamp_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGetDeviceTimestamp_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGetDeviceTimestamp_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDeviceTimestamp_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.device_timestamp_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGetDeviceTimestamp_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 device_timestamp = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_device_timestamp(&has_bits);
          _impl_.device_timestamp_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGetDeviceTimestamp_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDeviceTimestamp_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 device_timestamp = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_device_timestamp(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDeviceTimestamp_response)
  return target;
}

size_t xclGetDeviceTimestamp_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDeviceTimestamp_response)
  size_t total_size = 0;

  // required uint64 device_timestamp = 1;
  if (_internal_has_device_timestamp()) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_device_timestamp());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGetDeviceTimestamp_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGetDeviceTimestamp_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGetDeviceTimestamp_response::GetClassData() const { return &_class_data_; }


void xclGetDeviceTimestamp_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGetDeviceTimestamp_response*>(&to_msg);
  auto& from = static_cast<const xclGetDeviceTimestamp_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDeviceTimestamp_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_device_timestamp()) {
    _this->_internal_set_device_timestamp(from._internal_device_timestamp());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDeviceTimestamp_response::CopyFrom(const xclGetDeviceTimestamp_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDeviceTimestamp_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGetDeviceTimestamp_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGetDeviceTimestamp_response::InternalSwap(xclGetDeviceTimestamp_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.device_timestamp_, other->_impl_.device_timestamp_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGetDeviceTimestamp_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[40]);
}

// ===================================================================

class xclReadBusStatus_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadBusStatus_call>()._impl_._has_bits_);
  static void set_has_slot_n(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclReadBusStatus_call::xclReadBusStatus_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadBusStatus_call)
}
xclReadBusStatus_call::xclReadBusStatus_call(const xclReadBusStatus_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadBusStatus_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slot_n_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.slot_n_ = from._impl_.slot_n_;
  // @@protoc_insertion_point(copy_constructor:xclReadBusStatus_call)
}

inline void xclReadBusStatus_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slot_n_){0}
  };
}

xclReadBusStatus_call::~xclReadBusStatus_call() {
  // @@protoc_insertion_point(destructor:xclReadBusStatus_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadBusStatus_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclReadBusStatus_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadBusStatus_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadBusStatus_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.slot_n_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadBusStatus_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional int32 slot_n = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_slot_n(&has_bits);
          _impl_.slot_n_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadBusStatus_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadBusStatus_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional int32 slot_n = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteInt32ToArray(1, this->_internal_slot_n(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadBusStatus_call)
  return target;
}

size_t xclReadBusStatus_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadBusStatus_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional int32 slot_n = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(this->_internal_slot_n());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadBusStatus_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadBusStatus_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadBusStatus_call::GetClassData() const { return &_class_data_; }


void xclReadBusStatus_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadBusStatus_call*>(&to_msg);
  auto& from = static_cast<const xclReadBusStatus_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadBusStatus_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_slot_n()) {
    _this->_internal_set_slot_n(from._internal_slot_n());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadBusStatus_call::CopyFrom(const xclReadBusStatus_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadBusStatus_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadBusStatus_call::IsInitialized() const {
  return true;
}

void xclReadBusStatus_call::InternalSwap(xclReadBusStatus_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.slot_n_, other->_impl_.slot_n_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadBusStatus_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[41]);
}

// ===================================================================

class xclReadBusStatus_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadBusStatus_response>()._impl_._has_bits_);
  static void set_has_uncomplete_rd_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_uncomplete_wr_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_idle_bus_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_uncomplete_rd_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_uncomplete_wr_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclReadBusStatus_response::xclReadBusStatus_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadBusStatus_response)
}
xclReadBusStatus_response::xclReadBusStatus_response(const xclReadBusStatus_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadBusStatus_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.uncomplete_rd_cycles_){}
    , decltype(_impl_.uncomplete_wr_cycles_){}
    , decltype(_impl_.idle_bus_cycles_){}
    , decltype(_impl_.uncomplete_rd_id_){}
    , decltype(_impl_.uncomplete_wr_id_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.uncomplete_rd_cycles_, &from._impl_.uncomplete_rd_cycles_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.uncomplete_wr_id_) -
    reinterpret_cast<char*>(&_impl_.uncomplete_rd_cycles_)) + sizeof(_impl_.uncomplete_wr_id_));
  // @@protoc_insertion_point(copy_constructor:xclReadBusStatus_response)
}

inline void xclReadBusStatus_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.uncomplete_rd_cycles_){uint64_t{0u}}
    , decltype(_impl_.uncomplete_wr_cycles_){uint64_t{0u}}
    , decltype(_impl_.idle_bus_cycles_){uint64_t{0u}}
    , decltype(_impl_.uncomplete_rd_id_){uint64_t{0u}}
    , decltype(_impl_.uncomplete_wr_id_){uint64_t{0u}}
  };
}

xclReadBusStatus_response::~xclReadBusStatus_response() {
  // @@protoc_insertion_point(destructor:xclReadBusStatus_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadBusStatus_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclReadBusStatus_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadBusStatus_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadBusStatus_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    ::memset(&_impl_.uncomplete_rd_cycles_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.uncomplete_wr_id_) -
        reinterpret_cast<char*>(&_impl_.uncomplete_rd_cycles_)) + sizeof(_impl_.uncomplete_wr_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadBusStatus_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 uncomplete_rd_cycles = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_uncomplete_rd_cycles(&has_bits);
          _impl_.uncomplete_rd_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 uncomplete_wr_cycles = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_uncomplete_wr_cycles(&has_bits);
          _impl_.uncomplete_wr_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 idle_bus_cycles = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_idle_bus_cycles(&has_bits);
          _impl_.idle_bus_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 uncomplete_rd_id = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_uncomplete_rd_id(&has_bits);
          _impl_.uncomplete_rd_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 uncomplete_wr_id = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_uncomplete_wr_id(&has_bits);
          _impl_.uncomplete_wr_id_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadBusStatus_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadBusStatus_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 uncomplete_rd_cycles = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_uncomplete_rd_cycles(), target);
  }

  // optional uint64 uncomplete_wr_cycles = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_uncomplete_wr_cycles(), target);
  }

  // optional uint64 idle_bus_cycles = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_idle_bus_cycles(), target);
  }

  // optional uint64 uncomplete_rd_id = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_uncomplete_rd_id(), target);
  }

  // optional uint64 uncomplete_wr_id = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_uncomplete_wr_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadBusStatus_response)
  return target;
}

size_t xclReadBusStatus_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadBusStatus_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional uint64 uncomplete_rd_cycles = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_uncomplete_rd_cycles());
    }

    // optional uint64 uncomplete_wr_cycles = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_uncomplete_wr_cycles());
    }

    // optional uint64 idle_bus_cycles = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_idle_bus_cycles());
    }

    // optional uint64 uncomplete_rd_id = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_uncomplete_rd_id());
    }

    // optional uint64 uncomplete_wr_id = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_uncomplete_wr_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadBusStatus_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadBusStatus_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadBusStatus_response::GetClassData() const { return &_class_data_; }


void xclReadBusStatus_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadBusStatus_response*>(&to_msg);
  auto& from = static_cast<const xclReadBusStatus_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadBusStatus_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.uncomplete_rd_cycles_ = from._impl_.uncomplete_rd_cycles_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.uncomplete_wr_cycles_ = from._impl_.uncomplete_wr_cycles_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.idle_bus_cycles_ = from._impl_.idle_bus_cycles_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.uncomplete_rd_id_ = from._impl_.uncomplete_rd_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.uncomplete_wr_id_ = from._impl_.uncomplete_wr_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadBusStatus_response::CopyFrom(const xclReadBusStatus_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadBusStatus_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadBusStatus_response::IsInitialized() const {
  return true;
}

void xclReadBusStatus_response::InternalSwap(xclReadBusStatus_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_id_)
      + sizeof(xclReadBusStatus_response::_impl_.uncomplete_wr_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_)>(
          reinterpret_cast<char*>(&_impl_.uncomplete_rd_cycles_),
          reinterpret_cast<char*>(&other->_impl_.uncomplete_rd_cycles_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadBusStatus_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[42]);
}

// ===================================================================

class xclGetDebugMessages_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDebugMessages_call>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_force(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclGetDebugMessages_call::xclGetDebugMessages_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGetDebugMessages_call)
}
xclGetDebugMessages_call::xclGetDebugMessages_call(const xclGetDebugMessages_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGetDebugMessages_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}
    , decltype(_impl_.force_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.ack_, &from._impl_.ack_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.force_) -
    reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.force_));
  // @@protoc_insertion_point(copy_constructor:xclGetDebugMessages_call)
}

inline void xclGetDebugMessages_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
    , decltype(_impl_.force_){false}
  };
}

xclGetDebugMessages_call::~xclGetDebugMessages_call() {
  // @@protoc_insertion_point(destructor:xclGetDebugMessages_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGetDebugMessages_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGetDebugMessages_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGetDebugMessages_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDebugMessages_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  ::memset(&_impl_.ack_, 0, static_cast<size_t>(
      reinterpret_cast<char*>(&_impl_.force_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.force_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGetDebugMessages_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool force = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_force(&has_bits);
          _impl_.force_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGetDebugMessages_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDebugMessages_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  // optional bool force = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(2, this->_internal_force(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDebugMessages_call)
  return target;
}

size_t xclGetDebugMessages_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDebugMessages_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bool ack = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + 1;
    }

    // optional bool force = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGetDebugMessages_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGetDebugMessages_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGetDebugMessages_call::GetClassData() const { return &_class_data_; }


void xclGetDebugMessages_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGetDebugMessages_call*>(&to_msg);
  auto& from = static_cast<const xclGetDebugMessages_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDebugMessages_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.force_ = from._impl_.force_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDebugMessages_call::CopyFrom(const xclGetDebugMessages_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDebugMessages_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGetDebugMessages_call::IsInitialized() const {
  return true;
}

void xclGetDebugMessages_call::InternalSwap(xclGetDebugMessages_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.force_)
      + sizeof(xclGetDebugMessages_call::_impl_.force_)
      - PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGetDebugMessages_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[43]);
}

// ===================================================================

class xclGetDebugMessages_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDebugMessages_response>()._impl_._has_bits_);
  static void set_has_display_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_log_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_stop_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclGetDebugMessages_response::xclGetDebugMessages_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGetDebugMessages_response)
}
xclGetDebugMessages_response::xclGetDebugMessages_response(const xclGetDebugMessages_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGetDebugMessages_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.display_msgs_){}
    , decltype(_impl_.log_msgs_){}
    , decltype(_impl_.stop_msgs_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.display_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.display_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_display_msgs()) {
    _this->_impl_.display_msgs_.Set(from._internal_display_msgs(), 
      _this->GetArenaForAllocation());
  }
  _impl_.log_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.log_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_log_msgs()) {
    _this->_impl_.log_msgs_.Set(from._internal_log_msgs(), 
      _this->GetArenaForAllocation());
  }
  _impl_.stop_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.stop_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_stop_msgs()) {
    _this->_impl_.stop_msgs_.Set(from._internal_stop_msgs(), 
      _this->GetArenaForAllocation());
  }
  // @@protoc_insertion_point(copy_constructor:xclGetDebugMessages_response)
}

inline void xclGetDebugMessages_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.display_msgs_){}
    , decltype(_impl_.log_msgs_){}
    , decltype(_impl_.stop_msgs_){}
  };
  _impl_.display_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.display_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.log_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.log_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.stop_msgs_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.stop_msgs_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGetDebugMessages_response::~xclGetDebugMessages_response() {
  // @@protoc_insertion_point(destructor:xclGetDebugMessages_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGetDebugMessages_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.display_msgs_.Destroy();
  _impl_.log_msgs_.Destroy();
  _impl_.stop_msgs_.Destroy();
}

void xclGetDebugMessages_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGetDebugMessages_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDebugMessages_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.display_msgs_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.log_msgs_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.stop_msgs_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGetDebugMessages_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bytes display_msgs = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_display_msgs();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes log_msgs = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_log_msgs();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes stop_msgs = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_stop_msgs();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGetDebugMessages_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDebugMessages_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes display_msgs = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        1, this->_internal_display_msgs(), target);
  }

  // optional bytes log_msgs = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_log_msgs(), target);
  }

  // optional bytes stop_msgs = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_stop_msgs(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDebugMessages_response)
  return target;
}

size_t xclGetDebugMessages_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDebugMessages_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional bytes display_msgs = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_display_msgs());
    }

    // optional bytes log_msgs = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_log_msgs());
    }

    // optional bytes stop_msgs = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_stop_msgs());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGetDebugMessages_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGetDebugMessages_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGetDebugMessages_response::GetClassData() const { return &_class_data_; }


void xclGetDebugMessages_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGetDebugMessages_response*>(&to_msg);
  auto& from = static_cast<const xclGetDebugMessages_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDebugMessages_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_display_msgs(from._internal_display_msgs());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_log_msgs(from._internal_log_msgs());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_stop_msgs(from._internal_stop_msgs());
    }
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDebugMessages_response::CopyFrom(const xclGetDebugMessages_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDebugMessages_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGetDebugMessages_response::IsInitialized() const {
  return true;
}

void xclGetDebugMessages_response::InternalSwap(xclGetDebugMessages_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.display_msgs_, lhs_arena,
      &other->_impl_.display_msgs_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.log_msgs_, lhs_arena,
      &other->_impl_.log_msgs_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.stop_msgs_, lhs_arena,
      &other->_impl_.stop_msgs_, rhs_arena
  );
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGetDebugMessages_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[44]);
}

// ===================================================================

class xclImportBO_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclImportBO_call>()._impl_._has_bits_);
  static void set_has_dst_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclImportBO_call::xclImportBO_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclImportBO_call)
}
xclImportBO_call::xclImportBO_call(const xclImportBO_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclImportBO_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dst_filename_){}
    , decltype(_impl_.offset_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dst_filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dst_filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dst_filename()) {
    _this->_impl_.dst_filename_.Set(from._internal_dst_filename(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.offset_, &from._impl_.offset_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.size_) -
    reinterpret_cast<char*>(&_impl_.offset_)) + sizeof(_impl_.size_));
  // @@protoc_insertion_point(copy_constructor:xclImportBO_call)
}

inline void xclImportBO_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dst_filename_){}
    , decltype(_impl_.offset_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
  _impl_.dst_filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dst_filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclImportBO_call::~xclImportBO_call() {
  // @@protoc_insertion_point(destructor:xclImportBO_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclImportBO_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dst_filename_.Destroy();
}

void xclImportBO_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclImportBO_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclImportBO_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dst_filename_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.offset_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.offset_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclImportBO_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional string dst_filename = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_dst_filename();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclImportBO_call.dst_filename");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional uint64 offset = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_offset(&has_bits);
          _impl_.offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclImportBO_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclImportBO_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional string dst_filename = 1;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_dst_filename().data(), static_cast<int>(this->_internal_dst_filename().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclImportBO_call.dst_filename");
    target = stream->WriteStringMaybeAliased(
        1, this->_internal_dst_filename(), target);
  }

  // optional uint64 offset = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_offset(), target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclImportBO_call)
  return target;
}

size_t xclImportBO_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclImportBO_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string dst_filename = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_dst_filename());
    }

    // optional uint64 offset = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_offset());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclImportBO_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclImportBO_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclImportBO_call::GetClassData() const { return &_class_data_; }


void xclImportBO_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclImportBO_call*>(&to_msg);
  auto& from = static_cast<const xclImportBO_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclImportBO_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dst_filename(from._internal_dst_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclImportBO_call::CopyFrom(const xclImportBO_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclImportBO_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclImportBO_call::IsInitialized() const {
  return true;
}

void xclImportBO_call::InternalSwap(xclImportBO_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dst_filename_, lhs_arena,
      &other->_impl_.dst_filename_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.size_)
      + sizeof(xclImportBO_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.offset_)>(
          reinterpret_cast<char*>(&_impl_.offset_),
          reinterpret_cast<char*>(&other->_impl_.offset_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclImportBO_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[45]);
}

// ===================================================================

class xclImportBO_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclImportBO_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclImportBO_response::xclImportBO_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclImportBO_response)
}
xclImportBO_response::xclImportBO_response(const xclImportBO_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclImportBO_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclImportBO_response)
}

inline void xclImportBO_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclImportBO_response::~xclImportBO_response() {
  // @@protoc_insertion_point(destructor:xclImportBO_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclImportBO_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclImportBO_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclImportBO_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclImportBO_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclImportBO_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclImportBO_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclImportBO_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclImportBO_response)
  return target;
}

size_t xclImportBO_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclImportBO_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclImportBO_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclImportBO_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclImportBO_response::GetClassData() const { return &_class_data_; }


void xclImportBO_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclImportBO_response*>(&to_msg);
  auto& from = static_cast<const xclImportBO_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclImportBO_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclImportBO_response::CopyFrom(const xclImportBO_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclImportBO_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclImportBO_response::IsInitialized() const {
  return true;
}

void xclImportBO_response::InternalSwap(xclImportBO_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclImportBO_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[46]);
}

// ===================================================================

class xclCopyBO_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBO_call>()._impl_._has_bits_);
  static void set_has_src_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dst_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_src_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_dst_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclCopyBO_call::xclCopyBO_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBO_call)
}
xclCopyBO_call::xclCopyBO_call(const xclCopyBO_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBO_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dst_filename_){}
    , decltype(_impl_.src_handle_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.src_offset_){}
    , decltype(_impl_.dst_offset_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dst_filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dst_filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dst_filename()) {
    _this->_impl_.dst_filename_.Set(from._internal_dst_filename(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.src_handle_, &from._impl_.src_handle_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.dst_offset_) -
    reinterpret_cast<char*>(&_impl_.src_handle_)) + sizeof(_impl_.dst_offset_));
  // @@protoc_insertion_point(copy_constructor:xclCopyBO_call)
}

inline void xclCopyBO_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dst_filename_){}
    , decltype(_impl_.src_handle_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.src_offset_){uint64_t{0u}}
    , decltype(_impl_.dst_offset_){uint64_t{0u}}
  };
  _impl_.dst_filename_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dst_filename_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclCopyBO_call::~xclCopyBO_call() {
  // @@protoc_insertion_point(destructor:xclCopyBO_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBO_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dst_filename_.Destroy();
}

void xclCopyBO_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBO_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBO_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dst_filename_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.src_handle_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.dst_offset_) -
        reinterpret_cast<char*>(&_impl_.src_handle_)) + sizeof(_impl_.dst_offset_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBO_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 src_handle = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_src_handle(&has_bits);
          _impl_.src_handle_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string dst_filename = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_dst_filename();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclCopyBO_call.dst_filename");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 src_offset = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_src_offset(&has_bits);
          _impl_.src_offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 dst_offset = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_dst_offset(&has_bits);
          _impl_.dst_offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBO_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBO_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 src_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_src_handle(), target);
  }

  // optional string dst_filename = 2;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_dst_filename().data(), static_cast<int>(this->_internal_dst_filename().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclCopyBO_call.dst_filename");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_dst_filename(), target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_size(), target);
  }

  // optional uint64 src_offset = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_src_offset(), target);
  }

  // optional uint64 dst_offset = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_dst_offset(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBO_call)
  return target;
}

size_t xclCopyBO_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBO_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional string dst_filename = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_dst_filename());
    }

    // optional uint64 src_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_src_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

    // optional uint64 src_offset = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_src_offset());
    }

    // optional uint64 dst_offset = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_dst_offset());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBO_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBO_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBO_call::GetClassData() const { return &_class_data_; }


void xclCopyBO_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBO_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBO_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBO_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dst_filename(from._internal_dst_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.src_handle_ = from._impl_.src_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.src_offset_ = from._impl_.src_offset_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.dst_offset_ = from._impl_.dst_offset_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBO_call::CopyFrom(const xclCopyBO_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBO_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBO_call::IsInitialized() const {
  return true;
}

void xclCopyBO_call::InternalSwap(xclCopyBO_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dst_filename_, lhs_arena,
      &other->_impl_.dst_filename_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_offset_)
      + sizeof(xclCopyBO_call::_impl_.dst_offset_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_handle_)>(
          reinterpret_cast<char*>(&_impl_.src_handle_),
          reinterpret_cast<char*>(&other->_impl_.src_handle_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBO_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[47]);
}

// ===================================================================

class xclCopyBO_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBO_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclCopyBO_response::xclCopyBO_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCopyBO_response)
}
xclCopyBO_response::xclCopyBO_response(const xclCopyBO_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCopyBO_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclCopyBO_response)
}

inline void xclCopyBO_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclCopyBO_response::~xclCopyBO_response() {
  // @@protoc_insertion_point(destructor:xclCopyBO_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCopyBO_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclCopyBO_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCopyBO_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBO_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCopyBO_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCopyBO_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBO_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBO_response)
  return target;
}

size_t xclCopyBO_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBO_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCopyBO_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCopyBO_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCopyBO_response::GetClassData() const { return &_class_data_; }


void xclCopyBO_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBO_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBO_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBO_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBO_response::CopyFrom(const xclCopyBO_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBO_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCopyBO_response::IsInitialized() const {
  return true;
}

void xclCopyBO_response::InternalSwap(xclCopyBO_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCopyBO_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[48]);
}

// ===================================================================

class xclCreateQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCreateQueue_call>()._impl_._has_bits_);
  static void set_has_write(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_type(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_state(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_route(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_flow(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_qsize(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_desc_size(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
  static void set_has_flags(HasBits* has_bits) {
    (*has_bits)[0] |= 128u;
  }
};

xclCreateQueue_call::xclCreateQueue_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCreateQueue_call)
}
xclCreateQueue_call::xclCreateQueue_call(const xclCreateQueue_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCreateQueue_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.write_){}
    , decltype(_impl_.type_){}
    , decltype(_impl_.state_){}
    , decltype(_impl_.route_){}
    , decltype(_impl_.flow_){}
    , decltype(_impl_.qsize_){}
    , decltype(_impl_.desc_size_){}
    , decltype(_impl_.flags_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.write_, &from._impl_.write_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.flags_) -
    reinterpret_cast<char*>(&_impl_.write_)) + sizeof(_impl_.flags_));
  // @@protoc_insertion_point(copy_constructor:xclCreateQueue_call)
}

inline void xclCreateQueue_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.write_){false}
    , decltype(_impl_.type_){0u}
    , decltype(_impl_.state_){0u}
    , decltype(_impl_.route_){0u}
    , decltype(_impl_.flow_){uint64_t{0u}}
    , decltype(_impl_.qsize_){0u}
    , decltype(_impl_.desc_size_){0u}
    , decltype(_impl_.flags_){uint64_t{0u}}
  };
}

xclCreateQueue_call::~xclCreateQueue_call() {
  // @@protoc_insertion_point(destructor:xclCreateQueue_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCreateQueue_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclCreateQueue_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCreateQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCreateQueue_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    ::memset(&_impl_.write_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.flags_) -
        reinterpret_cast<char*>(&_impl_.write_)) + sizeof(_impl_.flags_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCreateQueue_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool write = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_write(&has_bits);
          _impl_.write_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 type = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_type(&has_bits);
          _impl_.type_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 state = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_state(&has_bits);
          _impl_.state_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 route = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_route(&has_bits);
          _impl_.route_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 flow = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_flow(&has_bits);
          _impl_.flow_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 qsize = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_qsize(&has_bits);
          _impl_.qsize_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 desc_size = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_desc_size(&has_bits);
          _impl_.desc_size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 flags = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 64)) {
          _Internal::set_has_flags(&has_bits);
          _impl_.flags_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCreateQueue_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCreateQueue_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool write = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_write(), target);
  }

  // optional uint32 type = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_type(), target);
  }

  // optional uint32 state = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_state(), target);
  }

  // optional uint32 route = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(4, this->_internal_route(), target);
  }

  // optional uint64 flow = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_flow(), target);
  }

  // optional uint32 qsize = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(6, this->_internal_qsize(), target);
  }

  // optional uint32 desc_size = 7;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(7, this->_internal_desc_size(), target);
  }

  // optional uint64 flags = 8;
  if (cached_has_bits & 0x00000080u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(8, this->_internal_flags(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCreateQueue_call)
  return target;
}

size_t xclCreateQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCreateQueue_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    // optional bool write = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + 1;
    }

    // optional uint32 type = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_type());
    }

    // optional uint32 state = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_state());
    }

    // optional uint32 route = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_route());
    }

    // optional uint64 flow = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_flow());
    }

    // optional uint32 qsize = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_qsize());
    }

    // optional uint32 desc_size = 7;
    if (cached_has_bits & 0x00000040u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_desc_size());
    }

    // optional uint64 flags = 8;
    if (cached_has_bits & 0x00000080u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_flags());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCreateQueue_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCreateQueue_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCreateQueue_call::GetClassData() const { return &_class_data_; }


void xclCreateQueue_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCreateQueue_call*>(&to_msg);
  auto& from = static_cast<const xclCreateQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCreateQueue_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.write_ = from._impl_.write_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.type_ = from._impl_.type_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.state_ = from._impl_.state_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.route_ = from._impl_.route_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.flow_ = from._impl_.flow_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.qsize_ = from._impl_.qsize_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.desc_size_ = from._impl_.desc_size_;
    }
    if (cached_has_bits & 0x00000080u) {
      _this->_impl_.flags_ = from._impl_.flags_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCreateQueue_call::CopyFrom(const xclCreateQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCreateQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCreateQueue_call::IsInitialized() const {
  return true;
}

void xclCreateQueue_call::InternalSwap(xclCreateQueue_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flags_)
      + sizeof(xclCreateQueue_call::_impl_.flags_)
      - PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.write_)>(
          reinterpret_cast<char*>(&_impl_.write_),
          reinterpret_cast<char*>(&other->_impl_.write_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCreateQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[49]);
}

// ===================================================================

class xclCreateQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCreateQueue_response>()._impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclCreateQueue_response::xclCreateQueue_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclCreateQueue_response)
}
xclCreateQueue_response::xclCreateQueue_response(const xclCreateQueue_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclCreateQueue_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.q_handle_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.q_handle_ = from._impl_.q_handle_;
  // @@protoc_insertion_point(copy_constructor:xclCreateQueue_response)
}

inline void xclCreateQueue_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.q_handle_){uint64_t{0u}}
  };
}

xclCreateQueue_response::~xclCreateQueue_response() {
  // @@protoc_insertion_point(destructor:xclCreateQueue_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclCreateQueue_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclCreateQueue_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclCreateQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCreateQueue_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.q_handle_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclCreateQueue_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 q_handle = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_q_handle(&has_bits);
          _impl_.q_handle_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclCreateQueue_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCreateQueue_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_q_handle(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCreateQueue_response)
  return target;
}

size_t xclCreateQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCreateQueue_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 q_handle = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_q_handle());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclCreateQueue_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclCreateQueue_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclCreateQueue_response::GetClassData() const { return &_class_data_; }


void xclCreateQueue_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclCreateQueue_response*>(&to_msg);
  auto& from = static_cast<const xclCreateQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCreateQueue_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_q_handle()) {
    _this->_internal_set_q_handle(from._internal_q_handle());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclCreateQueue_response::CopyFrom(const xclCreateQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCreateQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclCreateQueue_response::IsInitialized() const {
  return true;
}

void xclCreateQueue_response::InternalSwap(xclCreateQueue_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.q_handle_, other->_impl_.q_handle_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclCreateQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[50]);
}

// ===================================================================

class xclWriteQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteQueue_call>()._impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_nonblocking(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_eot(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclWriteQueue_call::xclWriteQueue_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteQueue_call)
}
xclWriteQueue_call::xclWriteQueue_call(const xclWriteQueue_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteQueue_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.src_){}
    , decltype(_impl_.q_handle_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.req_){}
    , decltype(_impl_.nonblocking_){}
    , decltype(_impl_.eot_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.src_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.src_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_src()) {
    _this->_impl_.src_.Set(from._internal_src(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.q_handle_, &from._impl_.q_handle_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.eot_) -
    reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  // @@protoc_insertion_point(copy_constructor:xclWriteQueue_call)
}

inline void xclWriteQueue_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.src_){}
    , decltype(_impl_.q_handle_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.req_){uint64_t{0u}}
    , decltype(_impl_.nonblocking_){false}
    , decltype(_impl_.eot_){false}
  };
  _impl_.src_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.src_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclWriteQueue_call::~xclWriteQueue_call() {
  // @@protoc_insertion_point(destructor:xclWriteQueue_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteQueue_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.src_.Destroy();
}

void xclWriteQueue_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteQueue_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.src_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.q_handle_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.eot_) -
        reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteQueue_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 q_handle = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_q_handle(&has_bits);
          _impl_.q_handle_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes src = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_src();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 req = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_req(&has_bits);
          _impl_.req_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool nonblocking = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_nonblocking(&has_bits);
          _impl_.nonblocking_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool eot = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_eot(&has_bits);
          _impl_.eot_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteQueue_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteQueue_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_q_handle(), target);
  }

  // optional bytes src = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_src(), target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_size(), target);
  }

  // optional uint64 req = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_req(), target);
  }

  // optional bool nonblocking = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(5, this->_internal_nonblocking(), target);
  }

  // optional bool eot = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(6, this->_internal_eot(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteQueue_call)
  return target;
}

size_t xclWriteQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteQueue_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional bytes src = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_src());
    }

    // optional uint64 q_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_q_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

    // optional uint64 req = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_req());
    }

    // optional bool nonblocking = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += 1 + 1;
    }

    // optional bool eot = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteQueue_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteQueue_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteQueue_call::GetClassData() const { return &_class_data_; }


void xclWriteQueue_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteQueue_call*>(&to_msg);
  auto& from = static_cast<const xclWriteQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteQueue_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_src(from._internal_src());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.q_handle_ = from._impl_.q_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.req_ = from._impl_.req_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.nonblocking_ = from._impl_.nonblocking_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.eot_ = from._impl_.eot_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteQueue_call::CopyFrom(const xclWriteQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteQueue_call::IsInitialized() const {
  return true;
}

void xclWriteQueue_call::InternalSwap(xclWriteQueue_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.src_, lhs_arena,
      &other->_impl_.src_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.eot_)
      + sizeof(xclWriteQueue_call::_impl_.eot_)
      - PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.q_handle_)>(
          reinterpret_cast<char*>(&_impl_.q_handle_),
          reinterpret_cast<char*>(&other->_impl_.q_handle_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[51]);
}

// ===================================================================

class xclWriteQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteQueue_response>()._impl_._has_bits_);
  static void set_has_written_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclWriteQueue_response::xclWriteQueue_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclWriteQueue_response)
}
xclWriteQueue_response::xclWriteQueue_response(const xclWriteQueue_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclWriteQueue_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.written_size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.written_size_ = from._impl_.written_size_;
  // @@protoc_insertion_point(copy_constructor:xclWriteQueue_response)
}

inline void xclWriteQueue_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.written_size_){uint64_t{0u}}
  };
}

xclWriteQueue_response::~xclWriteQueue_response() {
  // @@protoc_insertion_point(destructor:xclWriteQueue_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclWriteQueue_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclWriteQueue_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclWriteQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteQueue_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.written_size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclWriteQueue_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 written_size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_written_size(&has_bits);
          _impl_.written_size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclWriteQueue_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteQueue_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 written_size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_written_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteQueue_response)
  return target;
}

size_t xclWriteQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteQueue_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 written_size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_written_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclWriteQueue_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclWriteQueue_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclWriteQueue_response::GetClassData() const { return &_class_data_; }


void xclWriteQueue_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclWriteQueue_response*>(&to_msg);
  auto& from = static_cast<const xclWriteQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteQueue_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_written_size()) {
    _this->_internal_set_written_size(from._internal_written_size());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteQueue_response::CopyFrom(const xclWriteQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclWriteQueue_response::IsInitialized() const {
  return true;
}

void xclWriteQueue_response::InternalSwap(xclWriteQueue_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.written_size_, other->_impl_.written_size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclWriteQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[52]);
}

// ===================================================================

class xclReadQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadQueue_call>()._impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_nonblocking(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_eot(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclReadQueue_call::xclReadQueue_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadQueue_call)
}
xclReadQueue_call::xclReadQueue_call(const xclReadQueue_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadQueue_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.q_handle_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.req_){}
    , decltype(_impl_.nonblocking_){}
    , decltype(_impl_.eot_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dest()) {
    _this->_impl_.dest_.Set(from._internal_dest(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.q_handle_, &from._impl_.q_handle_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.eot_) -
    reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  // @@protoc_insertion_point(copy_constructor:xclReadQueue_call)
}

inline void xclReadQueue_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.q_handle_){uint64_t{0u}}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.req_){uint64_t{0u}}
    , decltype(_impl_.nonblocking_){false}
    , decltype(_impl_.eot_){false}
  };
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadQueue_call::~xclReadQueue_call() {
  // @@protoc_insertion_point(destructor:xclReadQueue_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadQueue_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dest_.Destroy();
}

void xclReadQueue_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadQueue_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.q_handle_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.eot_) -
        reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadQueue_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 q_handle = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_q_handle(&has_bits);
          _impl_.q_handle_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bytes dest = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_dest();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 req = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_req(&has_bits);
          _impl_.req_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool nonblocking = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_nonblocking(&has_bits);
          _impl_.nonblocking_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional bool eot = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_eot(&has_bits);
          _impl_.eot_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadQueue_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadQueue_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_q_handle(), target);
  }

  // optional bytes dest = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_dest(), target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_size(), target);
  }

  // optional uint64 req = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_req(), target);
  }

  // optional bool nonblocking = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(5, this->_internal_nonblocking(), target);
  }

  // optional bool eot = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(6, this->_internal_eot(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadQueue_call)
  return target;
}

size_t xclReadQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadQueue_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional bytes dest = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_dest());
    }

    // optional uint64 q_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_q_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

    // optional uint64 req = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_req());
    }

    // optional bool nonblocking = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += 1 + 1;
    }

    // optional bool eot = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadQueue_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadQueue_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadQueue_call::GetClassData() const { return &_class_data_; }


void xclReadQueue_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadQueue_call*>(&to_msg);
  auto& from = static_cast<const xclReadQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadQueue_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.q_handle_ = from._impl_.q_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.req_ = from._impl_.req_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.nonblocking_ = from._impl_.nonblocking_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.eot_ = from._impl_.eot_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadQueue_call::CopyFrom(const xclReadQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadQueue_call::IsInitialized() const {
  return true;
}

void xclReadQueue_call::InternalSwap(xclReadQueue_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dest_, lhs_arena,
      &other->_impl_.dest_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.eot_)
      + sizeof(xclReadQueue_call::_impl_.eot_)
      - PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.q_handle_)>(
          reinterpret_cast<char*>(&_impl_.q_handle_),
          reinterpret_cast<char*>(&other->_impl_.q_handle_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[53]);
}

// ===================================================================

class xclReadQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadQueue_response>()._impl_._has_bits_);
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclReadQueue_response::xclReadQueue_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclReadQueue_response)
}
xclReadQueue_response::xclReadQueue_response(const xclReadQueue_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclReadQueue_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dest()) {
    _this->_impl_.dest_.Set(from._internal_dest(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclReadQueue_response)
}

inline void xclReadQueue_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclReadQueue_response::~xclReadQueue_response() {
  // @@protoc_insertion_point(destructor:xclReadQueue_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclReadQueue_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dest_.Destroy();
}

void xclReadQueue_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclReadQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadQueue_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclReadQueue_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bytes dest = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_dest();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclReadQueue_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadQueue_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes dest = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        1, this->_internal_dest(), target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadQueue_response)
  return target;
}

size_t xclReadQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadQueue_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes dest = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_dest());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclReadQueue_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclReadQueue_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclReadQueue_response::GetClassData() const { return &_class_data_; }


void xclReadQueue_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclReadQueue_response*>(&to_msg);
  auto& from = static_cast<const xclReadQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadQueue_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadQueue_response::CopyFrom(const xclReadQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclReadQueue_response::IsInitialized() const {
  return true;
}

void xclReadQueue_response::InternalSwap(xclReadQueue_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dest_, lhs_arena,
      &other->_impl_.dest_, rhs_arena
  );
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclReadQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[54]);
}

// ===================================================================

class xclPollCompletion_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPollCompletion_call>()._impl_._has_bits_);
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPollCompletion_call::xclPollCompletion_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_call)
}
xclPollCompletion_call::xclPollCompletion_call(const xclPollCompletion_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPollCompletion_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.req_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.req_ = from._impl_.req_;
  // @@protoc_insertion_point(copy_constructor:xclPollCompletion_call)
}

inline void xclPollCompletion_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.req_){uint64_t{0u}}
  };
}

xclPollCompletion_call::~xclPollCompletion_call() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPollCompletion_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPollCompletion_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPollCompletion_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.req_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPollCompletion_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 req = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_req(&has_bits);
          _impl_.req_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPollCompletion_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 req = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_req(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_call)
  return target;
}

size_t xclPollCompletion_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 req = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_req());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPollCompletion_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPollCompletion_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPollCompletion_call::GetClassData() const { return &_class_data_; }


void xclPollCompletion_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_call*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_req()) {
    _this->_internal_set_req(from._internal_req());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_call::CopyFrom(const xclPollCompletion_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPollCompletion_call::IsInitialized() const {
  return true;
}

void xclPollCompletion_call::InternalSwap(xclPollCompletion_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.req_, other->_impl_.req_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPollCompletion_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[55]);
}

// ===================================================================

class xclPollCompletion_response_request::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPollCompletion_response_request>()._impl_._has_bits_);
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclPollCompletion_response_request::xclPollCompletion_response_request(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_response.request)
}
xclPollCompletion_response_request::xclPollCompletion_response_request(const xclPollCompletion_response_request& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPollCompletion_response_request* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_dest()) {
    _this->_impl_.dest_.Set(from._internal_dest(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclPollCompletion_response.request)
}

inline void xclPollCompletion_response_request::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.dest_){}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
  _impl_.dest_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.dest_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPollCompletion_response_request::~xclPollCompletion_response_request() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_response.request)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPollCompletion_response_request::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.dest_.Destroy();
}

void xclPollCompletion_response_request::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPollCompletion_response_request::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_response.request)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPollCompletion_response_request::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bytes dest = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_dest();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPollCompletion_response_request::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_response.request)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes dest = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        1, this->_internal_dest(), target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_response.request)
  return target;
}

size_t xclPollCompletion_response_request::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_response.request)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes dest = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
          this->_internal_dest());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPollCompletion_response_request::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPollCompletion_response_request::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPollCompletion_response_request::GetClassData() const { return &_class_data_; }


void xclPollCompletion_response_request::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_response_request*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_response_request&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_response.request)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_response_request::CopyFrom(const xclPollCompletion_response_request& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_response.request)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPollCompletion_response_request::IsInitialized() const {
  return true;
}

void xclPollCompletion_response_request::InternalSwap(xclPollCompletion_response_request* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.dest_, lhs_arena,
      &other->_impl_.dest_, rhs_arena
  );
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPollCompletion_response_request::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[56]);
}

// ===================================================================

class xclPollCompletion_response::_Internal {
 public:
};

xclPollCompletion_response::xclPollCompletion_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_response)
}
xclPollCompletion_response::xclPollCompletion_response(const xclPollCompletion_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPollCompletion_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_.fullrequest_){from._impl_.fullrequest_}
    , /*decltype(_impl_._cached_size_)*/{}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  // @@protoc_insertion_point(copy_constructor:xclPollCompletion_response)
}

inline void xclPollCompletion_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_.fullrequest_){arena}
    , /*decltype(_impl_._cached_size_)*/{}
  };
}

xclPollCompletion_response::~xclPollCompletion_response() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPollCompletion_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.fullrequest_.~RepeatedPtrField();
}

void xclPollCompletion_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPollCompletion_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.fullrequest_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPollCompletion_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // repeated .xclPollCompletion_response.request fullrequest = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_fullrequest(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<26>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPollCompletion_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  // repeated .xclPollCompletion_response.request fullrequest = 3;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_fullrequest_size()); i < n; i++) {
    const auto& repfield = this->_internal_fullrequest(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(3, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_response)
  return target;
}

size_t xclPollCompletion_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPollCompletion_response.request fullrequest = 3;
  total_size += 1UL * this->_internal_fullrequest_size();
  for (const auto& msg : this->_impl_.fullrequest_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPollCompletion_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPollCompletion_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPollCompletion_response::GetClassData() const { return &_class_data_; }


void xclPollCompletion_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_response*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.fullrequest_.MergeFrom(from._impl_.fullrequest_);
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_response::CopyFrom(const xclPollCompletion_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPollCompletion_response::IsInitialized() const {
  return true;
}

void xclPollCompletion_response::InternalSwap(xclPollCompletion_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.fullrequest_.InternalSwap(&other->_impl_.fullrequest_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPollCompletion_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[57]);
}

// ===================================================================

class xclDestroyQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclDestroyQueue_call>()._impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclDestroyQueue_call::xclDestroyQueue_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclDestroyQueue_call)
}
xclDestroyQueue_call::xclDestroyQueue_call(const xclDestroyQueue_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclDestroyQueue_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.q_handle_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.q_handle_ = from._impl_.q_handle_;
  // @@protoc_insertion_point(copy_constructor:xclDestroyQueue_call)
}

inline void xclDestroyQueue_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.q_handle_){uint64_t{0u}}
  };
}

xclDestroyQueue_call::~xclDestroyQueue_call() {
  // @@protoc_insertion_point(destructor:xclDestroyQueue_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclDestroyQueue_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclDestroyQueue_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclDestroyQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclDestroyQueue_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.q_handle_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclDestroyQueue_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 q_handle = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_q_handle(&has_bits);
          _impl_.q_handle_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclDestroyQueue_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclDestroyQueue_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_q_handle(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclDestroyQueue_call)
  return target;
}

size_t xclDestroyQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclDestroyQueue_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 q_handle = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_q_handle());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclDestroyQueue_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclDestroyQueue_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclDestroyQueue_call::GetClassData() const { return &_class_data_; }


void xclDestroyQueue_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclDestroyQueue_call*>(&to_msg);
  auto& from = static_cast<const xclDestroyQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclDestroyQueue_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_q_handle()) {
    _this->_internal_set_q_handle(from._internal_q_handle());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclDestroyQueue_call::CopyFrom(const xclDestroyQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclDestroyQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclDestroyQueue_call::IsInitialized() const {
  return true;
}

void xclDestroyQueue_call::InternalSwap(xclDestroyQueue_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.q_handle_, other->_impl_.q_handle_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclDestroyQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[58]);
}

// ===================================================================

class xclDestroyQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclDestroyQueue_response>()._impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclDestroyQueue_response::xclDestroyQueue_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclDestroyQueue_response)
}
xclDestroyQueue_response::xclDestroyQueue_response(const xclDestroyQueue_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclDestroyQueue_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.success_ = from._impl_.success_;
  // @@protoc_insertion_point(copy_constructor:xclDestroyQueue_response)
}

inline void xclDestroyQueue_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){false}
  };
}

xclDestroyQueue_response::~xclDestroyQueue_response() {
  // @@protoc_insertion_point(destructor:xclDestroyQueue_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclDestroyQueue_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclDestroyQueue_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclDestroyQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclDestroyQueue_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclDestroyQueue_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool success = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_success(&has_bits);
          _impl_.success_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclDestroyQueue_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclDestroyQueue_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclDestroyQueue_response)
  return target;
}

size_t xclDestroyQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclDestroyQueue_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclDestroyQueue_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclDestroyQueue_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclDestroyQueue_response::GetClassData() const { return &_class_data_; }


void xclDestroyQueue_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclDestroyQueue_response*>(&to_msg);
  auto& from = static_cast<const xclDestroyQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclDestroyQueue_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_success()) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclDestroyQueue_response::CopyFrom(const xclDestroyQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclDestroyQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclDestroyQueue_response::IsInitialized() const {
  return true;
}

void xclDestroyQueue_response::InternalSwap(xclDestroyQueue_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.success_, other->_impl_.success_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclDestroyQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[59]);
}

// ===================================================================

class xclSetupInstance_call_argflowpair::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_call_argflowpair>()._impl_._has_bits_);
  static void set_has_arg(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_flow(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_tag(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_call.argflowpair)
}
xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(const xclSetupInstance_call_argflowpair& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetupInstance_call_argflowpair* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.tag_){}
    , decltype(_impl_.arg_){}
    , decltype(_impl_.flow_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.tag_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.tag_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_tag()) {
    _this->_impl_.tag_.Set(from._internal_tag(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.arg_, &from._impl_.arg_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.flow_) -
    reinterpret_cast<char*>(&_impl_.arg_)) + sizeof(_impl_.flow_));
  // @@protoc_insertion_point(copy_constructor:xclSetupInstance_call.argflowpair)
}

inline void xclSetupInstance_call_argflowpair::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.tag_){}
    , decltype(_impl_.arg_){uint64_t{0u}}
    , decltype(_impl_.flow_){uint64_t{0u}}
  };
  _impl_.tag_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.tag_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclSetupInstance_call_argflowpair::~xclSetupInstance_call_argflowpair() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_call.argflowpair)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetupInstance_call_argflowpair::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.tag_.Destroy();
}

void xclSetupInstance_call_argflowpair::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetupInstance_call_argflowpair::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_call.argflowpair)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.tag_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.arg_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.flow_) -
        reinterpret_cast<char*>(&_impl_.arg_)) + sizeof(_impl_.flow_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetupInstance_call_argflowpair::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 arg = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_arg(&has_bits);
          _impl_.arg_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 flow = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_flow(&has_bits);
          _impl_.flow_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string tag = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_tag();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclSetupInstance_call.argflowpair.tag");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetupInstance_call_argflowpair::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_call.argflowpair)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 arg = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(2, this->_internal_arg(), target);
  }

  // optional uint64 flow = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_flow(), target);
  }

  // optional string tag = 4;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_tag().data(), static_cast<int>(this->_internal_tag().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclSetupInstance_call.argflowpair.tag");
    target = stream->WriteStringMaybeAliased(
        4, this->_internal_tag(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_call.argflowpair)
  return target;
}

size_t xclSetupInstance_call_argflowpair::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_call.argflowpair)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string tag = 4;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_tag());
    }

    // optional uint64 arg = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_arg());
    }

    // optional uint64 flow = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_flow());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetupInstance_call_argflowpair::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetupInstance_call_argflowpair::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetupInstance_call_argflowpair::GetClassData() const { return &_class_data_; }


void xclSetupInstance_call_argflowpair::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_call_argflowpair*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_call_argflowpair&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_call.argflowpair)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_tag(from._internal_tag());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.arg_ = from._impl_.arg_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.flow_ = from._impl_.flow_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_call_argflowpair::CopyFrom(const xclSetupInstance_call_argflowpair& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_call.argflowpair)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetupInstance_call_argflowpair::IsInitialized() const {
  return true;
}

void xclSetupInstance_call_argflowpair::InternalSwap(xclSetupInstance_call_argflowpair* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.tag_, lhs_arena,
      &other->_impl_.tag_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.flow_)
      + sizeof(xclSetupInstance_call_argflowpair::_impl_.flow_)
      - PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.arg_)>(
          reinterpret_cast<char*>(&_impl_.arg_),
          reinterpret_cast<char*>(&other->_impl_.arg_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetupInstance_call_argflowpair::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[60]);
}

// ===================================================================

class xclSetupInstance_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_call>()._impl_._has_bits_);
  static void set_has_route(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_call::xclSetupInstance_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_call)
}
xclSetupInstance_call::xclSetupInstance_call(const xclSetupInstance_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetupInstance_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.setup_){from._impl_.setup_}
    , decltype(_impl_.route_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.route_ = from._impl_.route_;
  // @@protoc_insertion_point(copy_constructor:xclSetupInstance_call)
}

inline void xclSetupInstance_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.setup_){arena}
    , decltype(_impl_.route_){uint64_t{0u}}
  };
}

xclSetupInstance_call::~xclSetupInstance_call() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetupInstance_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.setup_.~RepeatedPtrField();
}

void xclSetupInstance_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetupInstance_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.setup_.Clear();
  _impl_.route_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetupInstance_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 route = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_route(&has_bits);
          _impl_.route_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // repeated .xclSetupInstance_call.argflowpair setup = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 42)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_setup(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<42>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetupInstance_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 route = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_route(), target);
  }

  // repeated .xclSetupInstance_call.argflowpair setup = 5;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_setup_size()); i < n; i++) {
    const auto& repfield = this->_internal_setup(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(5, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_call)
  return target;
}

size_t xclSetupInstance_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclSetupInstance_call.argflowpair setup = 5;
  total_size += 1UL * this->_internal_setup_size();
  for (const auto& msg : this->_impl_.setup_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  // optional uint64 route = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_route());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetupInstance_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetupInstance_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetupInstance_call::GetClassData() const { return &_class_data_; }


void xclSetupInstance_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_call*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.setup_.MergeFrom(from._impl_.setup_);
  if (from._internal_has_route()) {
    _this->_internal_set_route(from._internal_route());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_call::CopyFrom(const xclSetupInstance_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetupInstance_call::IsInitialized() const {
  return true;
}

void xclSetupInstance_call::InternalSwap(xclSetupInstance_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.setup_.InternalSwap(&other->_impl_.setup_);
  swap(_impl_.route_, other->_impl_.route_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetupInstance_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[61]);
}

// ===================================================================

class xclSetupInstance_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_response>()._impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_response::xclSetupInstance_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_response)
}
xclSetupInstance_response::xclSetupInstance_response(const xclSetupInstance_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSetupInstance_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.success_ = from._impl_.success_;
  // @@protoc_insertion_point(copy_constructor:xclSetupInstance_response)
}

inline void xclSetupInstance_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){false}
  };
}

xclSetupInstance_response::~xclSetupInstance_response() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSetupInstance_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclSetupInstance_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSetupInstance_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSetupInstance_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool success = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_success(&has_bits);
          _impl_.success_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSetupInstance_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_response)
  return target;
}

size_t xclSetupInstance_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSetupInstance_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSetupInstance_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSetupInstance_response::GetClassData() const { return &_class_data_; }


void xclSetupInstance_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_response*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_success()) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_response::CopyFrom(const xclSetupInstance_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSetupInstance_response::IsInitialized() const {
  return true;
}

void xclSetupInstance_response::InternalSwap(xclSetupInstance_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.success_, other->_impl_.success_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSetupInstance_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[62]);
}

// ===================================================================

class swemuDriverVersion_call::_Internal {
 public:
  using HasBits = decltype(std::declval<swemuDriverVersion_call>()._impl_._has_bits_);
  static void set_has_version(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

swemuDriverVersion_call::swemuDriverVersion_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:swemuDriverVersion_call)
}
swemuDriverVersion_call::swemuDriverVersion_call(const swemuDriverVersion_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  swemuDriverVersion_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.version_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.version_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.version_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_version()) {
    _this->_impl_.version_.Set(from._internal_version(), 
      _this->GetArenaForAllocation());
  }
  // @@protoc_insertion_point(copy_constructor:swemuDriverVersion_call)
}

inline void swemuDriverVersion_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.version_){}
  };
  _impl_.version_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.version_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

swemuDriverVersion_call::~swemuDriverVersion_call() {
  // @@protoc_insertion_point(destructor:swemuDriverVersion_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void swemuDriverVersion_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.version_.Destroy();
}

void swemuDriverVersion_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void swemuDriverVersion_call::Clear() {
// @@protoc_insertion_point(message_clear_start:swemuDriverVersion_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.version_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* swemuDriverVersion_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required string version = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_version();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "swemuDriverVersion_call.version");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* swemuDriverVersion_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:swemuDriverVersion_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string version = 1;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_version().data(), static_cast<int>(this->_internal_version().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "swemuDriverVersion_call.version");
    target = stream->WriteStringMaybeAliased(
        1, this->_internal_version(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:swemuDriverVersion_call)
  return target;
}

size_t swemuDriverVersion_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:swemuDriverVersion_call)
  size_t total_size = 0;

  // required string version = 1;
  if (_internal_has_version()) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
        this->_internal_version());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData swemuDriverVersion_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    swemuDriverVersion_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*swemuDriverVersion_call::GetClassData() const { return &_class_data_; }


void swemuDriverVersion_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<swemuDriverVersion_call*>(&to_msg);
  auto& from = static_cast<const swemuDriverVersion_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:swemuDriverVersion_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_version()) {
    _this->_internal_set_version(from._internal_version());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void swemuDriverVersion_call::CopyFrom(const swemuDriverVersion_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:swemuDriverVersion_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool swemuDriverVersion_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void swemuDriverVersion_call::InternalSwap(swemuDriverVersion_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.version_, lhs_arena,
      &other->_impl_.version_, rhs_arena
  );
}

::PROTOBUF_NAMESPACE_ID::Metadata swemuDriverVersion_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[63]);
}

// ===================================================================

class swemuDriverVersion_response::_Internal {
 public:
  using HasBits = decltype(std::declval<swemuDriverVersion_response>()._impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

swemuDriverVersion_response::swemuDriverVersion_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:swemuDriverVersion_response)
}
swemuDriverVersion_response::swemuDriverVersion_response(const swemuDriverVersion_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  swemuDriverVersion_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.success_ = from._impl_.success_;
  // @@protoc_insertion_point(copy_constructor:swemuDriverVersion_response)
}

inline void swemuDriverVersion_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.success_){false}
  };
}

swemuDriverVersion_response::~swemuDriverVersion_response() {
  // @@protoc_insertion_point(destructor:swemuDriverVersion_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void swemuDriverVersion_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void swemuDriverVersion_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void swemuDriverVersion_response::Clear() {
// @@protoc_insertion_point(message_clear_start:swemuDriverVersion_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* swemuDriverVersion_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool success = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_success(&has_bits);
          _impl_.success_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* swemuDriverVersion_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:swemuDriverVersion_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:swemuDriverVersion_response)
  return target;
}

size_t swemuDriverVersion_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:swemuDriverVersion_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + 1;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData swemuDriverVersion_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    swemuDriverVersion_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*swemuDriverVersion_response::GetClassData() const { return &_class_data_; }


void swemuDriverVersion_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<swemuDriverVersion_response*>(&to_msg);
  auto& from = static_cast<const swemuDriverVersion_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:swemuDriverVersion_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_success()) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void swemuDriverVersion_response::CopyFrom(const swemuDriverVersion_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:swemuDriverVersion_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool swemuDriverVersion_response::IsInitialized() const {
  return true;
}

void swemuDriverVersion_response::InternalSwap(swemuDriverVersion_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.success_, other->_impl_.success_);
}

::PROTOBUF_NAMESPACE_ID::Metadata swemuDriverVersion_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[64]);
}

// ===================================================================

class xclPerfMonReadCounters_Streaming_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_Streaming_call>()._impl_._has_bits_);
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_Streaming_call)
}
xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(const xclPerfMonReadCounters_Streaming_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadCounters_Streaming_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_slotname()) {
    _this->_impl_.slotname_.Set(from._internal_slotname(), 
      _this->GetArenaForAllocation());
  }
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_Streaming_call)
}

inline void xclPerfMonReadCounters_Streaming_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
  };
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPerfMonReadCounters_Streaming_call::~xclPerfMonReadCounters_Streaming_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_Streaming_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadCounters_Streaming_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.slotname_.Destroy();
}

void xclPerfMonReadCounters_Streaming_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadCounters_Streaming_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_Streaming_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_Streaming_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required string slotname = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 10)) {
          auto str = _internal_mutable_slotname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclPerfMonReadCounters_Streaming_call.slotname");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadCounters_Streaming_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_Streaming_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string slotname = 1;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_slotname().data(), static_cast<int>(this->_internal_slotname().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclPerfMonReadCounters_Streaming_call.slotname");
    target = stream->WriteStringMaybeAliased(
        1, this->_internal_slotname(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_Streaming_call)
  return target;
}

size_t xclPerfMonReadCounters_Streaming_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_Streaming_call)
  size_t total_size = 0;

  // required string slotname = 1;
  if (_internal_has_slotname()) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
        this->_internal_slotname());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadCounters_Streaming_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadCounters_Streaming_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadCounters_Streaming_call::GetClassData() const { return &_class_data_; }


void xclPerfMonReadCounters_Streaming_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_Streaming_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_Streaming_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_Streaming_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_slotname()) {
    _this->_internal_set_slotname(from._internal_slotname());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_Streaming_call::CopyFrom(const xclPerfMonReadCounters_Streaming_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_Streaming_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadCounters_Streaming_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclPerfMonReadCounters_Streaming_call::InternalSwap(xclPerfMonReadCounters_Streaming_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.slotname_, lhs_arena,
      &other->_impl_.slotname_, rhs_arena
  );
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadCounters_Streaming_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[65]);
}

// ===================================================================

class xclPerfMonReadCounters_Streaming_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_Streaming_response>()._impl_._has_bits_);
  static void set_has_str_num_tranx(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_str_data_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_str_busy_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_str_stall_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_str_starve_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_Streaming_response)
}
xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(const xclPerfMonReadCounters_Streaming_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadCounters_Streaming_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.str_num_tranx_){}
    , decltype(_impl_.str_data_bytes_){}
    , decltype(_impl_.str_busy_cycles_){}
    , decltype(_impl_.str_stall_cycles_){}
    , decltype(_impl_.str_starve_cycles_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.str_num_tranx_, &from._impl_.str_num_tranx_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.str_starve_cycles_) -
    reinterpret_cast<char*>(&_impl_.str_num_tranx_)) + sizeof(_impl_.str_starve_cycles_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_Streaming_response)
}

inline void xclPerfMonReadCounters_Streaming_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.str_num_tranx_){uint64_t{0u}}
    , decltype(_impl_.str_data_bytes_){uint64_t{0u}}
    , decltype(_impl_.str_busy_cycles_){uint64_t{0u}}
    , decltype(_impl_.str_stall_cycles_){uint64_t{0u}}
    , decltype(_impl_.str_starve_cycles_){uint64_t{0u}}
  };
}

xclPerfMonReadCounters_Streaming_response::~xclPerfMonReadCounters_Streaming_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_Streaming_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadCounters_Streaming_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPerfMonReadCounters_Streaming_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadCounters_Streaming_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_Streaming_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    ::memset(&_impl_.str_num_tranx_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.str_starve_cycles_) -
        reinterpret_cast<char*>(&_impl_.str_num_tranx_)) + sizeof(_impl_.str_starve_cycles_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_Streaming_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint64 str_num_tranx = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_str_num_tranx(&has_bits);
          _impl_.str_num_tranx_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 str_data_bytes = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 64)) {
          _Internal::set_has_str_data_bytes(&has_bits);
          _impl_.str_data_bytes_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 str_busy_cycles = 9;
      case 9:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 72)) {
          _Internal::set_has_str_busy_cycles(&has_bits);
          _impl_.str_busy_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 str_stall_cycles = 10;
      case 10:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 80)) {
          _Internal::set_has_str_stall_cycles(&has_bits);
          _impl_.str_stall_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 str_starve_cycles = 11;
      case 11:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 88)) {
          _Internal::set_has_str_starve_cycles(&has_bits);
          _impl_.str_starve_cycles_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadCounters_Streaming_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_Streaming_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 str_num_tranx = 7;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(7, this->_internal_str_num_tranx(), target);
  }

  // optional uint64 str_data_bytes = 8;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(8, this->_internal_str_data_bytes(), target);
  }

  // optional uint64 str_busy_cycles = 9;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(9, this->_internal_str_busy_cycles(), target);
  }

  // optional uint64 str_stall_cycles = 10;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(10, this->_internal_str_stall_cycles(), target);
  }

  // optional uint64 str_starve_cycles = 11;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(11, this->_internal_str_starve_cycles(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_Streaming_response)
  return target;
}

size_t xclPerfMonReadCounters_Streaming_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_Streaming_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional uint64 str_num_tranx = 7;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_str_num_tranx());
    }

    // optional uint64 str_data_bytes = 8;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_str_data_bytes());
    }

    // optional uint64 str_busy_cycles = 9;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_str_busy_cycles());
    }

    // optional uint64 str_stall_cycles = 10;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_str_stall_cycles());
    }

    // optional uint64 str_starve_cycles = 11;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_str_starve_cycles());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadCounters_Streaming_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadCounters_Streaming_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadCounters_Streaming_response::GetClassData() const { return &_class_data_; }


void xclPerfMonReadCounters_Streaming_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_Streaming_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_Streaming_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_Streaming_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.str_num_tranx_ = from._impl_.str_num_tranx_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.str_data_bytes_ = from._impl_.str_data_bytes_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.str_busy_cycles_ = from._impl_.str_busy_cycles_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.str_stall_cycles_ = from._impl_.str_stall_cycles_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.str_starve_cycles_ = from._impl_.str_starve_cycles_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_Streaming_response::CopyFrom(const xclPerfMonReadCounters_Streaming_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_Streaming_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadCounters_Streaming_response::IsInitialized() const {
  return true;
}

void xclPerfMonReadCounters_Streaming_response::InternalSwap(xclPerfMonReadCounters_Streaming_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_)
      + sizeof(xclPerfMonReadCounters_Streaming_response::_impl_.str_starve_cycles_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_)>(
          reinterpret_cast<char*>(&_impl_.str_num_tranx_),
          reinterpret_cast<char*>(&other->_impl_.str_num_tranx_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadCounters_Streaming_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[66]);
}

// ===================================================================

class xclPerfMonReadTrace_Streaming_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_Streaming_call>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_call)
}
xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(const xclPerfMonReadTrace_Streaming_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_Streaming_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_slotname()) {
    _this->_impl_.slotname_.Set(from._internal_slotname(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_Streaming_call)
}

inline void xclPerfMonReadTrace_Streaming_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.slotname_){}
    , decltype(_impl_.ack_){false}
  };
  _impl_.slotname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.slotname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclPerfMonReadTrace_Streaming_call::~xclPerfMonReadTrace_Streaming_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_Streaming_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.slotname_.Destroy();
}

void xclPerfMonReadTrace_Streaming_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_Streaming_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional string slotname = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_slotname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
          #ifndef NDEBUG
          ::_pbi::VerifyUTF8(str, "xclPerfMonReadTrace_Streaming_call.slotname");
          #endif  // !NDEBUG
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_Streaming_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::VerifyUTF8StringNamedField(
      this->_internal_slotname().data(), static_cast<int>(this->_internal_slotname().length()),
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormat::SERIALIZE,
      "xclPerfMonReadTrace_Streaming_call.slotname");
    target = stream->WriteStringMaybeAliased(
        2, this->_internal_slotname(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_call)
  return target;
}

size_t xclPerfMonReadTrace_Streaming_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_call)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 +
        ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::StringSize(
          this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + 1;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_Streaming_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_Streaming_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_Streaming_call::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_Streaming_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_call::CopyFrom(const xclPerfMonReadTrace_Streaming_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_Streaming_call::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_Streaming_call::InternalSwap(xclPerfMonReadTrace_Streaming_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.slotname_, lhs_arena,
      &other->_impl_.slotname_, rhs_arena
  );
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_Streaming_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[67]);
}

// ===================================================================

class xclPerfMonReadTrace_Streaming_response_events::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_Streaming_response_events>()._impl_._has_bits_);
  static void set_has_eventflags(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_host_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_response.events)
}
xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(const xclPerfMonReadTrace_Streaming_response_events& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_Streaming_response_events* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.eventflags_){}
    , decltype(_impl_.timestamp_){}
    , decltype(_impl_.host_timestamp_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.eventflags_, &from._impl_.eventflags_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.host_timestamp_) -
    reinterpret_cast<char*>(&_impl_.eventflags_)) + sizeof(_impl_.host_timestamp_));
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_Streaming_response.events)
}

inline void xclPerfMonReadTrace_Streaming_response_events::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.eventflags_){0u}
    , decltype(_impl_.timestamp_){0u}
    , decltype(_impl_.host_timestamp_){uint64_t{0u}}
  };
}

xclPerfMonReadTrace_Streaming_response_events::~xclPerfMonReadTrace_Streaming_response_events() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_response.events)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_Streaming_response_events::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclPerfMonReadTrace_Streaming_response_events::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_Streaming_response_events::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_response.events)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    ::memset(&_impl_.eventflags_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.host_timestamp_) -
        reinterpret_cast<char*>(&_impl_.eventflags_)) + sizeof(_impl_.host_timestamp_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_response_events::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // optional uint32 eventflags = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_eventflags(&has_bits);
          _impl_.eventflags_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint32 timestamp = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_timestamp(&has_bits);
          _impl_.timestamp_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // optional uint64 host_timestamp = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_host_timestamp(&has_bits);
          _impl_.host_timestamp_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_Streaming_response_events::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_response.events)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 eventflags = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_eventflags(), target);
  }

  // optional uint32 timestamp = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(4, this->_internal_timestamp(), target);
  }

  // optional uint64 host_timestamp = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_host_timestamp(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_response.events)
  return target;
}

size_t xclPerfMonReadTrace_Streaming_response_events::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_response.events)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional uint32 eventflags = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_eventflags());
    }

    // optional uint32 timestamp = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_timestamp());
    }

    // optional uint64 host_timestamp = 5;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_host_timestamp());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_Streaming_response_events::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_Streaming_response_events::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_Streaming_response_events::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_Streaming_response_events::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_response_events*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_response_events&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_response.events)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.eventflags_ = from._impl_.eventflags_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.timestamp_ = from._impl_.timestamp_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.host_timestamp_ = from._impl_.host_timestamp_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_response_events::CopyFrom(const xclPerfMonReadTrace_Streaming_response_events& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_response.events)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_Streaming_response_events::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_Streaming_response_events::InternalSwap(xclPerfMonReadTrace_Streaming_response_events* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_)
      + sizeof(xclPerfMonReadTrace_Streaming_response_events::_impl_.host_timestamp_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_)>(
          reinterpret_cast<char*>(&_impl_.eventflags_),
          reinterpret_cast<char*>(&other->_impl_.eventflags_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_Streaming_response_events::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[68]);
}

// ===================================================================

class xclPerfMonReadTrace_Streaming_response::_Internal {
 public:
};

xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_response)
}
xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(const xclPerfMonReadTrace_Streaming_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclPerfMonReadTrace_Streaming_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_.output_data_){from._impl_.output_data_}
    , /*decltype(_impl_._cached_size_)*/{}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_Streaming_response)
}

inline void xclPerfMonReadTrace_Streaming_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_.output_data_){arena}
    , /*decltype(_impl_._cached_size_)*/{}
  };
}

xclPerfMonReadTrace_Streaming_response::~xclPerfMonReadTrace_Streaming_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclPerfMonReadTrace_Streaming_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.output_data_.~RepeatedPtrField();
}

void xclPerfMonReadTrace_Streaming_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclPerfMonReadTrace_Streaming_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.output_data_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 66)) {
          ptr -= 1;
          do {
            ptr += 1;
            ptr = ctx->ParseMessage(_internal_add_output_data(), ptr);
            CHK_(ptr);
            if (!ctx->DataAvailable(ptr)) break;
          } while (::PROTOBUF_NAMESPACE_ID::internal::ExpectTag<66>(ptr));
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclPerfMonReadTrace_Streaming_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_output_data_size()); i < n; i++) {
    const auto& repfield = this->_internal_output_data(i);
    target = ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_response)
  return target;
}

size_t xclPerfMonReadTrace_Streaming_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_response)
  size_t total_size = 0;

  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
  total_size += 1UL * this->_internal_output_data_size();
  for (const auto& msg : this->_impl_.output_data_) {
    total_size +=
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::MessageSize(msg);
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclPerfMonReadTrace_Streaming_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclPerfMonReadTrace_Streaming_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclPerfMonReadTrace_Streaming_response::GetClassData() const { return &_class_data_; }


void xclPerfMonReadTrace_Streaming_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_impl_.output_data_.MergeFrom(from._impl_.output_data_);
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_response::CopyFrom(const xclPerfMonReadTrace_Streaming_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclPerfMonReadTrace_Streaming_response::IsInitialized() const {
  return true;
}

void xclPerfMonReadTrace_Streaming_response::InternalSwap(xclPerfMonReadTrace_Streaming_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.output_data_.InternalSwap(&other->_impl_.output_data_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclPerfMonReadTrace_Streaming_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[69]);
}

// ===================================================================

class xclSlaveReadReq_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveReadReq_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclSlaveReadReq_call::xclSlaveReadReq_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSlaveReadReq_call)
}
xclSlaveReadReq_call::xclSlaveReadReq_call(const xclSlaveReadReq_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSlaveReadReq_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.size_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  // @@protoc_insertion_point(copy_constructor:xclSlaveReadReq_call)
}

inline void xclSlaveReadReq_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
  };
}

xclSlaveReadReq_call::~xclSlaveReadReq_call() {
  // @@protoc_insertion_point(destructor:xclSlaveReadReq_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSlaveReadReq_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclSlaveReadReq_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSlaveReadReq_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveReadReq_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSlaveReadReq_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSlaveReadReq_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveReadReq_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_addr(), target);
  }

  // required uint32 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveReadReq_call)
  return target;
}

size_t xclSlaveReadReq_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclSlaveReadReq_call)
  size_t total_size = 0;

  if (_internal_has_addr()) {
    // required uint64 addr = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclSlaveReadReq_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveReadReq_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required uint64 addr = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSlaveReadReq_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSlaveReadReq_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSlaveReadReq_call::GetClassData() const { return &_class_data_; }


void xclSlaveReadReq_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveReadReq_call*>(&to_msg);
  auto& from = static_cast<const xclSlaveReadReq_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveReadReq_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveReadReq_call::CopyFrom(const xclSlaveReadReq_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveReadReq_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSlaveReadReq_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSlaveReadReq_call::InternalSwap(xclSlaveReadReq_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.size_)
      + sizeof(xclSlaveReadReq_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSlaveReadReq_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[70]);
}

// ===================================================================

class xclSlaveReadReq_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveReadReq_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclSlaveReadReq_response::xclSlaveReadReq_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSlaveReadReq_response)
}
xclSlaveReadReq_response::xclSlaveReadReq_response(const xclSlaveReadReq_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSlaveReadReq_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclSlaveReadReq_response)
}

inline void xclSlaveReadReq_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.valid_){false}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclSlaveReadReq_response::~xclSlaveReadReq_response() {
  // @@protoc_insertion_point(destructor:xclSlaveReadReq_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSlaveReadReq_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclSlaveReadReq_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSlaveReadReq_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveReadReq_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSlaveReadReq_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes data = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSlaveReadReq_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveReadReq_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  // required bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_data(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveReadReq_response)
  return target;
}

size_t xclSlaveReadReq_response::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclSlaveReadReq_response)
  size_t total_size = 0;

  if (_internal_has_data()) {
    // required bytes data = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (_internal_has_valid()) {
    // required bool valid = 1;
    total_size += 1 + 1;
  }

  return total_size;
}
size_t xclSlaveReadReq_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveReadReq_response)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required bytes data = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());

    // required bool valid = 1;
    total_size += 1 + 1;

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSlaveReadReq_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSlaveReadReq_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSlaveReadReq_response::GetClassData() const { return &_class_data_; }


void xclSlaveReadReq_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveReadReq_response*>(&to_msg);
  auto& from = static_cast<const xclSlaveReadReq_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveReadReq_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveReadReq_response::CopyFrom(const xclSlaveReadReq_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveReadReq_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSlaveReadReq_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSlaveReadReq_response::InternalSwap(xclSlaveReadReq_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSlaveReadReq_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[71]);
}

// ===================================================================

class xclSlaveWriteReq_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveWriteReq_call>()._impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclSlaveWriteReq_call::xclSlaveWriteReq_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSlaveWriteReq_call)
}
xclSlaveWriteReq_call::xclSlaveWriteReq_call(const xclSlaveWriteReq_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSlaveWriteReq_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_data()) {
    _this->_impl_.data_.Set(from._internal_data(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.addr_, &from._impl_.addr_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.size_) -
    reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  // @@protoc_insertion_point(copy_constructor:xclSlaveWriteReq_call)
}

inline void xclSlaveWriteReq_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.data_){}
    , decltype(_impl_.addr_){uint64_t{0u}}
    , decltype(_impl_.size_){0u}
  };
  _impl_.data_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.data_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclSlaveWriteReq_call::~xclSlaveWriteReq_call() {
  // @@protoc_insertion_point(destructor:xclSlaveWriteReq_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSlaveWriteReq_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.data_.Destroy();
}

void xclSlaveWriteReq_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSlaveWriteReq_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveWriteReq_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.addr_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSlaveWriteReq_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 addr = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_addr(&has_bits);
          _impl_.addr_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 size = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes data = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_data();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSlaveWriteReq_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveWriteReq_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_addr(), target);
  }

  // required uint32 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_size(), target);
  }

  // required bytes data = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_data(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveWriteReq_call)
  return target;
}

size_t xclSlaveWriteReq_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclSlaveWriteReq_call)
  size_t total_size = 0;

  if (_internal_has_data()) {
    // required bytes data = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());
  }

  if (_internal_has_addr()) {
    // required uint64 addr = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());
  }

  if (_internal_has_size()) {
    // required uint32 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclSlaveWriteReq_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveWriteReq_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000007) ^ 0x00000007) == 0) {  // All required fields are present.
    // required bytes data = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_data());

    // required uint64 addr = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_addr());

    // required uint32 size = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSlaveWriteReq_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSlaveWriteReq_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSlaveWriteReq_call::GetClassData() const { return &_class_data_; }


void xclSlaveWriteReq_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveWriteReq_call*>(&to_msg);
  auto& from = static_cast<const xclSlaveWriteReq_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveWriteReq_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveWriteReq_call::CopyFrom(const xclSlaveWriteReq_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveWriteReq_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSlaveWriteReq_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSlaveWriteReq_call::InternalSwap(xclSlaveWriteReq_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.data_, lhs_arena,
      &other->_impl_.data_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.size_)
      + sizeof(xclSlaveWriteReq_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSlaveWriteReq_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[72]);
}

// ===================================================================

class xclSlaveWriteReq_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveWriteReq_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclSlaveWriteReq_response::xclSlaveWriteReq_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSlaveWriteReq_response)
}
xclSlaveWriteReq_response::xclSlaveWriteReq_response(const xclSlaveWriteReq_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSlaveWriteReq_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclSlaveWriteReq_response)
}

inline void xclSlaveWriteReq_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclSlaveWriteReq_response::~xclSlaveWriteReq_response() {
  // @@protoc_insertion_point(destructor:xclSlaveWriteReq_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSlaveWriteReq_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclSlaveWriteReq_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSlaveWriteReq_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveWriteReq_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSlaveWriteReq_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSlaveWriteReq_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveWriteReq_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveWriteReq_response)
  return target;
}

size_t xclSlaveWriteReq_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveWriteReq_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSlaveWriteReq_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSlaveWriteReq_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSlaveWriteReq_response::GetClassData() const { return &_class_data_; }


void xclSlaveWriteReq_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveWriteReq_response*>(&to_msg);
  auto& from = static_cast<const xclSlaveWriteReq_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveWriteReq_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveWriteReq_response::CopyFrom(const xclSlaveWriteReq_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveWriteReq_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSlaveWriteReq_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSlaveWriteReq_response::InternalSwap(xclSlaveWriteReq_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSlaveWriteReq_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[73]);
}

// ===================================================================

class xclInterruptOccured_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclInterruptOccured_call>()._impl_._has_bits_);
  static void set_has_interrupt_line(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclInterruptOccured_call::xclInterruptOccured_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclInterruptOccured_call)
}
xclInterruptOccured_call::xclInterruptOccured_call(const xclInterruptOccured_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclInterruptOccured_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.interrupt_line_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.interrupt_line_ = from._impl_.interrupt_line_;
  // @@protoc_insertion_point(copy_constructor:xclInterruptOccured_call)
}

inline void xclInterruptOccured_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.interrupt_line_){0u}
  };
}

xclInterruptOccured_call::~xclInterruptOccured_call() {
  // @@protoc_insertion_point(destructor:xclInterruptOccured_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclInterruptOccured_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclInterruptOccured_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclInterruptOccured_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclInterruptOccured_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.interrupt_line_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclInterruptOccured_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 interrupt_line = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_interrupt_line(&has_bits);
          _impl_.interrupt_line_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclInterruptOccured_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclInterruptOccured_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 interrupt_line = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_interrupt_line(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclInterruptOccured_call)
  return target;
}

size_t xclInterruptOccured_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclInterruptOccured_call)
  size_t total_size = 0;

  // required uint32 interrupt_line = 1;
  if (_internal_has_interrupt_line()) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_interrupt_line());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclInterruptOccured_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclInterruptOccured_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclInterruptOccured_call::GetClassData() const { return &_class_data_; }


void xclInterruptOccured_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclInterruptOccured_call*>(&to_msg);
  auto& from = static_cast<const xclInterruptOccured_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclInterruptOccured_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_interrupt_line()) {
    _this->_internal_set_interrupt_line(from._internal_interrupt_line());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclInterruptOccured_call::CopyFrom(const xclInterruptOccured_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclInterruptOccured_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclInterruptOccured_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclInterruptOccured_call::InternalSwap(xclInterruptOccured_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.interrupt_line_, other->_impl_.interrupt_line_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclInterruptOccured_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[74]);
}

// ===================================================================

class xclInterruptOccured_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclInterruptOccured_response>()._impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclInterruptOccured_response::xclInterruptOccured_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclInterruptOccured_response)
}
xclInterruptOccured_response::xclInterruptOccured_response(const xclInterruptOccured_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclInterruptOccured_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.valid_ = from._impl_.valid_;
  // @@protoc_insertion_point(copy_constructor:xclInterruptOccured_response)
}

inline void xclInterruptOccured_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.valid_){false}
  };
}

xclInterruptOccured_response::~xclInterruptOccured_response() {
  // @@protoc_insertion_point(destructor:xclInterruptOccured_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclInterruptOccured_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclInterruptOccured_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclInterruptOccured_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclInterruptOccured_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclInterruptOccured_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool valid = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_valid(&has_bits);
          _impl_.valid_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclInterruptOccured_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclInterruptOccured_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclInterruptOccured_response)
  return target;
}

size_t xclInterruptOccured_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclInterruptOccured_response)
  size_t total_size = 0;

  // required bool valid = 1;
  if (_internal_has_valid()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclInterruptOccured_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclInterruptOccured_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclInterruptOccured_response::GetClassData() const { return &_class_data_; }


void xclInterruptOccured_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclInterruptOccured_response*>(&to_msg);
  auto& from = static_cast<const xclInterruptOccured_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclInterruptOccured_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_valid()) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclInterruptOccured_response::CopyFrom(const xclInterruptOccured_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclInterruptOccured_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclInterruptOccured_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclInterruptOccured_response::InternalSwap(xclInterruptOccured_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.valid_, other->_impl_.valid_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclInterruptOccured_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[75]);
}

// ===================================================================

class xclGraphInit_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphInit_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_graphname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphInit_call::xclGraphInit_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphInit_call)
}
xclGraphInit_call::xclGraphInit_call(const xclGraphInit_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphInit_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.graphname_){}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.graphname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.graphname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_graphname()) {
    _this->_impl_.graphname_.Set(from._internal_graphname(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.gh_ = from._impl_.gh_;
  // @@protoc_insertion_point(copy_constructor:xclGraphInit_call)
}

inline void xclGraphInit_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.graphname_){}
    , decltype(_impl_.gh_){0u}
  };
  _impl_.graphname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.graphname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGraphInit_call::~xclGraphInit_call() {
  // @@protoc_insertion_point(destructor:xclGraphInit_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphInit_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.graphname_.Destroy();
}

void xclGraphInit_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphInit_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphInit_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.graphname_.ClearNonDefaultToEmpty();
  }
  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphInit_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes graphname = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_graphname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphInit_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphInit_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required bytes graphname = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_graphname(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphInit_call)
  return target;
}

size_t xclGraphInit_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphInit_call)
  size_t total_size = 0;

  if (_internal_has_graphname()) {
    // required bytes graphname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_graphname());
  }

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  return total_size;
}
size_t xclGraphInit_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphInit_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required bytes graphname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_graphname());

    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphInit_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphInit_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphInit_call::GetClassData() const { return &_class_data_; }


void xclGraphInit_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphInit_call*>(&to_msg);
  auto& from = static_cast<const xclGraphInit_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphInit_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_graphname(from._internal_graphname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphInit_call::CopyFrom(const xclGraphInit_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphInit_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphInit_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphInit_call::InternalSwap(xclGraphInit_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.graphname_, lhs_arena,
      &other->_impl_.graphname_, rhs_arena
  );
  swap(_impl_.gh_, other->_impl_.gh_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphInit_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[76]);
}

// ===================================================================

class xclGraphInit_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphInit_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphInit_response::xclGraphInit_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphInit_response)
}
xclGraphInit_response::xclGraphInit_response(const xclGraphInit_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphInit_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphInit_response)
}

inline void xclGraphInit_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphInit_response::~xclGraphInit_response() {
  // @@protoc_insertion_point(destructor:xclGraphInit_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphInit_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphInit_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphInit_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphInit_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphInit_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphInit_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphInit_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphInit_response)
  return target;
}

size_t xclGraphInit_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphInit_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphInit_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphInit_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphInit_response::GetClassData() const { return &_class_data_; }


void xclGraphInit_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphInit_response*>(&to_msg);
  auto& from = static_cast<const xclGraphInit_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphInit_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphInit_response::CopyFrom(const xclGraphInit_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphInit_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphInit_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphInit_response::InternalSwap(xclGraphInit_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphInit_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[77]);
}

// ===================================================================

class xclGraphRun_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphRun_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_iterations(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphRun_call::xclGraphRun_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphRun_call)
}
xclGraphRun_call::xclGraphRun_call(const xclGraphRun_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphRun_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){}
    , decltype(_impl_.iterations_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.gh_, &from._impl_.gh_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.iterations_) -
    reinterpret_cast<char*>(&_impl_.gh_)) + sizeof(_impl_.iterations_));
  // @@protoc_insertion_point(copy_constructor:xclGraphRun_call)
}

inline void xclGraphRun_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){0u}
    , decltype(_impl_.iterations_){0u}
  };
}

xclGraphRun_call::~xclGraphRun_call() {
  // @@protoc_insertion_point(destructor:xclGraphRun_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphRun_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphRun_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphRun_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphRun_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.gh_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.iterations_) -
        reinterpret_cast<char*>(&_impl_.gh_)) + sizeof(_impl_.iterations_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphRun_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 iterations = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_iterations(&has_bits);
          _impl_.iterations_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphRun_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphRun_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required uint32 iterations = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_iterations(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphRun_call)
  return target;
}

size_t xclGraphRun_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphRun_call)
  size_t total_size = 0;

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  if (_internal_has_iterations()) {
    // required uint32 iterations = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_iterations());
  }

  return total_size;
}
size_t xclGraphRun_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphRun_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

    // required uint32 iterations = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_iterations());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphRun_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphRun_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphRun_call::GetClassData() const { return &_class_data_; }


void xclGraphRun_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphRun_call*>(&to_msg);
  auto& from = static_cast<const xclGraphRun_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphRun_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.iterations_ = from._impl_.iterations_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphRun_call::CopyFrom(const xclGraphRun_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphRun_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphRun_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphRun_call::InternalSwap(xclGraphRun_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.iterations_)
      + sizeof(xclGraphRun_call::_impl_.iterations_)
      - PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.gh_)>(
          reinterpret_cast<char*>(&_impl_.gh_),
          reinterpret_cast<char*>(&other->_impl_.gh_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphRun_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[78]);
}

// ===================================================================

class xclGraphRun_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphRun_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphRun_response::xclGraphRun_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphRun_response)
}
xclGraphRun_response::xclGraphRun_response(const xclGraphRun_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphRun_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphRun_response)
}

inline void xclGraphRun_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphRun_response::~xclGraphRun_response() {
  // @@protoc_insertion_point(destructor:xclGraphRun_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphRun_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphRun_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphRun_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphRun_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphRun_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphRun_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphRun_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphRun_response)
  return target;
}

size_t xclGraphRun_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphRun_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphRun_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphRun_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphRun_response::GetClassData() const { return &_class_data_; }


void xclGraphRun_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphRun_response*>(&to_msg);
  auto& from = static_cast<const xclGraphRun_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphRun_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphRun_response::CopyFrom(const xclGraphRun_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphRun_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphRun_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphRun_response::InternalSwap(xclGraphRun_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphRun_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[79]);
}

// ===================================================================

class xclGraphWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphWait_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphWait_call::xclGraphWait_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphWait_call)
}
xclGraphWait_call::xclGraphWait_call(const xclGraphWait_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphWait_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.gh_ = from._impl_.gh_;
  // @@protoc_insertion_point(copy_constructor:xclGraphWait_call)
}

inline void xclGraphWait_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){0u}
  };
}

xclGraphWait_call::~xclGraphWait_call() {
  // @@protoc_insertion_point(destructor:xclGraphWait_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphWait_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphWait_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphWait_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphWait_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphWait_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphWait_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphWait_call)
  return target;
}

size_t xclGraphWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphWait_call)
  size_t total_size = 0;

  // required uint32 gh = 2;
  if (_internal_has_gh()) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphWait_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphWait_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphWait_call::GetClassData() const { return &_class_data_; }


void xclGraphWait_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphWait_call*>(&to_msg);
  auto& from = static_cast<const xclGraphWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphWait_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_gh()) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphWait_call::CopyFrom(const xclGraphWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphWait_call::InternalSwap(xclGraphWait_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.gh_, other->_impl_.gh_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[80]);
}

// ===================================================================

class xclGraphWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphWait_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphWait_response::xclGraphWait_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphWait_response)
}
xclGraphWait_response::xclGraphWait_response(const xclGraphWait_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphWait_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphWait_response)
}

inline void xclGraphWait_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphWait_response::~xclGraphWait_response() {
  // @@protoc_insertion_point(destructor:xclGraphWait_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphWait_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphWait_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphWait_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphWait_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphWait_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphWait_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphWait_response)
  return target;
}

size_t xclGraphWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphWait_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphWait_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphWait_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphWait_response::GetClassData() const { return &_class_data_; }


void xclGraphWait_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphWait_response*>(&to_msg);
  auto& from = static_cast<const xclGraphWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphWait_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphWait_response::CopyFrom(const xclGraphWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphWait_response::InternalSwap(xclGraphWait_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[81]);
}

// ===================================================================

class xclGraphEnd_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphEnd_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphEnd_call::xclGraphEnd_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphEnd_call)
}
xclGraphEnd_call::xclGraphEnd_call(const xclGraphEnd_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphEnd_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.gh_ = from._impl_.gh_;
  // @@protoc_insertion_point(copy_constructor:xclGraphEnd_call)
}

inline void xclGraphEnd_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){0u}
  };
}

xclGraphEnd_call::~xclGraphEnd_call() {
  // @@protoc_insertion_point(destructor:xclGraphEnd_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphEnd_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphEnd_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphEnd_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphEnd_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphEnd_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphEnd_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphEnd_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphEnd_call)
  return target;
}

size_t xclGraphEnd_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphEnd_call)
  size_t total_size = 0;

  // required uint32 gh = 2;
  if (_internal_has_gh()) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphEnd_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphEnd_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphEnd_call::GetClassData() const { return &_class_data_; }


void xclGraphEnd_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphEnd_call*>(&to_msg);
  auto& from = static_cast<const xclGraphEnd_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphEnd_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_gh()) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphEnd_call::CopyFrom(const xclGraphEnd_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphEnd_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphEnd_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphEnd_call::InternalSwap(xclGraphEnd_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.gh_, other->_impl_.gh_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphEnd_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[82]);
}

// ===================================================================

class xclGraphEnd_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphEnd_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphEnd_response::xclGraphEnd_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphEnd_response)
}
xclGraphEnd_response::xclGraphEnd_response(const xclGraphEnd_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphEnd_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphEnd_response)
}

inline void xclGraphEnd_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){0u}
  };
}

xclGraphEnd_response::~xclGraphEnd_response() {
  // @@protoc_insertion_point(destructor:xclGraphEnd_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphEnd_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphEnd_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphEnd_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphEnd_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphEnd_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphEnd_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphEnd_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphEnd_response)
  return target;
}

size_t xclGraphEnd_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphEnd_response)
  size_t total_size = 0;

  // required uint32 ack = 1;
  if (_internal_has_ack()) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_ack());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphEnd_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphEnd_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphEnd_response::GetClassData() const { return &_class_data_; }


void xclGraphEnd_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphEnd_response*>(&to_msg);
  auto& from = static_cast<const xclGraphEnd_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphEnd_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphEnd_response::CopyFrom(const xclGraphEnd_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphEnd_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphEnd_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphEnd_response::InternalSwap(xclGraphEnd_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphEnd_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[83]);
}

// ===================================================================

class xclGraphUpdateRTP_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphUpdateRTP_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_portname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000f) ^ 0x0000000f) != 0;
  }
};

xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphUpdateRTP_call)
}
xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(const xclGraphUpdateRTP_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphUpdateRTP_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.portname_){}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.portname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.portname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_portname()) {
    _this->_impl_.portname_.Set(from._internal_portname(), 
      _this->GetArenaForAllocation());
  }
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_buffer()) {
    _this->_impl_.buffer_.Set(from._internal_buffer(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.gh_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  // @@protoc_insertion_point(copy_constructor:xclGraphUpdateRTP_call)
}

inline void xclGraphUpdateRTP_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.portname_){}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.gh_){0u}
  };
  _impl_.portname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.portname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGraphUpdateRTP_call::~xclGraphUpdateRTP_call() {
  // @@protoc_insertion_point(destructor:xclGraphUpdateRTP_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphUpdateRTP_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.portname_.Destroy();
  _impl_.buffer_.Destroy();
}

void xclGraphUpdateRTP_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphUpdateRTP_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphUpdateRTP_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.portname_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.buffer_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphUpdateRTP_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes portname = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_portname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes buffer = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_buffer();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphUpdateRTP_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphUpdateRTP_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required bytes portname = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_portname(), target);
  }

  // required bytes buffer = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_buffer(), target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphUpdateRTP_call)
  return target;
}

size_t xclGraphUpdateRTP_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphUpdateRTP_call)
  size_t total_size = 0;

  if (_internal_has_portname()) {
    // required bytes portname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_portname());
  }

  if (_internal_has_buffer()) {
    // required bytes buffer = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());
  }

  if (_internal_has_size()) {
    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  return total_size;
}
size_t xclGraphUpdateRTP_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphUpdateRTP_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000000f) ^ 0x0000000f) == 0) {  // All required fields are present.
    // required bytes portname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_portname());

    // required bytes buffer = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());

    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphUpdateRTP_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphUpdateRTP_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphUpdateRTP_call::GetClassData() const { return &_class_data_; }


void xclGraphUpdateRTP_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphUpdateRTP_call*>(&to_msg);
  auto& from = static_cast<const xclGraphUpdateRTP_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphUpdateRTP_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_portname(from._internal_portname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphUpdateRTP_call::CopyFrom(const xclGraphUpdateRTP_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphUpdateRTP_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphUpdateRTP_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphUpdateRTP_call::InternalSwap(xclGraphUpdateRTP_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.portname_, lhs_arena,
      &other->_impl_.portname_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.buffer_, lhs_arena,
      &other->_impl_.buffer_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.gh_)
      + sizeof(xclGraphUpdateRTP_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphUpdateRTP_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[84]);
}

// ===================================================================

class xclGraphUpdateRTP_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphUpdateRTP_response>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphUpdateRTP_response)
}
xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(const xclGraphUpdateRTP_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphUpdateRTP_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclGraphUpdateRTP_response)
}

inline void xclGraphUpdateRTP_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
}

xclGraphUpdateRTP_response::~xclGraphUpdateRTP_response() {
  // @@protoc_insertion_point(destructor:xclGraphUpdateRTP_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphUpdateRTP_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphUpdateRTP_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphUpdateRTP_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphUpdateRTP_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphUpdateRTP_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphUpdateRTP_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphUpdateRTP_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphUpdateRTP_response)
  return target;
}

size_t xclGraphUpdateRTP_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphUpdateRTP_response)
  size_t total_size = 0;

  // required uint64 size = 1;
  if (_internal_has_size()) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphUpdateRTP_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphUpdateRTP_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphUpdateRTP_response::GetClassData() const { return &_class_data_; }


void xclGraphUpdateRTP_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphUpdateRTP_response*>(&to_msg);
  auto& from = static_cast<const xclGraphUpdateRTP_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphUpdateRTP_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_size()) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphUpdateRTP_response::CopyFrom(const xclGraphUpdateRTP_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphUpdateRTP_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphUpdateRTP_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphUpdateRTP_response::InternalSwap(xclGraphUpdateRTP_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphUpdateRTP_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[85]);
}

// ===================================================================

class xclGraphReadRTP_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphReadRTP_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_portname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000f) ^ 0x0000000f) != 0;
  }
};

xclGraphReadRTP_call::xclGraphReadRTP_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphReadRTP_call)
}
xclGraphReadRTP_call::xclGraphReadRTP_call(const xclGraphReadRTP_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphReadRTP_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.portname_){}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.portname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.portname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_portname()) {
    _this->_impl_.portname_.Set(from._internal_portname(), 
      _this->GetArenaForAllocation());
  }
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_buffer()) {
    _this->_impl_.buffer_.Set(from._internal_buffer(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.gh_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  // @@protoc_insertion_point(copy_constructor:xclGraphReadRTP_call)
}

inline void xclGraphReadRTP_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.portname_){}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.gh_){0u}
  };
  _impl_.portname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.portname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGraphReadRTP_call::~xclGraphReadRTP_call() {
  // @@protoc_insertion_point(destructor:xclGraphReadRTP_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphReadRTP_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.portname_.Destroy();
  _impl_.buffer_.Destroy();
}

void xclGraphReadRTP_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphReadRTP_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphReadRTP_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.portname_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.buffer_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphReadRTP_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes portname = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 26)) {
          auto str = _internal_mutable_portname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes buffer = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_buffer();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphReadRTP_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphReadRTP_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required bytes portname = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        3, this->_internal_portname(), target);
  }

  // required bytes buffer = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_buffer(), target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphReadRTP_call)
  return target;
}

size_t xclGraphReadRTP_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphReadRTP_call)
  size_t total_size = 0;

  if (_internal_has_portname()) {
    // required bytes portname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_portname());
  }

  if (_internal_has_buffer()) {
    // required bytes buffer = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());
  }

  if (_internal_has_size()) {
    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  return total_size;
}
size_t xclGraphReadRTP_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphReadRTP_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000000f) ^ 0x0000000f) == 0) {  // All required fields are present.
    // required bytes portname = 3;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_portname());

    // required bytes buffer = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());

    // required uint64 size = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphReadRTP_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphReadRTP_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphReadRTP_call::GetClassData() const { return &_class_data_; }


void xclGraphReadRTP_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphReadRTP_call*>(&to_msg);
  auto& from = static_cast<const xclGraphReadRTP_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphReadRTP_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_portname(from._internal_portname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphReadRTP_call::CopyFrom(const xclGraphReadRTP_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphReadRTP_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphReadRTP_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphReadRTP_call::InternalSwap(xclGraphReadRTP_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.portname_, lhs_arena,
      &other->_impl_.portname_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.buffer_, lhs_arena,
      &other->_impl_.buffer_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.gh_)
      + sizeof(xclGraphReadRTP_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphReadRTP_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[86]);
}

// ===================================================================

class xclGraphReadRTP_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphReadRTP_response>()._impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphReadRTP_response::xclGraphReadRTP_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphReadRTP_response)
}
xclGraphReadRTP_response::xclGraphReadRTP_response(const xclGraphReadRTP_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphReadRTP_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_buffer()) {
    _this->_impl_.buffer_.Set(from._internal_buffer(), 
      _this->GetArenaForAllocation());
  }
  _this->_impl_.size_ = from._impl_.size_;
  // @@protoc_insertion_point(copy_constructor:xclGraphReadRTP_response)
}

inline void xclGraphReadRTP_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.buffer_){}
    , decltype(_impl_.size_){uint64_t{0u}}
  };
  _impl_.buffer_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.buffer_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGraphReadRTP_response::~xclGraphReadRTP_response() {
  // @@protoc_insertion_point(destructor:xclGraphReadRTP_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphReadRTP_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.buffer_.Destroy();
}

void xclGraphReadRTP_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphReadRTP_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphReadRTP_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.buffer_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphReadRTP_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint64 size = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes buffer = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_buffer();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphReadRTP_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphReadRTP_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(1, this->_internal_size(), target);
  }

  // required bytes buffer = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_buffer(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphReadRTP_response)
  return target;
}

size_t xclGraphReadRTP_response::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphReadRTP_response)
  size_t total_size = 0;

  if (_internal_has_buffer()) {
    // required bytes buffer = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());
  }

  if (_internal_has_size()) {
    // required uint64 size = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  return total_size;
}
size_t xclGraphReadRTP_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphReadRTP_response)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required bytes buffer = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_buffer());

    // required uint64 size = 1;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphReadRTP_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphReadRTP_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphReadRTP_response::GetClassData() const { return &_class_data_; }


void xclGraphReadRTP_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphReadRTP_response*>(&to_msg);
  auto& from = static_cast<const xclGraphReadRTP_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphReadRTP_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphReadRTP_response::CopyFrom(const xclGraphReadRTP_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphReadRTP_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphReadRTP_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphReadRTP_response::InternalSwap(xclGraphReadRTP_response* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.buffer_, lhs_arena,
      &other->_impl_.buffer_, rhs_arena
  );
  swap(_impl_.size_, other->_impl_.size_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphReadRTP_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[87]);
}

// ===================================================================

class xclSyncBOAIENB_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSyncBOAIENB_call>()._impl_._has_bits_);
  static void set_has_gmioname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dir(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_boh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclSyncBOAIENB_call::xclSyncBOAIENB_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSyncBOAIENB_call)
}
xclSyncBOAIENB_call::xclSyncBOAIENB_call(const xclSyncBOAIENB_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSyncBOAIENB_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gmioname_){}
    , decltype(_impl_.size_){}
    , decltype(_impl_.offset_){}
    , decltype(_impl_.boh_){}
    , decltype(_impl_.dir_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.gmioname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.gmioname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_gmioname()) {
    _this->_impl_.gmioname_.Set(from._internal_gmioname(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.size_, &from._impl_.size_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.dir_) -
    reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.dir_));
  // @@protoc_insertion_point(copy_constructor:xclSyncBOAIENB_call)
}

inline void xclSyncBOAIENB_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gmioname_){}
    , decltype(_impl_.size_){uint64_t{0u}}
    , decltype(_impl_.offset_){uint64_t{0u}}
    , decltype(_impl_.boh_){uint64_t{0u}}
    , decltype(_impl_.dir_){0u}
  };
  _impl_.gmioname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.gmioname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclSyncBOAIENB_call::~xclSyncBOAIENB_call() {
  // @@protoc_insertion_point(destructor:xclSyncBOAIENB_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSyncBOAIENB_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.gmioname_.Destroy();
}

void xclSyncBOAIENB_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSyncBOAIENB_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSyncBOAIENB_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.gmioname_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.size_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.dir_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.dir_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSyncBOAIENB_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes gmioname = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_gmioname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint32 dir = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_dir(&has_bits);
          _impl_.dir_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 size = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 32)) {
          _Internal::set_has_size(&has_bits);
          _impl_.size_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 offset = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_offset(&has_bits);
          _impl_.offset_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 boh = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 48)) {
          _Internal::set_has_boh(&has_bits);
          _impl_.boh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSyncBOAIENB_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSyncBOAIENB_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes gmioname = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_gmioname(), target);
  }

  // required uint32 dir = 3;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(3, this->_internal_dir(), target);
  }

  // required uint64 size = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(4, this->_internal_size(), target);
  }

  // required uint64 offset = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_offset(), target);
  }

  // required uint64 boh = 6;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(6, this->_internal_boh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSyncBOAIENB_call)
  return target;
}

size_t xclSyncBOAIENB_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclSyncBOAIENB_call)
  size_t total_size = 0;

  if (_internal_has_gmioname()) {
    // required bytes gmioname = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_gmioname());
  }

  if (_internal_has_size()) {
    // required uint64 size = 4;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());
  }

  if (_internal_has_offset()) {
    // required uint64 offset = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_offset());
  }

  if (_internal_has_boh()) {
    // required uint64 boh = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_boh());
  }

  if (_internal_has_dir()) {
    // required uint32 dir = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_dir());
  }

  return total_size;
}
size_t xclSyncBOAIENB_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSyncBOAIENB_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000001f) ^ 0x0000001f) == 0) {  // All required fields are present.
    // required bytes gmioname = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_gmioname());

    // required uint64 size = 4;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_size());

    // required uint64 offset = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_offset());

    // required uint64 boh = 6;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_boh());

    // required uint32 dir = 3;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_dir());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSyncBOAIENB_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSyncBOAIENB_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSyncBOAIENB_call::GetClassData() const { return &_class_data_; }


void xclSyncBOAIENB_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSyncBOAIENB_call*>(&to_msg);
  auto& from = static_cast<const xclSyncBOAIENB_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSyncBOAIENB_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_gmioname(from._internal_gmioname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.boh_ = from._impl_.boh_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.dir_ = from._impl_.dir_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSyncBOAIENB_call::CopyFrom(const xclSyncBOAIENB_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSyncBOAIENB_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSyncBOAIENB_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSyncBOAIENB_call::InternalSwap(xclSyncBOAIENB_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.gmioname_, lhs_arena,
      &other->_impl_.gmioname_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.dir_)
      + sizeof(xclSyncBOAIENB_call::_impl_.dir_)
      - PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSyncBOAIENB_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[88]);
}

// ===================================================================

class xclSyncBOAIENB_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSyncBOAIENB_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclSyncBOAIENB_response::xclSyncBOAIENB_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclSyncBOAIENB_response)
}
xclSyncBOAIENB_response::xclSyncBOAIENB_response(const xclSyncBOAIENB_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclSyncBOAIENB_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclSyncBOAIENB_response)
}

inline void xclSyncBOAIENB_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclSyncBOAIENB_response::~xclSyncBOAIENB_response() {
  // @@protoc_insertion_point(destructor:xclSyncBOAIENB_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclSyncBOAIENB_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclSyncBOAIENB_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclSyncBOAIENB_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSyncBOAIENB_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclSyncBOAIENB_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclSyncBOAIENB_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSyncBOAIENB_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSyncBOAIENB_response)
  return target;
}

size_t xclSyncBOAIENB_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSyncBOAIENB_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclSyncBOAIENB_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclSyncBOAIENB_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclSyncBOAIENB_response::GetClassData() const { return &_class_data_; }


void xclSyncBOAIENB_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclSyncBOAIENB_response*>(&to_msg);
  auto& from = static_cast<const xclSyncBOAIENB_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSyncBOAIENB_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclSyncBOAIENB_response::CopyFrom(const xclSyncBOAIENB_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSyncBOAIENB_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclSyncBOAIENB_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclSyncBOAIENB_response::InternalSwap(xclSyncBOAIENB_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclSyncBOAIENB_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[89]);
}

// ===================================================================

class xclGMIOWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGMIOWait_call>()._impl_._has_bits_);
  static void set_has_gmioname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGMIOWait_call::xclGMIOWait_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGMIOWait_call)
}
xclGMIOWait_call::xclGMIOWait_call(const xclGMIOWait_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGMIOWait_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gmioname_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.gmioname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.gmioname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_gmioname()) {
    _this->_impl_.gmioname_.Set(from._internal_gmioname(), 
      _this->GetArenaForAllocation());
  }
  // @@protoc_insertion_point(copy_constructor:xclGMIOWait_call)
}

inline void xclGMIOWait_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gmioname_){}
  };
  _impl_.gmioname_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.gmioname_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclGMIOWait_call::~xclGMIOWait_call() {
  // @@protoc_insertion_point(destructor:xclGMIOWait_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGMIOWait_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.gmioname_.Destroy();
}

void xclGMIOWait_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGMIOWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGMIOWait_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.gmioname_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGMIOWait_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes gmioname = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_gmioname();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGMIOWait_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGMIOWait_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes gmioname = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_gmioname(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGMIOWait_call)
  return target;
}

size_t xclGMIOWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGMIOWait_call)
  size_t total_size = 0;

  // required bytes gmioname = 2;
  if (_internal_has_gmioname()) {
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_gmioname());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGMIOWait_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGMIOWait_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGMIOWait_call::GetClassData() const { return &_class_data_; }


void xclGMIOWait_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGMIOWait_call*>(&to_msg);
  auto& from = static_cast<const xclGMIOWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGMIOWait_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_gmioname()) {
    _this->_internal_set_gmioname(from._internal_gmioname());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGMIOWait_call::CopyFrom(const xclGMIOWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGMIOWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGMIOWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGMIOWait_call::InternalSwap(xclGMIOWait_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.gmioname_, lhs_arena,
      &other->_impl_.gmioname_, rhs_arena
  );
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGMIOWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[90]);
}

// ===================================================================

class xclGMIOWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGMIOWait_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGMIOWait_response::xclGMIOWait_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGMIOWait_response)
}
xclGMIOWait_response::xclGMIOWait_response(const xclGMIOWait_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGMIOWait_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGMIOWait_response)
}

inline void xclGMIOWait_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGMIOWait_response::~xclGMIOWait_response() {
  // @@protoc_insertion_point(destructor:xclGMIOWait_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGMIOWait_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGMIOWait_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGMIOWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGMIOWait_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGMIOWait_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGMIOWait_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGMIOWait_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGMIOWait_response)
  return target;
}

size_t xclGMIOWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGMIOWait_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGMIOWait_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGMIOWait_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGMIOWait_response::GetClassData() const { return &_class_data_; }


void xclGMIOWait_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGMIOWait_response*>(&to_msg);
  auto& from = static_cast<const xclGMIOWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGMIOWait_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGMIOWait_response::CopyFrom(const xclGMIOWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGMIOWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGMIOWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGMIOWait_response::InternalSwap(xclGMIOWait_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGMIOWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[91]);
}

// ===================================================================

class xclGraphTimedWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedWait_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_cycletimeout(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphTimedWait_call::xclGraphTimedWait_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedWait_call)
}
xclGraphTimedWait_call::xclGraphTimedWait_call(const xclGraphTimedWait_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphTimedWait_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.cycletimeout_){}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.cycletimeout_, &from._impl_.cycletimeout_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.gh_) -
    reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  // @@protoc_insertion_point(copy_constructor:xclGraphTimedWait_call)
}

inline void xclGraphTimedWait_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.cycletimeout_){uint64_t{0u}}
    , decltype(_impl_.gh_){0u}
  };
}

xclGraphTimedWait_call::~xclGraphTimedWait_call() {
  // @@protoc_insertion_point(destructor:xclGraphTimedWait_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphTimedWait_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphTimedWait_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphTimedWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedWait_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.cycletimeout_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphTimedWait_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 cycletimeout = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_cycletimeout(&has_bits);
          _impl_.cycletimeout_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphTimedWait_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedWait_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required uint64 cycletimeout = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_cycletimeout(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedWait_call)
  return target;
}

size_t xclGraphTimedWait_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphTimedWait_call)
  size_t total_size = 0;

  if (_internal_has_cycletimeout()) {
    // required uint64 cycletimeout = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_cycletimeout());
  }

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  return total_size;
}
size_t xclGraphTimedWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedWait_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required uint64 cycletimeout = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_cycletimeout());

    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphTimedWait_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphTimedWait_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphTimedWait_call::GetClassData() const { return &_class_data_; }


void xclGraphTimedWait_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedWait_call*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedWait_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.cycletimeout_ = from._impl_.cycletimeout_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedWait_call::CopyFrom(const xclGraphTimedWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphTimedWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphTimedWait_call::InternalSwap(xclGraphTimedWait_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.gh_)
      + sizeof(xclGraphTimedWait_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.cycletimeout_)>(
          reinterpret_cast<char*>(&_impl_.cycletimeout_),
          reinterpret_cast<char*>(&other->_impl_.cycletimeout_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphTimedWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[92]);
}

// ===================================================================

class xclGraphTimedWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedWait_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphTimedWait_response::xclGraphTimedWait_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedWait_response)
}
xclGraphTimedWait_response::xclGraphTimedWait_response(const xclGraphTimedWait_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphTimedWait_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphTimedWait_response)
}

inline void xclGraphTimedWait_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphTimedWait_response::~xclGraphTimedWait_response() {
  // @@protoc_insertion_point(destructor:xclGraphTimedWait_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphTimedWait_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphTimedWait_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphTimedWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedWait_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphTimedWait_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphTimedWait_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedWait_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedWait_response)
  return target;
}

size_t xclGraphTimedWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedWait_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphTimedWait_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphTimedWait_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphTimedWait_response::GetClassData() const { return &_class_data_; }


void xclGraphTimedWait_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedWait_response*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedWait_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedWait_response::CopyFrom(const xclGraphTimedWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphTimedWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphTimedWait_response::InternalSwap(xclGraphTimedWait_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphTimedWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[93]);
}

// ===================================================================

class xclGraphTimedEnd_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedEnd_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_cycletimeout(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphTimedEnd_call::xclGraphTimedEnd_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedEnd_call)
}
xclGraphTimedEnd_call::xclGraphTimedEnd_call(const xclGraphTimedEnd_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphTimedEnd_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.cycletimeout_){}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  ::memcpy(&_impl_.cycletimeout_, &from._impl_.cycletimeout_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.gh_) -
    reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  // @@protoc_insertion_point(copy_constructor:xclGraphTimedEnd_call)
}

inline void xclGraphTimedEnd_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.cycletimeout_){uint64_t{0u}}
    , decltype(_impl_.gh_){0u}
  };
}

xclGraphTimedEnd_call::~xclGraphTimedEnd_call() {
  // @@protoc_insertion_point(destructor:xclGraphTimedEnd_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphTimedEnd_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphTimedEnd_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphTimedEnd_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedEnd_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.cycletimeout_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphTimedEnd_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 cycletimeout = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_cycletimeout(&has_bits);
          _impl_.cycletimeout_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphTimedEnd_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedEnd_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  // required uint64 cycletimeout = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_cycletimeout(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedEnd_call)
  return target;
}

size_t xclGraphTimedEnd_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclGraphTimedEnd_call)
  size_t total_size = 0;

  if (_internal_has_cycletimeout()) {
    // required uint64 cycletimeout = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_cycletimeout());
  }

  if (_internal_has_gh()) {
    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }

  return total_size;
}
size_t xclGraphTimedEnd_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedEnd_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x00000003) ^ 0x00000003) == 0) {  // All required fields are present.
    // required uint64 cycletimeout = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_cycletimeout());

    // required uint32 gh = 2;
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphTimedEnd_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphTimedEnd_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphTimedEnd_call::GetClassData() const { return &_class_data_; }


void xclGraphTimedEnd_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedEnd_call*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedEnd_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedEnd_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.cycletimeout_ = from._impl_.cycletimeout_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedEnd_call::CopyFrom(const xclGraphTimedEnd_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedEnd_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphTimedEnd_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphTimedEnd_call::InternalSwap(xclGraphTimedEnd_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.gh_)
      + sizeof(xclGraphTimedEnd_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.cycletimeout_)>(
          reinterpret_cast<char*>(&_impl_.cycletimeout_),
          reinterpret_cast<char*>(&other->_impl_.cycletimeout_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphTimedEnd_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[94]);
}

// ===================================================================

class xclGraphTimedEnd_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedEnd_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphTimedEnd_response::xclGraphTimedEnd_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedEnd_response)
}
xclGraphTimedEnd_response::xclGraphTimedEnd_response(const xclGraphTimedEnd_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphTimedEnd_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphTimedEnd_response)
}

inline void xclGraphTimedEnd_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphTimedEnd_response::~xclGraphTimedEnd_response() {
  // @@protoc_insertion_point(destructor:xclGraphTimedEnd_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphTimedEnd_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphTimedEnd_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphTimedEnd_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedEnd_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphTimedEnd_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphTimedEnd_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedEnd_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedEnd_response)
  return target;
}

size_t xclGraphTimedEnd_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedEnd_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphTimedEnd_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphTimedEnd_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphTimedEnd_response::GetClassData() const { return &_class_data_; }


void xclGraphTimedEnd_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedEnd_response*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedEnd_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedEnd_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedEnd_response::CopyFrom(const xclGraphTimedEnd_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedEnd_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphTimedEnd_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphTimedEnd_response::InternalSwap(xclGraphTimedEnd_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphTimedEnd_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[95]);
}

// ===================================================================

class xclGraphResume_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphResume_call>()._impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphResume_call::xclGraphResume_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphResume_call)
}
xclGraphResume_call::xclGraphResume_call(const xclGraphResume_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphResume_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.gh_ = from._impl_.gh_;
  // @@protoc_insertion_point(copy_constructor:xclGraphResume_call)
}

inline void xclGraphResume_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.gh_){0u}
  };
}

xclGraphResume_call::~xclGraphResume_call() {
  // @@protoc_insertion_point(destructor:xclGraphResume_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphResume_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphResume_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphResume_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphResume_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphResume_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required uint32 gh = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 16)) {
          _Internal::set_has_gh(&has_bits);
          _impl_.gh_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint32(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphResume_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphResume_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphResume_call)
  return target;
}

size_t xclGraphResume_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphResume_call)
  size_t total_size = 0;

  // required uint32 gh = 2;
  if (_internal_has_gh()) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(this->_internal_gh());
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphResume_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphResume_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphResume_call::GetClassData() const { return &_class_data_; }


void xclGraphResume_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphResume_call*>(&to_msg);
  auto& from = static_cast<const xclGraphResume_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphResume_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_gh()) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphResume_call::CopyFrom(const xclGraphResume_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphResume_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphResume_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphResume_call::InternalSwap(xclGraphResume_call* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.gh_, other->_impl_.gh_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphResume_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[96]);
}

// ===================================================================

class xclGraphResume_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphResume_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphResume_response::xclGraphResume_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclGraphResume_response)
}
xclGraphResume_response::xclGraphResume_response(const xclGraphResume_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclGraphResume_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclGraphResume_response)
}

inline void xclGraphResume_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclGraphResume_response::~xclGraphResume_response() {
  // @@protoc_insertion_point(destructor:xclGraphResume_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclGraphResume_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclGraphResume_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclGraphResume_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphResume_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclGraphResume_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclGraphResume_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphResume_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphResume_response)
  return target;
}

size_t xclGraphResume_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphResume_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclGraphResume_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclGraphResume_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclGraphResume_response::GetClassData() const { return &_class_data_; }


void xclGraphResume_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclGraphResume_response*>(&to_msg);
  auto& from = static_cast<const xclGraphResume_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphResume_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphResume_response::CopyFrom(const xclGraphResume_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphResume_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclGraphResume_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclGraphResume_response::InternalSwap(xclGraphResume_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclGraphResume_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[97]);
}

// ===================================================================

class xclLoadXclbinContent_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadXclbinContent_call>()._impl_._has_bits_);
  static void set_has_xmlbuff(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xmlbuffsize(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_sharedbin(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_sharedbinsize(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_emuldata(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_emuldatasize(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_keepdir(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000007f) ^ 0x0000007f) != 0;
  }
};

xclLoadXclbinContent_call::xclLoadXclbinContent_call(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclLoadXclbinContent_call)
}
xclLoadXclbinContent_call::xclLoadXclbinContent_call(const xclLoadXclbinContent_call& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclLoadXclbinContent_call* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xmlbuff_){}
    , decltype(_impl_.sharedbin_){}
    , decltype(_impl_.emuldata_){}
    , decltype(_impl_.xmlbuffsize_){}
    , decltype(_impl_.sharedbinsize_){}
    , decltype(_impl_.emuldatasize_){}
    , decltype(_impl_.keepdir_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _impl_.xmlbuff_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xmlbuff_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_xmlbuff()) {
    _this->_impl_.xmlbuff_.Set(from._internal_xmlbuff(), 
      _this->GetArenaForAllocation());
  }
  _impl_.sharedbin_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.sharedbin_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_sharedbin()) {
    _this->_impl_.sharedbin_.Set(from._internal_sharedbin(), 
      _this->GetArenaForAllocation());
  }
  _impl_.emuldata_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.emuldata_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  if (from._internal_has_emuldata()) {
    _this->_impl_.emuldata_.Set(from._internal_emuldata(), 
      _this->GetArenaForAllocation());
  }
  ::memcpy(&_impl_.xmlbuffsize_, &from._impl_.xmlbuffsize_,
    static_cast<size_t>(reinterpret_cast<char*>(&_impl_.keepdir_) -
    reinterpret_cast<char*>(&_impl_.xmlbuffsize_)) + sizeof(_impl_.keepdir_));
  // @@protoc_insertion_point(copy_constructor:xclLoadXclbinContent_call)
}

inline void xclLoadXclbinContent_call::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.xmlbuff_){}
    , decltype(_impl_.sharedbin_){}
    , decltype(_impl_.emuldata_){}
    , decltype(_impl_.xmlbuffsize_){uint64_t{0u}}
    , decltype(_impl_.sharedbinsize_){uint64_t{0u}}
    , decltype(_impl_.emuldatasize_){uint64_t{0u}}
    , decltype(_impl_.keepdir_){false}
  };
  _impl_.xmlbuff_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.xmlbuff_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.sharedbin_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.sharedbin_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
  _impl_.emuldata_.InitDefault();
  #ifdef PROTOBUF_FORCE_COPY_DEFAULT_STRING
    _impl_.emuldata_.Set("", GetArenaForAllocation());
  #endif // PROTOBUF_FORCE_COPY_DEFAULT_STRING
}

xclLoadXclbinContent_call::~xclLoadXclbinContent_call() {
  // @@protoc_insertion_point(destructor:xclLoadXclbinContent_call)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclLoadXclbinContent_call::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
  _impl_.xmlbuff_.Destroy();
  _impl_.sharedbin_.Destroy();
  _impl_.emuldata_.Destroy();
}

void xclLoadXclbinContent_call::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclLoadXclbinContent_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadXclbinContent_call)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xmlbuff_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.sharedbin_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.emuldata_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x00000078u) {
    ::memset(&_impl_.xmlbuffsize_, 0, static_cast<size_t>(
        reinterpret_cast<char*>(&_impl_.keepdir_) -
        reinterpret_cast<char*>(&_impl_.xmlbuffsize_)) + sizeof(_impl_.keepdir_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclLoadXclbinContent_call::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bytes xmlbuff = 2;
      case 2:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 18)) {
          auto str = _internal_mutable_xmlbuff();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 xmlbuffsize = 3;
      case 3:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 24)) {
          _Internal::set_has_xmlbuffsize(&has_bits);
          _impl_.xmlbuffsize_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes sharedbin = 4;
      case 4:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 34)) {
          auto str = _internal_mutable_sharedbin();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 sharedbinsize = 5;
      case 5:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 40)) {
          _Internal::set_has_sharedbinsize(&has_bits);
          _impl_.sharedbinsize_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bytes emuldata = 6;
      case 6:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 50)) {
          auto str = _internal_mutable_emuldata();
          ptr = ::_pbi::InlineGreedyStringParser(str, ptr, ctx);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required uint64 emuldatasize = 7;
      case 7:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 56)) {
          _Internal::set_has_emuldatasize(&has_bits);
          _impl_.emuldatasize_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      // required bool keepdir = 8;
      case 8:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 64)) {
          _Internal::set_has_keepdir(&has_bits);
          _impl_.keepdir_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclLoadXclbinContent_call::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadXclbinContent_call)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xmlbuff = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->WriteBytesMaybeAliased(
        2, this->_internal_xmlbuff(), target);
  }

  // required uint64 xmlbuffsize = 3;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(3, this->_internal_xmlbuffsize(), target);
  }

  // required bytes sharedbin = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->WriteBytesMaybeAliased(
        4, this->_internal_sharedbin(), target);
  }

  // required uint64 sharedbinsize = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(5, this->_internal_sharedbinsize(), target);
  }

  // required bytes emuldata = 6;
  if (cached_has_bits & 0x00000004u) {
    target = stream->WriteBytesMaybeAliased(
        6, this->_internal_emuldata(), target);
  }

  // required uint64 emuldatasize = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(7, this->_internal_emuldatasize(), target);
  }

  // required bool keepdir = 8;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(8, this->_internal_keepdir(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadXclbinContent_call)
  return target;
}

size_t xclLoadXclbinContent_call::RequiredFieldsByteSizeFallback() const {
// @@protoc_insertion_point(required_fields_byte_size_fallback_start:xclLoadXclbinContent_call)
  size_t total_size = 0;

  if (_internal_has_xmlbuff()) {
    // required bytes xmlbuff = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xmlbuff());
  }

  if (_internal_has_sharedbin()) {
    // required bytes sharedbin = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_sharedbin());
  }

  if (_internal_has_emuldata()) {
    // required bytes emuldata = 6;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_emuldata());
  }

  if (_internal_has_xmlbuffsize()) {
    // required uint64 xmlbuffsize = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_xmlbuffsize());
  }

  if (_internal_has_sharedbinsize()) {
    // required uint64 sharedbinsize = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_sharedbinsize());
  }

  if (_internal_has_emuldatasize()) {
    // required uint64 emuldatasize = 7;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_emuldatasize());
  }

  if (_internal_has_keepdir()) {
    // required bool keepdir = 8;
    total_size += 1 + 1;
  }

  return total_size;
}
size_t xclLoadXclbinContent_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadXclbinContent_call)
  size_t total_size = 0;

  if (((_impl_._has_bits_[0] & 0x0000007f) ^ 0x0000007f) == 0) {  // All required fields are present.
    // required bytes xmlbuff = 2;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_xmlbuff());

    // required bytes sharedbin = 4;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_sharedbin());

    // required bytes emuldata = 6;
    total_size += 1 +
      ::PROTOBUF_NAMESPACE_ID::internal::WireFormatLite::BytesSize(
        this->_internal_emuldata());

    // required uint64 xmlbuffsize = 3;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_xmlbuffsize());

    // required uint64 sharedbinsize = 5;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_sharedbinsize());

    // required uint64 emuldatasize = 7;
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(this->_internal_emuldatasize());

    // required bool keepdir = 8;
    total_size += 1 + 1;

  } else {
    total_size += RequiredFieldsByteSizeFallback();
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclLoadXclbinContent_call::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclLoadXclbinContent_call::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclLoadXclbinContent_call::GetClassData() const { return &_class_data_; }


void xclLoadXclbinContent_call::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclLoadXclbinContent_call*>(&to_msg);
  auto& from = static_cast<const xclLoadXclbinContent_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadXclbinContent_call)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xmlbuff(from._internal_xmlbuff());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_sharedbin(from._internal_sharedbin());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_emuldata(from._internal_emuldata());
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.xmlbuffsize_ = from._impl_.xmlbuffsize_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.sharedbinsize_ = from._impl_.sharedbinsize_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.emuldatasize_ = from._impl_.emuldatasize_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.keepdir_ = from._impl_.keepdir_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadXclbinContent_call::CopyFrom(const xclLoadXclbinContent_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadXclbinContent_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclLoadXclbinContent_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclLoadXclbinContent_call::InternalSwap(xclLoadXclbinContent_call* other) {
  using std::swap;
  auto* lhs_arena = GetArenaForAllocation();
  auto* rhs_arena = other->GetArenaForAllocation();
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.xmlbuff_, lhs_arena,
      &other->_impl_.xmlbuff_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.sharedbin_, lhs_arena,
      &other->_impl_.sharedbin_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::ArenaStringPtr::InternalSwap(
      &_impl_.emuldata_, lhs_arena,
      &other->_impl_.emuldata_, rhs_arena
  );
  ::PROTOBUF_NAMESPACE_ID::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.keepdir_)
      + sizeof(xclLoadXclbinContent_call::_impl_.keepdir_)
      - PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuffsize_)>(
          reinterpret_cast<char*>(&_impl_.xmlbuffsize_),
          reinterpret_cast<char*>(&other->_impl_.xmlbuffsize_));
}

::PROTOBUF_NAMESPACE_ID::Metadata xclLoadXclbinContent_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[98]);
}

// ===================================================================

class xclLoadXclbinContent_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadXclbinContent_response>()._impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclLoadXclbinContent_response::xclLoadXclbinContent_response(::PROTOBUF_NAMESPACE_ID::Arena* arena,
                         bool is_message_owned)
  : ::PROTOBUF_NAMESPACE_ID::Message(arena, is_message_owned) {
  SharedCtor(arena, is_message_owned);
  // @@protoc_insertion_point(arena_constructor:xclLoadXclbinContent_response)
}
xclLoadXclbinContent_response::xclLoadXclbinContent_response(const xclLoadXclbinContent_response& from)
  : ::PROTOBUF_NAMESPACE_ID::Message() {
  xclLoadXclbinContent_response* const _this = this; (void)_this;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){from._impl_._has_bits_}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){}};

  _internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
  _this->_impl_.ack_ = from._impl_.ack_;
  // @@protoc_insertion_point(copy_constructor:xclLoadXclbinContent_response)
}

inline void xclLoadXclbinContent_response::SharedCtor(
    ::_pb::Arena* arena, bool is_message_owned) {
  (void)arena;
  (void)is_message_owned;
  new (&_impl_) Impl_{
      decltype(_impl_._has_bits_){}
    , /*decltype(_impl_._cached_size_)*/{}
    , decltype(_impl_.ack_){false}
  };
}

xclLoadXclbinContent_response::~xclLoadXclbinContent_response() {
  // @@protoc_insertion_point(destructor:xclLoadXclbinContent_response)
  if (auto *arena = _internal_metadata_.DeleteReturnArena<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>()) {
  (void)arena;
    return;
  }
  SharedDtor();
}

inline void xclLoadXclbinContent_response::SharedDtor() {
  GOOGLE_DCHECK(GetArenaForAllocation() == nullptr);
}

void xclLoadXclbinContent_response::SetCachedSize(int size) const {
  _impl_._cached_size_.Set(size);
}

void xclLoadXclbinContent_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadXclbinContent_response)
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>();
}

const char* xclLoadXclbinContent_response::_InternalParse(const char* ptr, ::_pbi::ParseContext* ctx) {
#define CHK_(x) if (PROTOBUF_PREDICT_FALSE(!(x))) goto failure
  _Internal::HasBits has_bits{};
  while (!ctx->Done(&ptr)) {
    uint32_t tag;
    ptr = ::_pbi::ReadTag(ptr, &tag);
    switch (tag >> 3) {
      // required bool ack = 1;
      case 1:
        if (PROTOBUF_PREDICT_TRUE(static_cast<uint8_t>(tag) == 8)) {
          _Internal::set_has_ack(&has_bits);
          _impl_.ack_ = ::PROTOBUF_NAMESPACE_ID::internal::ReadVarint64(&ptr);
          CHK_(ptr);
        } else
          goto handle_unusual;
        continue;
      default:
        goto handle_unusual;
    }  // switch
  handle_unusual:
    if ((tag == 0) || ((tag & 7) == 4)) {
      CHK_(ptr);
      ctx->SetLastTag(tag);
      goto message_done;
    }
    ptr = UnknownFieldParse(
        tag,
        _internal_metadata_.mutable_unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(),
        ptr, ctx);
    CHK_(ptr != nullptr);
  }  // while
message_done:
  _impl_._has_bits_.Or(has_bits);
  return ptr;
failure:
  ptr = nullptr;
  goto message_done;
#undef CHK_
}

uint8_t* xclLoadXclbinContent_response::_InternalSerialize(
    uint8_t* target, ::PROTOBUF_NAMESPACE_ID::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadXclbinContent_response)
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target = ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
        _internal_metadata_.unknown_fields<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(::PROTOBUF_NAMESPACE_ID::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadXclbinContent_response)
  return target;
}

size_t xclLoadXclbinContent_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadXclbinContent_response)
  size_t total_size = 0;

  // required bool ack = 1;
  if (_internal_has_ack()) {
    total_size += 1 + 1;
  }
  uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::PROTOBUF_NAMESPACE_ID::Message::ClassData xclLoadXclbinContent_response::_class_data_ = {
    ::PROTOBUF_NAMESPACE_ID::Message::CopyWithSourceCheck,
    xclLoadXclbinContent_response::MergeImpl
};
const ::PROTOBUF_NAMESPACE_ID::Message::ClassData*xclLoadXclbinContent_response::GetClassData() const { return &_class_data_; }


void xclLoadXclbinContent_response::MergeImpl(::PROTOBUF_NAMESPACE_ID::Message& to_msg, const ::PROTOBUF_NAMESPACE_ID::Message& from_msg) {
  auto* const _this = static_cast<xclLoadXclbinContent_response*>(&to_msg);
  auto& from = static_cast<const xclLoadXclbinContent_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadXclbinContent_response)
  GOOGLE_DCHECK_NE(&from, _this);
  uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if (from._internal_has_ack()) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::PROTOBUF_NAMESPACE_ID::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadXclbinContent_response::CopyFrom(const xclLoadXclbinContent_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadXclbinContent_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

bool xclLoadXclbinContent_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) return false;
  return true;
}

void xclLoadXclbinContent_response::InternalSwap(xclLoadXclbinContent_response* other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  swap(_impl_.ack_, other->_impl_.ack_);
}

::PROTOBUF_NAMESPACE_ID::Metadata xclLoadXclbinContent_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[99]);
}

// @@protoc_insertion_point(namespace_scope)
PROTOBUF_NAMESPACE_OPEN
template<> PROTOBUF_NOINLINE ::call_packet_info*
Arena::CreateMaybeMessage< ::call_packet_info >(Arena* arena) {
  return Arena::CreateMessageInternal< ::call_packet_info >(arena);
}
template<> PROTOBUF_NOINLINE ::response_packet_info*
Arena::CreateMaybeMessage< ::response_packet_info >(Arena* arena) {
  return Arena::CreateMessageInternal< ::response_packet_info >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetEnvironment_call_namevaluepair*
Arena::CreateMaybeMessage< ::xclSetEnvironment_call_namevaluepair >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetEnvironment_call_namevaluepair >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetEnvironment_call*
Arena::CreateMaybeMessage< ::xclSetEnvironment_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetEnvironment_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetEnvironment_response*
Arena::CreateMaybeMessage< ::xclSetEnvironment_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetEnvironment_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclLoadBitstream_call_ddrbank*
Arena::CreateMaybeMessage< ::xclLoadBitstream_call_ddrbank >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclLoadBitstream_call_ddrbank >(arena);
}
template<> PROTOBUF_NOINLINE ::xclLoadBitstream_call*
Arena::CreateMaybeMessage< ::xclLoadBitstream_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclLoadBitstream_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclLoadBitstream_response*
Arena::CreateMaybeMessage< ::xclLoadBitstream_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclLoadBitstream_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclAllocDeviceBuffer_call*
Arena::CreateMaybeMessage< ::xclAllocDeviceBuffer_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclAllocDeviceBuffer_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclAllocDeviceBuffer_response*
Arena::CreateMaybeMessage< ::xclAllocDeviceBuffer_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclAllocDeviceBuffer_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclFreeDeviceBuffer_call*
Arena::CreateMaybeMessage< ::xclFreeDeviceBuffer_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclFreeDeviceBuffer_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclFreeDeviceBuffer_response*
Arena::CreateMaybeMessage< ::xclFreeDeviceBuffer_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclFreeDeviceBuffer_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclClose_call*
Arena::CreateMaybeMessage< ::xclClose_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclClose_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclClose_response*
Arena::CreateMaybeMessage< ::xclClose_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclClose_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBufferHost2Device_call*
Arena::CreateMaybeMessage< ::xclCopyBufferHost2Device_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBufferHost2Device_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBufferHost2Device_response*
Arena::CreateMaybeMessage< ::xclCopyBufferHost2Device_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBufferHost2Device_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBufferDevice2Host_call*
Arena::CreateMaybeMessage< ::xclCopyBufferDevice2Host_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBufferDevice2Host_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBufferDevice2Host_response*
Arena::CreateMaybeMessage< ::xclCopyBufferDevice2Host_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBufferDevice2Host_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteAddrSpaceDeviceRam_call*
Arena::CreateMaybeMessage< ::xclWriteAddrSpaceDeviceRam_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteAddrSpaceDeviceRam_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteAddrSpaceDeviceRam_response*
Arena::CreateMaybeMessage< ::xclWriteAddrSpaceDeviceRam_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteAddrSpaceDeviceRam_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteAddrKernelCtrl_call_kernelInfo*
Arena::CreateMaybeMessage< ::xclWriteAddrKernelCtrl_call_kernelInfo >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteAddrKernelCtrl_call_kernelInfo >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteAddrKernelCtrl_call*
Arena::CreateMaybeMessage< ::xclWriteAddrKernelCtrl_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteAddrKernelCtrl_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteAddrKernelCtrl_response*
Arena::CreateMaybeMessage< ::xclWriteAddrKernelCtrl_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteAddrKernelCtrl_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclRegWrite_call*
Arena::CreateMaybeMessage< ::xclRegWrite_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclRegWrite_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclRegWrite_response*
Arena::CreateMaybeMessage< ::xclRegWrite_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclRegWrite_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadAddrSpaceDeviceRam_call*
Arena::CreateMaybeMessage< ::xclReadAddrSpaceDeviceRam_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadAddrSpaceDeviceRam_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadAddrSpaceDeviceRam_response*
Arena::CreateMaybeMessage< ::xclReadAddrSpaceDeviceRam_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadAddrSpaceDeviceRam_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadAddrKernelCtrl_call*
Arena::CreateMaybeMessage< ::xclReadAddrKernelCtrl_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadAddrKernelCtrl_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadAddrKernelCtrl_response*
Arena::CreateMaybeMessage< ::xclReadAddrKernelCtrl_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadAddrKernelCtrl_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclRegRead_call*
Arena::CreateMaybeMessage< ::xclRegRead_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclRegRead_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclRegRead_response*
Arena::CreateMaybeMessage< ::xclRegRead_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclRegRead_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadCounters_call*
Arena::CreateMaybeMessage< ::xclPerfMonReadCounters_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadCounters_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadCounters_response*
Arena::CreateMaybeMessage< ::xclPerfMonReadCounters_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadCounters_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonGetTraceCount_call*
Arena::CreateMaybeMessage< ::xclPerfMonGetTraceCount_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonGetTraceCount_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonGetTraceCount_response*
Arena::CreateMaybeMessage< ::xclPerfMonGetTraceCount_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonGetTraceCount_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_call*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_response_events*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_response_events >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_response_events >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_response*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_response >(arena);
}
template<> PROTOBUF_NOINLINE ::ddr_mem_msg*
Arena::CreateMaybeMessage< ::ddr_mem_msg >(Arena* arena) {
  return Arena::CreateMessageInternal< ::ddr_mem_msg >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGetDeviceTimestamp_call*
Arena::CreateMaybeMessage< ::xclGetDeviceTimestamp_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGetDeviceTimestamp_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGetDeviceTimestamp_response*
Arena::CreateMaybeMessage< ::xclGetDeviceTimestamp_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGetDeviceTimestamp_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadBusStatus_call*
Arena::CreateMaybeMessage< ::xclReadBusStatus_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadBusStatus_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadBusStatus_response*
Arena::CreateMaybeMessage< ::xclReadBusStatus_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadBusStatus_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGetDebugMessages_call*
Arena::CreateMaybeMessage< ::xclGetDebugMessages_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGetDebugMessages_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGetDebugMessages_response*
Arena::CreateMaybeMessage< ::xclGetDebugMessages_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGetDebugMessages_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclImportBO_call*
Arena::CreateMaybeMessage< ::xclImportBO_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclImportBO_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclImportBO_response*
Arena::CreateMaybeMessage< ::xclImportBO_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclImportBO_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBO_call*
Arena::CreateMaybeMessage< ::xclCopyBO_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBO_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCopyBO_response*
Arena::CreateMaybeMessage< ::xclCopyBO_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCopyBO_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCreateQueue_call*
Arena::CreateMaybeMessage< ::xclCreateQueue_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCreateQueue_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclCreateQueue_response*
Arena::CreateMaybeMessage< ::xclCreateQueue_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclCreateQueue_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteQueue_call*
Arena::CreateMaybeMessage< ::xclWriteQueue_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteQueue_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclWriteQueue_response*
Arena::CreateMaybeMessage< ::xclWriteQueue_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclWriteQueue_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadQueue_call*
Arena::CreateMaybeMessage< ::xclReadQueue_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadQueue_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclReadQueue_response*
Arena::CreateMaybeMessage< ::xclReadQueue_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclReadQueue_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPollCompletion_call*
Arena::CreateMaybeMessage< ::xclPollCompletion_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPollCompletion_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPollCompletion_response_request*
Arena::CreateMaybeMessage< ::xclPollCompletion_response_request >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPollCompletion_response_request >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPollCompletion_response*
Arena::CreateMaybeMessage< ::xclPollCompletion_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPollCompletion_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclDestroyQueue_call*
Arena::CreateMaybeMessage< ::xclDestroyQueue_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclDestroyQueue_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclDestroyQueue_response*
Arena::CreateMaybeMessage< ::xclDestroyQueue_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclDestroyQueue_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetupInstance_call_argflowpair*
Arena::CreateMaybeMessage< ::xclSetupInstance_call_argflowpair >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetupInstance_call_argflowpair >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetupInstance_call*
Arena::CreateMaybeMessage< ::xclSetupInstance_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetupInstance_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSetupInstance_response*
Arena::CreateMaybeMessage< ::xclSetupInstance_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSetupInstance_response >(arena);
}
template<> PROTOBUF_NOINLINE ::swemuDriverVersion_call*
Arena::CreateMaybeMessage< ::swemuDriverVersion_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::swemuDriverVersion_call >(arena);
}
template<> PROTOBUF_NOINLINE ::swemuDriverVersion_response*
Arena::CreateMaybeMessage< ::swemuDriverVersion_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::swemuDriverVersion_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadCounters_Streaming_call*
Arena::CreateMaybeMessage< ::xclPerfMonReadCounters_Streaming_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadCounters_Streaming_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadCounters_Streaming_response*
Arena::CreateMaybeMessage< ::xclPerfMonReadCounters_Streaming_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadCounters_Streaming_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_Streaming_call*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_Streaming_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_Streaming_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_Streaming_response_events*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_Streaming_response_events >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_Streaming_response_events >(arena);
}
template<> PROTOBUF_NOINLINE ::xclPerfMonReadTrace_Streaming_response*
Arena::CreateMaybeMessage< ::xclPerfMonReadTrace_Streaming_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclPerfMonReadTrace_Streaming_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSlaveReadReq_call*
Arena::CreateMaybeMessage< ::xclSlaveReadReq_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSlaveReadReq_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSlaveReadReq_response*
Arena::CreateMaybeMessage< ::xclSlaveReadReq_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSlaveReadReq_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSlaveWriteReq_call*
Arena::CreateMaybeMessage< ::xclSlaveWriteReq_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSlaveWriteReq_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSlaveWriteReq_response*
Arena::CreateMaybeMessage< ::xclSlaveWriteReq_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSlaveWriteReq_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclInterruptOccured_call*
Arena::CreateMaybeMessage< ::xclInterruptOccured_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclInterruptOccured_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclInterruptOccured_response*
Arena::CreateMaybeMessage< ::xclInterruptOccured_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclInterruptOccured_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphInit_call*
Arena::CreateMaybeMessage< ::xclGraphInit_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphInit_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphInit_response*
Arena::CreateMaybeMessage< ::xclGraphInit_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphInit_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphRun_call*
Arena::CreateMaybeMessage< ::xclGraphRun_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphRun_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphRun_response*
Arena::CreateMaybeMessage< ::xclGraphRun_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphRun_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphWait_call*
Arena::CreateMaybeMessage< ::xclGraphWait_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphWait_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphWait_response*
Arena::CreateMaybeMessage< ::xclGraphWait_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphWait_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphEnd_call*
Arena::CreateMaybeMessage< ::xclGraphEnd_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphEnd_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphEnd_response*
Arena::CreateMaybeMessage< ::xclGraphEnd_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphEnd_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphUpdateRTP_call*
Arena::CreateMaybeMessage< ::xclGraphUpdateRTP_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphUpdateRTP_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphUpdateRTP_response*
Arena::CreateMaybeMessage< ::xclGraphUpdateRTP_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphUpdateRTP_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphReadRTP_call*
Arena::CreateMaybeMessage< ::xclGraphReadRTP_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphReadRTP_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphReadRTP_response*
Arena::CreateMaybeMessage< ::xclGraphReadRTP_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphReadRTP_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSyncBOAIENB_call*
Arena::CreateMaybeMessage< ::xclSyncBOAIENB_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSyncBOAIENB_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclSyncBOAIENB_response*
Arena::CreateMaybeMessage< ::xclSyncBOAIENB_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclSyncBOAIENB_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGMIOWait_call*
Arena::CreateMaybeMessage< ::xclGMIOWait_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGMIOWait_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGMIOWait_response*
Arena::CreateMaybeMessage< ::xclGMIOWait_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGMIOWait_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphTimedWait_call*
Arena::CreateMaybeMessage< ::xclGraphTimedWait_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphTimedWait_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphTimedWait_response*
Arena::CreateMaybeMessage< ::xclGraphTimedWait_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphTimedWait_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphTimedEnd_call*
Arena::CreateMaybeMessage< ::xclGraphTimedEnd_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphTimedEnd_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphTimedEnd_response*
Arena::CreateMaybeMessage< ::xclGraphTimedEnd_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphTimedEnd_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphResume_call*
Arena::CreateMaybeMessage< ::xclGraphResume_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphResume_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclGraphResume_response*
Arena::CreateMaybeMessage< ::xclGraphResume_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclGraphResume_response >(arena);
}
template<> PROTOBUF_NOINLINE ::xclLoadXclbinContent_call*
Arena::CreateMaybeMessage< ::xclLoadXclbinContent_call >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclLoadXclbinContent_call >(arena);
}
template<> PROTOBUF_NOINLINE ::xclLoadXclbinContent_response*
Arena::CreateMaybeMessage< ::xclLoadXclbinContent_response >(Arena* arena) {
  return Arena::CreateMessageInternal< ::xclLoadXclbinContent_response >(arena);
}
PROTOBUF_NAMESPACE_CLOSE

// @@protoc_insertion_point(global_scope)
#include <google/protobuf/port_undef.inc>
